/*------------------------------------------------------------------------------------------
//  Creation Date:  06/22/23
//  Author:		Greg Smith (GKS)
//  Name:		imvuToolsInstaller.ms
//  Description:  IMVU tools installer script.
					Removes all existing toolkit files so there is a clean new install.
					Copies all of the new files to the user's AppData\Local\Autodesk\3dsMax\<version>\<lang> folder
					Updates\runs imvu menu
//  Limitations:
//  Revisions:
------------------------------------------------------------------------------------------*/
format "Imvu Tools Installer script loading\n"

(
	-- DEBUGGING
	local PRINTDEBUG = false

	local THIS_SCRIPT_FILEPATH = getFilenamePath(getThisScriptFilename())
	local THIS_SCRIPT_FILENAME = filenameFromPath(getThisScriptFilename())
	-- Install root is always expected to be "C:\Users\<username>\AppData\Local\Autodesk\3dsMax\<max_version>\<language>\"
	local INSTALLROOT = getFilenamePath(getDir #userScripts)
	local INSTALLER_INIFILE = getFilenamePath(getThisScriptFilename()) + "imvuToolsInstaller.ini"
	local IMVU_MENU_SCRIPTPATH = "scripts/imvu/ImvuMenu.ms"
	local IMVU_STARTUP_SCRIPTPATH = "scripts/startup/ImvuStartup.ms"

	-- Get data from installer INI
	local cleanDirs = for n in (getINISetting INSTALLER_INIFILE "CleanDirs") collect getINISetting INSTALLER_INIFILE "CleanDirs" n
	local cleanFiles = for n in (getINISetting INSTALLER_INIFILE "cleanFiles") collect getINISetting INSTALLER_INIFILE "cleanFiles" n
	local fileSources = for n in (getINISetting INSTALLER_INIFILE "Source") collect getINISetting INSTALLER_INIFILE "Source" n
	local fileDestinations = for n in (getINISetting INSTALLER_INIFILE "Destination") collect getINISetting INSTALLER_INIFILE "Destination" n

	-- Define regular expressions object
	local rgx = dotNetClass "System.Text.RegularExpressions.RegEx"

	---------- Installer Support Functions
	function printMsg msg msgType: functionName: =
	(
		/*
		Description:
			Prints a message to the console with the optional message type and function name from where it was called.
		Arguments:
			<string> msg: message string
			<name> msgType [optional]: Valid types are #ERROR, #WARN, #INFO. Defaults to no label
			<string> functionName [optional]: function name from where it was called
		Return:
			None
		*/
		local msgLabel
		case msgType of
		(
			#ERROR: msgLabel = "ERROR:"
			#WARN: msgLabel = "WARNING:"
			#INFO: msgLabel = "INFO:"
			#DEBUG: msgLabel = "DEBUG:"
			default: msgLabel = ""
		)
		if (PRINTDEBUG and msgType == #DEBUG) or msgType != #DEBUG then
		(
			if functionName == unsupplied then
				format "% % %\n" THIS_SCRIPT_FILENAME msgLabel msg
			else
				format "%::% % %\n" THIS_SCRIPT_FILENAME functionName msgLabel msg
		)
	)
		

	function isEmptyString theString =
	(
		/*
		Description:
			Check supplied string for empty conditions. Will also return true if the supplied string is undefined
		Arguments:
			<string> theString: String to be evaluated
		Return:
			<boolean> true is empty conditions are met, false if not.
		*/
		local ret = false
		local whitespacePattern = "^\s*$"
		local match
		if theString == undefined then ret = true
		else
		(
			match = rgx.Match theString whitespacePattern
			if match.value.count > 0 or theString == "" then ret = true
		)
		ret
	)


	fn getFilesRecursive root pattern =
	(
		/*
		Description:
			Looks for a file pattern and returns an array of file paths
		Arguments:
			<string> root: The root directory to start the search in
			<string> pattern: The pattern to look for
		Return:
			<array> of strings of file paths found
		*/
		local dirArray = GetDirectories (root+"/*")
		local filePaths = (getFiles (root+"/"+pattern))
		for dir in dirArray do
			join dirArray (GetDirectories (dir+"/*"))
		for file in dirArray do
			join filePaths (getFiles (file + pattern))
		-- Make sure our file path is Windows formatted
		for i = 1 to filePaths.count do
			filePaths[i] = (substituteString filePaths[i] "/" "\\")
		filePaths
	)


	function delTree dir =
	(
		/*
		Description:
			Deletes a directory and all of its contents recursively
		Arguments:
			<string> dir: The name of the directory to be deleted
		Return:
			<boolean> ret: true is everything executed without errors, false if errors
		TODO: FIX: This will return false if the 'dir' is a file, but will still perform the delete.
		*/
		local THIS_FUNCTION_NAME = (delTree as string)
		local fullDirPath = INSTALLROOT + dir
		local ret = false
		local cmd = "del /f/s/q \"" + dir + "\" 1>nul && rmdir /s/q \"" + dir + "\""

		if doesFileExist fullDirPath then
		(
			hiddenDOSCommand cmd startpath:INSTALLROOT exitCode:&exitCode
			if exitCode != 0 then printMsg ("Something went wrong deleting directory tree: " + fullDirPath) msgType:#ERROR functionName:THIS_FUNCTION_NAME
			else
			(
				printMsg ("Deleted " + fullDirPath) msgType:#DEBUG functionName:THIS_FUNCTION_NAME
				ret = true
			)
		)
		else
			printMsg ("Directory " + fullDirPath + " does not exist") msgType:#INFO functionName:THIS_FUNCTION_NAME
		ret
	)


	function delFile fileSpec =
	(
		/*
		Description:
			Tries to deletes a file with or with a path. If fails then looks in INSTALLROOT recursively for the file to delete and does some error reporting if there is a problem or FNF.
		Arguments:
			<string> fileSpec: The file, file path, or pattern to be deleted
		Return:
			<boolean> ret: true is everything executed without errors, false if errors
		*/
		local THIS_FUNCTION_NAME = (delFile as string)
		local file = ""
		local filesToDelete = #()
		local ret = false
		local cmd = ""

		join filesToDelete (getFilesRecursive INSTALLROOT fileSpec)
		if filesToDelete.count > 0 then
		(
			for file in filesToDelete do
			(
				cmd = "del /f/q \"" + file + "\""
				hiddenDOSCommand cmd startpath:INSTALLROOT exitCode:&exitCode
				if exitCode != 0 then printMsg ("Something went wrong deleting file: " + file) msgType:#ERROR functionName:THIS_FUNCTION_NAME
				else
				(
					printMsg ("Deleted " + file) msgType:#DEBUG functionName:THIS_FUNCTION_NAME
					ret = true
				)
			)
		)
		else
			printMsg ("File path or pattern " + fileSpec + " does not exist in " + INSTALLROOT) msgType:#INFO functionName:THIS_FUNCTION_NAME
		ret
	)


	function copyFiles sourcePath destinationPath =
	(
		/*
		Description:
			Preforms system copy and error reporting if there is a problem
		Arguments:
			<string> sourcePath: The file spec to be copied
			<string> destinationPath: The path the file(s) will be copied to
		Return:
			<boolean> ret: true is everything executed without errors, false if errors
		*/
		local THIS_FUNCTION_NAME = (copyFiles as string)
		local fullSourcePath = THIS_SCRIPT_FILEPATH + sourcePath
		local fullDestinationPath = INSTALLROOT + destinationPath
		local ret = false
		local cmd = "copy /y \"" + sourcePath + "\" \"" + fullDestinationPath + "\""

		if (getFiles fullSourcePath).count > 0 or doesFileExist fullSourcePath then
		(
			ret = makeDir fullDestinationPath
			if ret == true then
			(
				hiddenDOSCommand cmd startpath:THIS_SCRIPT_FILEPATH exitCode:&exitCode
				if exitCode != 0 then
				(
					printMsg ("Something went wrong copying " + fullSourcePath + " to " + fullDestinationPath) msgType:#ERROR functionName:THIS_FUNCTION_NAME
					ret = false
				)
				else
					printMsg ("Copied " + fullSourcePath + " to " + fullDestinationPath) msgType:#DEBUG functionName:THIS_FUNCTION_NAME
			)
		)
		else
			printMsg ("Source file path " + fullSourcePath + " does not exist") msgType:#INFO functionName:THIS_FUNCTION_NAME		
		ret
	)


	printMsg THIS_SCRIPT_FILEPATH msgType:#DEBUG

	-- Destroy/close the Toolkit if it currently running
	try (closeRolloutFloater imvuStudioToolkitFloater) catch()

	-- Clean out existing imvu files
	for i = 1 to cleanDirs.count do delTree(cleanDirs[i])
	for i = 1 to cleanFiles.count do delFile(cleanFiles[i])

	-- Install all new imvu files
	if fileSources.count > 0 and fileDestinations.count > 0 and fileSources.count == fileDestinations.count then
	(
		for i = 1 to fileSources.count do
		(
			if isEmptyString(fileSources[i]) then printMsg ("There is no source file specified for index [" + (i as string) + "] in the " + INSTALLER_INIFILE) msgType:#WARN
			else copyFiles fileSources[i] fileDestinations[i]
		)
	)
	else printMsg ("There is a problem with the source files and/or file destinations in the " + INSTALLER_INIFILE + "\nfileSources.count=" + (fileSources.count as string) + "\nfileDestinations.count=" + (fileDestinations.count as string)) msgType:#ERROR

	-- Evaluate Imvu startup script and create/update Imvu menu
	fileIn (INSTALLROOT + IMVU_STARTUP_SCRIPTPATH)
	fileIn (INSTALLROOT + IMVU_MENU_SCRIPTPATH)
)