/*
	-- TODO FUTURE:
	-- TODO: Improvement: Export should be binary FBX with maybe an option to export ASCII.
	-- DONE: Improvement: Installer should close any active toolkit (and reopen it?).
	-- TODO: Improvement: Investigate adding support for multiple animations in a single FBX.
	-- TODO: Improvement: Support multiple UV channels and consolidate them upon export.
	-- TODO: Attachment tool, check to make sure the selection does not include AttachmentRoot - Currently this throws an error.
	-- TODO: (maybe): Add Export options: All, and checkboxes for the individual tools that are active. i.e. an EXPORT_<too> layer has been created.
	-- TODO: Make resize room work to resize the room after creation.
	-- TODO: See-through node is currently a Cone. Cone needs to be excluded from export, but a bone of the same name needs to be exported.
	-- DONE: Add a Reset button to the Attachment Tool which sets the AttachmentRoot back to 0,0,0 pos and rot and removes the Skin modifier to all of the meshes that have been affected.
	-- TODO: Add pre-export checks:
	--			Number of meshes, Marvelous designer dozens of mesh parts. Would be good to consolidate the meshes (for clothing Tool).
	--			Material IDs
	--			Multiple UV channels
	--			Mesh poly limits
	--			Texture sizes
	-- TODO: Save Toolkit and window size and position
*/
/*------------------------------------------------------------------------------------------
//  Creation Date:  11/8/22
//  Author:		Greg Smith (gks)
//  Name:		ImvuStudioToolkit.ms
//  Description:  This script provides all of the functionality of the IMVU Studio Toolkit
//  Guidelines from Autodesk:
The ordering of user-interface items and functions and sub-rollouts in a utility is important when referencing other items and rollouts in handler or local function code.
Specifically, you can only reference a user-interface item or rollout or function in a handler or other function if that item or rollout or function has been defined earlier in the utility.

A recommended ordering can help ensure this:
	local variables
	structs
	user-interface items
	nested rollouts
	functions
	event handlers
In some situations, cross-referencing means there's no way to order definitions and still pre-define everything.
To remedy this, MAXScript lets you pre-declare local rollouts and functions, so code that comes before the function or rollout definition will see the correct object.
You do this by declaring such rollouts and functions as uninitialized locals.
By pre-declaring them as locals in the main utility, cross-referencing is possible.

//  NOTES:
	-- TOOLKIT TOOL STATE MANAGEMENT
	-- Each Tool will have 3 states
	-- 	DISABLED - Means that the user does not have access to the tool.
	--		Determined by the presence any active tool NOT in the tool's compatibility list.
	-- 	ENABLED - Means that the user has not yet initiated the tool.
	--		An enabled tool will become disabled if it is NOT in the compatibility list of a tool that becomes active.
	-- 	ACTIVE - Means that the user has initiated the tool.
	--		Tools cannot be deactivated or disabled once active.
	-- Through normal use of the toolkit individual tools will transition from ENABLED to either ACTIVE or DISABLED
	--		The Visuals and Export Tools (rollouts) are exceptions. They are always active.

	-- Each tool will have compatibility list that determines which tools can be active at the same time.
	-- 	Right now the assumption is that Room and Furniture tools are NOT compatible (only one will be ACTIVE at a time).
	--		The functions addSeat and createSeeThroughNode DOES distinguish between the tool calling the functions, so there is no reason they could not function together.

	-- Determining tools states and setting UI and scene is handled by the function	updateToolkitSceneAndUIState():
	--	Updates happen: At Toolkit startup; after file new, open, save; when user starts any individual tool.
------------------------------------------------------------------------------------------*/
(
	-- CONSTANTS ----------------------------------------------------

	local PRINT_DEBUG_ALL = false
	local THIS_SCRIPT_FILENAME = filenameFromPath(getThisScriptFilename())

	-- Callback param integer for "normal save" during File Notification callback #filePreOpenProcess https://help.autodesk.com/view/3DSMAX/2017/ENU/?guid=__files_GUID_C1F6495F_5831_4FC8_A00C_667C5F2EAE36_htm
	local CALLBACKPARAM_NORMALSAVE = 1

	-- Tool version and Update constants
	-- LOCAL_VERSION is the source of truth for the current version of the toolkit.
	--	This version number is matched against what is on the CEC "https://create.imvu.com/api/studio/3dsmax"
	--	The remote version number gets updated manually when a new version of the toolkit is released
	--	It is updated via the WordPress CEC Dashboard > IMVU APIs interface. "https://create.imvu.com/wp-login.php"
	local LOCAL_VERSION = "1.1.6"
	local updateAfter_FILESAVE = 0
	local updateAfter_FILEOPEN = 1
	local updateAfter_FILENEW = 2
	local updateAfter_TOOLSTART = 3

	-- Window default size
	local DEFAULT_WINDOW_WIDTH = 300
	local DEFAULT_WINDOW_HEIGHT = 700

	-- Mega scene constants
	local MEGASCENE = (substituteString (substituteString (getDir #userScripts) "scripts" "imvu") "\\" "/") + "/IMVUMegaScene.max"
	local MEGASCENE_GROUPS = #("AvatarRig")
	local megaSceneGroup_AVATAR_RIG = 1

	local MEGASCENE_NODE_NAME_FEMALE_MESH = "Female_"
	local MEGASCENE_NODE_NAME_MALE_MESH = "Male_"
	local MEGASCENE_NODE_NAME_PROXY_MESH = "Proxy_"
	local MEGASCENE_NODE_NAME_AVATAR_SKELETON_ROOT = "Female03MasterRoot"
	local MEGASCENE_NODE_NAME_RIG_ATTRIBUTES = "_RigAttributes"
	local MEGASCENE_NODE_NAME_PREFIX_CONTROL_BONE = "Control_"
	local MEGASCENE_NODE_NAME_CONTROL_SKELETON_ROOT = (MEGASCENE_NODE_NAME_PREFIX_CONTROL_BONE + MEGASCENE_NODE_NAME_AVATAR_SKELETON_ROOT)
	local MEGASCENE_NODE_NAME_PREFIX_RIG_WIDGETS = "WGT_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_IK_LEFT_ARM_WIDGETS = "WGT_IK_lfArm_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_IK_RIGHT_ARM_WIDGETS = "WGT_IK_rtArm_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_IK_LEFT_LEG_WIDGETS = "WGT_IK_lfLeg_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_IK_RIGHT_LEG_WIDGETS = "WGT_IK_rtLeg_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_FK_LEFT_ARM_WIDGETS = "WGT_FK_lfArm_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_FK_RIGHT_ARM_WIDGETS = "WGT_FK_rtArm_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_FK_LEFT_LEG_WIDGETS = "WGT_FK_lfLeg_"
	local MEGASCENE_NODE_NAME_PREFIX_RIG_FK_RIGHT_LEG_WIDGETS = "WGT_FK_rtLeg_"

	-- Mega scene INI
	local MEGASCENE_INIFILE = (substituteString (substituteString (getDir #userScripts) "scripts" "imvu") "\\" "/") + "/IMVUMegaScene.ini"

	-- General constants
	local const_SOURCE = 1
	local const_TARGET = 2

	local WORLD_TRANSFORM_MATRIX = matrix3 1

	-- Tool id constants
	local toolId_EXPORT = 0
	local toolId_VISUALS = 1
	local toolId_ROOM = 2
	local toolId_FURNITURE = 3
	local toolId_ATTACHMENT = 4
	local toolId_CLOTHING = 5
	local toolId_AVATARANIM = 6

	-- Tool name constants
	local TOOLKIT_NAME = "IMVU Studio Toolkit"
	local TOOLNAME_ROOM = "Room Tool"
	local TOOLNAME_FURNITURE = "Furniture Tool"
	local TOOLNAME_ATTACHMENT = "Attachment Tool"
	local TOOLNAME_CLOTHING = "Clothing Tool"
	local TOOLNAME_AVATARANIM = "Avatar Animation Tool"
	local TOOLNAME_EXPORT = "Export Tool"

	local EXPORT_LAYER_NAME_ROOM = "ROOM_EXPORT"
	local EXPORT_LAYER_NAME_FURNITURE = "FURNITURE_EXPORT"
	local EXPORT_LAYER_NAME_ATTACHMENT = "ATTACHMENT_EXPORT"
	local EXPORT_LAYER_NAME_CLOTHING = "CLOTHING_EXPORT"
	local EXPORT_LAYER_NAME_AVATAR_ANIMATION = "AVATAR_ANIMATION_EXPORT"
	local exportLayerNames = #(
		EXPORT_LAYER_NAME_ROOM,
		EXPORT_LAYER_NAME_FURNITURE,
		EXPORT_LAYER_NAME_ATTACHMENT,
		EXPORT_LAYER_NAME_CLOTHING,
		EXPORT_LAYER_NAME_AVATAR_ANIMATION)

	-- EXPORT SCALE - This is a tricky one for several reasons:
	--		The IMVU avatar is approximately 11.5 meters tall or about 5x actual human height
	--		Historically, IMVU has supplied 3ds Max starter files that are 100x in scale and the system unit set to meters. That is the avatar is approximately 1150 meters tall.
	--		Historically, IMVU has supplied 3ds Max starter files that have the system unit set to meters.
	--			With this set up the fbxExporterSetParam "ScaleFactor" can be left DEFAULT of 1.0 and the assets imported into Studio will appear 1:1 with how
	--			they were created with Scale set to 1.0 in the Studio FBX Import dialog.
	-- 		Blender system units are centimeters. Therefore if we make 11.5 meters tall object in Blender and export that with no adjustments to scale during export and
	--			import we get a result that appears 1:1 with how they were created.
	--		We "could" set the 3ds Max system unit to centimeters, but this has the potential to screw up the work flow of legacy creators.
	--			3ds Max will prompt you when a file is opened with different system units and even offer to scale the file, but from testing, this has negative results, especially when dealing with the avatar skeleton.
	--			We also don't want to make switching system units anything that happens frequently or even rarely. It is bad business.
	--		We don't want to have reference to two different scale systems. For consistency we want one reference scale.
	--			When working in Blender build your scene in meters to an avatar that is 11.5 meters tall
	--			When working in 3ds Max build your scene in meters to an avatar that is 11.5 meters tall, not 1150 meters tall.
	--	Therefore the best solution is to move away from building assets at 100x and toward consistency:
	--		Keep the IMVU standard for 3ds Max system units 1 unit == 1 meter for compatibility.
	--		TODO: Check for units.SystemType != #meters and units.SystemScale != 1.0 warn the customer.
	--		Build all assets and provide starter content 1:1 not 100:1. That is build to an avatar that is 11.5 meters tall.
	--			These assets in 3ds Max will be 1/100th the size of legacy assets, but the likelihood of creators merging legacy assets with new is hopefully not common.
	--		Set the EXPORT_SCALE_FACTOR (fbxExporterSetParam "ScaleFactor") to 100.0
	--		TODO (maybe): Provide a "Legacy" option that:
	--			Creates starter content at 100x
	--			Sets the EXPORT_SCALE_FACTOR (fbxExporterSetParam "ScaleFactor") to 1.0
	local EXPORT_SCALE_FACTOR = 100.0

	/*	CRUFT - Moved to imvuToolLib.ms
	-- Required system unit settings for Toolkit to work correctly. Toolkit relies on these for consistency from IMVUMegaScene to import into Studio
	-- These units also make moving assets back and forth between Blender and Maya consistent without issue with scale.
	local SYSTEM_UNIT_TYPE = #meters
	local SYSTEM_UNIT_SCALE = 1.0
	*/

	-- Uniform Toolkit node type to be used for all helper nodes (e.g.: Seat, Furniture, Catcher, Pitcher, etc.)
	local TOOLKIT_NODE_CLASS = Point
		
	-- Node type constants
	local nodeType_SEAT = 0
	local nodeType_FLOOR = 1
	local nodeType_WALL = 2
	local nodeType_CEILING = 3
	local nodeType_HANDLE = 4
	local nodeType_CATCHER_PITCHER = 5
	local nodeType_SEETHROUGH = 6

	local NODE_NAME_SEPARATOR = "."
	local LEGACY_NODE_NAME_SEPARATORS = "._"

	-- ROOM node names
	local NODE_NAME_ROOM_ROOT = "Skeleton" + NODE_NAME_SEPARATOR + "Room"
	local NODE_NAME_SEAT_PARENT = "AvatarNodes"
	local NODE_NAME_FURNITURE_PARENT = "FurnitureNodes"
	local NODE_NAME_FURNITURE = "Furniture"
	local NODE_NAME_FLOOR = "Floor"
	local NODE_NAME_WALL = "Wall"
	local NODE_NAME_CEILING = "Ceiling"
	local ROOM_REF_MESH_NAME = "Room_Ref"
	local CAMERA_NAME = "Camera" + NODE_NAME_SEPARATOR + "01" + NODE_NAME_SEPARATOR + "01"
	local CAMERA_TRAGET_NAME = "Camera" + NODE_NAME_SEPARATOR + "01" + NODE_NAME_SEPARATOR + "01" + NODE_NAME_SEPARATOR + "Target"
	local OMNI_LIGHT_NAME = "Omni01"

	-- FURNITURE node names
	local NODE_NAME_FURNITURE_ROOT = "Attachment" + NODE_NAME_SEPARATOR + "Root"
	local NODE_NAME_FURNITURE_NODES_PARENT = "Nodes"

	-- Shared ROOM and FURNITURE node names
	local NODE_NAME_SEAT = "Seat"
	local NODE_NAME_SEAT_STANDING = "Standing"
	local NODE_NAME_SEAT_SITTING = "Sitting"
	local NODE_NAME_SEAT_IDLE = "Idle"
	local NODE_NAME_HANDLE = "Handle"
	local NODE_NAME_CATCHER = "Catcher"
	local NODE_NAME_PITCHER = "Pitcher"
	local NODE_NAME_LIGHTING = "LightingNodes"
	local NODE_NAME_SEETHROUGH = "SeeThrough"
	local FURNITURE_REF_MESH_NAME = "ScaleCube_Ref"
	local MODIFIER_NAME_SEETHROUGHSKIN = "SeeThroughSkin"

	-- Shared values
	local WORLD_ORIGIN_POSITION = [0.0,0.0,0.0]
	local DEFAULT_NODE_SIZE = 2.0
	local NODE_SIZE_ZERO = 0.0
	local COLOR_WHITE = color 255 255 255
	local COLOR_BLACK = color 0 0 0		

	-- ROOM values in meters unless otherwise noted
	local DEFAULT_ROOM_REF_SIZE = 40.0
	local MAX_ROOM_REF_SIZE = 10000.0
	--local DEFAULT_ROOM_SCALE_FACTOR = 100.0
	local ROOM_REF_SCALE_FACTOR = 1.0
	local ROOM_REF_BOX_SIZE = [1.0, 1.0, 1.0]
	local CAMERA_POS = [10.0, -28.0, 5.0]
	local CAMERA_TARGET_POS = [10.0, 10.0, 6.0]
	local ROOM_COLOR = color 204 204 204
	local OMNI_LIGHT_POSITION = [5.0, 0.0, 25.0]
	local SCALE_REFERENCE_SIZE = [3.13, 3.13, 3.13]
	local SCALE_REFERENCE_POS = [-15.0, 0.0, 0.0]
	local NEW_SEAT_OFFSET = [3.0, 0.0, 0.0]
	local SEE_THROUGH_NODE_SIZE = [1.0, -1.0, 2.0]
		
	-- FURNITURE values
	local FURNITURE_ANIM_DEFAULT_NAME = "FurnAnim01"
	local FURNITURE_REFERENCE_CUBE_SIZE = [6.0, 3.13, 3.13]

	-- Shared ROOM and FURNITURE values in meters unless otherwise noted
	local SKELETON_ROOT_NODE_SIZE = 4.0
	local CATCHER_NODE_OFFSET = [0.0, -3.0, 0.0]
	local PITCHER_NODE_OFFSET = [0.0, -8.0, 0.0]
	local PITCHER_CATCHER_NODE_SIZE = 1.5
	local HANDLE_NODE_OFFSET = [0.0, 0.0, 5.0]
	local HANDLE_NODE_SIZE = 1.5

	-- ATTACHMENT node names
	local NODE_NAME_ATTACHMENT_ROOT = "AttachmentRoot"
	--local NODE_NAME_ATTACHMENT_NODE = "AttachmentNode"  -- TODO: I don't think this is a requirement and it might be more convenient to have any/all assets attached to the AttachmnetRoot node

	-- CLOTHING node names
	local MODIFIER_NAME_CLOTHINGSKIN = "ClothingSkin"
	
	-- CLOTHING values
	local BONES_PER_VERTEX_LIMIT = 6

	-- Shared ATTACHNMENT and CLOTHING avatar visibility constants
	local visAvatar_NONE = 0
	local visAvatar_PROXY = 1
	local visAvatar_FEMALE = 2
	local visAvatar_MALE = 3
	-- Order in which the avatar labels appear in the Visuals Tool rollout
	local VIS_AVATAR_LABELS = #("Male","Female", "Proxy", "Hide All")
	local VIS_AVATAR_LABEL_ORDER = #(visAvatar_MALE, visAvatar_FEMALE, visAvatar_PROXY, visAvatar_NONE)

	-- ANIMATION defaults and constants
	local AVATAR_ANIM_DEFAULT_NAME = "AviAnim01"
	local ANIM_AUTO_DEFAULT_OFF = false
	local ANIM_FRAMERATE = 30
	local ANIM_DEFUALT_LENGTH = 100
	local ANIM_POSITION_THRESHOLD = 0.1
	local ANIM_ROTATION_THRESHOLD = 0.1
	local ANIM_IK_ARMS_CAPTION = "IK Arms"
	local ANIM_IK_LEGS_CAPTION = "IK Legs"
	local animAlign_IK2FKARMS_L = 1
	local animAlign_IK2FKARMS_R = 2
	local animAlign_IK2FKLEGS_L = 3
	local animAlign_IK2FKLEGS_R = 4
	local animAlign_FK2IKARMS_L = 5
	local animAlign_FK2IKARMS_R = 6
	local animAlign_FK2IKLEGS_L = 7
	local animAlign_FK2IKLEGS_R = 8
	local ANIM_ALIGN_PAIR_NODE_NAMES = #(
		#(	-- IK->FK Arms - left
			#("WGT_IK_lfArm_Hand","WGT_FK_lfArm_Hand")),
		#(	-- right
			#("WGT_IK_rtArm_Hand","WGT_FK_rtArm_Hand")),
		#(	-- IK->FK Legs - left
			#("WGT_IK_lfLeg_Foot","WGT_FK_lfLeg_Foot")),
		#(	-- right
			#("WGT_IK_rtLeg_Foot","WGT_FK_rtLeg_Foot")),
		#(	-- FK->IK Arms - left
			#("WGT_FK_lfArm_Bicep","MCH_IK_lfBicep"),
			#("WGT_FK_lfArm_Elbow","MCH_IK_lfElbow"),
			#("WGT_FK_lfArm_Hand","WGT_IK_lfArm_Hand")),
		#(	-- right
			#("WGT_FK_rtArm_Bicep","MCH_IK_rtBicep"),
			#("WGT_FK_rtArm_Elbow","MCH_IK_rtElbow"),
			#("WGT_FK_rtArm_Hand","WGT_IK_rtArm_Hand")),
		#(	-- FK->IK Legs - left
			#("WGT_FK_lfLeg_Thigh","MCH_IK_lfThigh"),
			#("WGT_FK_lfLeg_Calf","MCH_IK_lfCalf"),
			#("WGT_FK_lfLeg_Foot","MCH_IK_lfFoot"),
			#("WGT_FK_lfLeg_Toes","MCH_IK_lfToes")),
		#(	-- right
			#("WGT_FK_rtLeg_Thigh","MCH_IK_rtThigh"),
			#("WGT_FK_rtLeg_Calf","MCH_IK_rtCalf"),
			#("WGT_FK_rtLeg_Foot","MCH_IK_rtFoot"),
			#("WGT_FK_rtLeg_Toes","MCH_IK_rtToes")))

	-- Tool states
	local toolState_DISABLED = 0
	local toolState_ENABLED = 1
	local toolState_ACTIVE = 2

	-- Tooltip strings
	local TOOLTIP_VISUALS_SHOWNAMES_BUTTON = "Show Object Names"
	local TOOLTIP_VISUALS_HIDENAMES_BUTTON = "Hide Object Names"
	local TOOLTIP_VISUALS_FREEZEAVATARS_BUTTON = "Freeze the Male and Female avatar meshes"
	local TOOLTIP_VISUALS_UNFREEZEAVATARS_BUTTON = "Unfreeze the Male and Female avatar meshes"

	local TOOLTIP_ROOM_TOOL_CREATE_BUTTON = "Create a room based on the above dimensions"
	local TOOLTIP_ROOM_TOOL_ADDFURNITURE_BUTTON = "Create a piece of furniture based on the above selection"
	local TOOLTIP_ROOM_TOOL_ROOMN_BUTTON = "Duplicate furniture up"
	local TOOLTIP_ROOM_TOOL_ROOMW_BUTTON = "Duplicate furniture left"
	local TOOLTIP_ROOM_TOOL_ROOME_BUTTON = "Duplicate furniture right"
	local TOOLTIP_ROOM_TOOL_ROOMS_BUTTON = "Duplicate furniture down"
	local TOOLTIP_ROOM_TOOL_FURNDIST_BUTTON = "Offset distance when duplicating furniture nodes"
	local TOOLTIP_ROOM_TOOL_ADDIDLE_BUTTON = "Create an idle seat node"
	local TOOLTIP_ROOM_TOOL_ADDSTANDING_BUTTON = "Create a standing seat node"
	local TOOLTIP_ROOM_TOOL_ADDSITTING_BUTTON = "Create a sitting seat node"
	local TOOLTIP_ROOM_TOOL_CUSTOMSEAT = "Name the custom seat node"
	local TOOLTIP_ROOM_TOOL_ADDCUSTOMSEAT_BUTTON = "Create a custom seat node"
	local TOOLTIP_ROOM_TOOL_ADDCATCHER_BUTTON = "Add a catcher node"
	local TOOLTIP_ROOM_TOOL_ADDHANDLE_BUTTON = "Remove a handle node"
	local TOOLTIP_ROOM_TOOL_REMOVECATCHER_BUTTON = "Add a catcher node"
	local TOOLTIP_ROOM_TOOL_REMOVEHANDLE_BUTTON = "Remove a handle node"
	--local TOOLTIP_ROOM_TOOL_CREATESEETHRUNODE_BUTTON = "Create See Through Node"
	--local TOOLTIP_ROOM_TOOL_LINKSEETHRUNODE_BUTTON = "Link To See Through Node"
	--local TOOLTIP_ROOM_TOOL_UNLINKSEETHRUNODE_BUTTON = "Unlink From See Through Node"

	local TOOLTIP_FURNITURE_TOOL_CREATE_BUTTON = "Create a furniture"
	local TOOLTIP_FURNITURE_TOOL_ADDIDLE_BUTTON = TOOLTIP_ROOM_TOOL_ADDIDLE_BUTTON
	local TOOLTIP_FURNITURE_TOOL_ADDSTANDING_BUTTON = TOOLTIP_ROOM_TOOL_ADDSTANDING_BUTTON
	local TOOLTIP_FURNITURE_TOOL_ADDSITTING_BUTTON = TOOLTIP_ROOM_TOOL_ADDSITTING_BUTTON
	local TOOLTIP_FURNITURE_TOOL_CUSTOMSEAT = TOOLTIP_ROOM_TOOL_CUSTOMSEAT
	local TOOLTIP_FURNITURE_TOOL_ADDCUSTOMSEAT_BUTTON = TOOLTIP_ROOM_TOOL_CUSTOMSEAT
	local TOOLTIP_FURNITURE_TOOL_ADDCATCHER_BUTTON = TOOLTIP_ROOM_TOOL_ADDCATCHER_BUTTON
	local TOOLTIP_FURNITURE_TOOL_ADDHANDLE_BUTTON = TOOLTIP_ROOM_TOOL_ADDHANDLE_BUTTON
	local TOOLTIP_FURNITURE_TOOL_REMOVECATCHER_BUTTON = TOOLTIP_ROOM_TOOL_REMOVECATCHER_BUTTON
	local TOOLTIP_FURNITURE_TOOL_REMOVEHANDLE_BUTTON = TOOLTIP_ROOM_TOOL_REMOVEHANDLE_BUTTON
	local TOOLTIP_FURNITURE_TOOL_ANIMNAME = "Name your furniture animation"
	local TOOLTIP_FURNITURE_TOOL_ANIMNAMEFROMOBJECT = "Set the name your furniture animation from the selected object. If more than one object is selected picks the first one."
	local TOOLTIP_FURNITURE_TOOL_SETSELECTEDFOREXPORT = "Adds the selected object(s) and their children to the furniture export layer and parents selected object(s) to the Attachment.Root node. Preserves the hierarchy of the selection."

	local TOOLTIP_ATTACHMENT_TOOL_CREATE_BUTTON = "Create a new Attachment product"
	local TOOLTIP_ATTACHMENT_TOOL_SETUPORIENTATION_CHECKBOX = "When checked the selected objects will be parented to the Attachment Root bone before the Attachment Root bone is aligned with the picked bone"
	local TOOLTIP_ATTACHMENT_TOOL_SETUP_BUTTON = "Aligns Attachment Root bone to the picked bone and sets up the selected object(s) for export"
	local TOOLTIP_ATTACHMENT_TOOL_RESET_BUTTON = "Returns attachment objects and Attachment Root to last position and rotation before setup and unparents attachment objects from the Attachment Root node"

	local TOOLTIP_CLOTHING_TOOL_CREATE_BUTTON = "Create a new Clothing product"
	local TOOLTIP_CLOTHING_TOOL_BIND_BUTTON = "Weight the selected object to the avatar skeleton"
	local TOOLTIP_CLOTHING_TOOL_PAINTWEIGHTS_BUTTON = "Adjust the bone weighting"
	local TOOLTIP_CLOTHING_TOOL_POSE_BUTTON = "Pose the avatar"
	local TOOLTIP_CLOTHING_TOOL_LIMITBONES_BUTTON = "Limit the bones"

	local TOOLTIP_AVATARANIM_TOOL_CREATE_BUTTON = "Create a new Avatar Animation product"
	local TOOLTIP_AVATARANIM_TOOL_MAKEPOSE = "On export automatically set the needed keys on frames 0, 1, and 3 for creating a static pose in IMVU Studio"
	local TOOLTIP_AVATARANIM_TOOL_EXPORTALLBONES = "On export key all avatar bones so that the animation replaces any idle animation with the action that is triggered. Useful for custom idles or dance actions."
	local TOOLTIP_AVATARANIM_TOOL_ANIMNAME = "Name your avatar animation"
	local TOOLTIP_AVATARANIM_TOOL_USEIK = "Use FK if unchecked"
	local TOOLTIP_AVATARANIM_TOOL_SELECTALLRIGWIDGETS_BUTTON = "Selects all rig widgets whether hidden or not. Useful for editing keys for all widgets at once."
	local TOOLTIP_AVATARANIM_TOOL_DELETEAVATARKEYS_BUTTON = "Delete keys at the current time slider position for selected widgets or all widgets if none are selected"
	local TOOLTIP_AVATARANIM_TOOL_DELETEALLAVATARKEYS_BUTTON = "Delete all keys for selected widgets or all widgets if none are selected"
	local TOOLTIP_AVATARANIM_TOOL_RESETPOSE_BUTTON = "Resets the pose of the entire avatar if no bones are selected, otherwise just resets the selected bones"


	local TOOLTIP_EXPORT_TOOL_EXPORT_BUTTON = "Export your project to FBX"
	local TOOLTIP_EXPORT_TOOL_OPENDIR_BUTTON = "Open the export directory"
	local TOOLTIP_EXPORT_TOOL_HELP_BUTTON = "Visit the Help Page"
	local TOOLTIP_EXPORT_TOOL_VERSION_BUTTON = "This version of Studio Toolkit is up to date"
	local TOOLTIP_EXPORT_TOOL_VERSION_BUTTON_BETA = "This version of Studio Toolkit is in Beta"
	local TOOLTIP_EXPORT_TOOL_VERSION_BUTTON_UNK = "Failed to retrieve current version from server! Please check download page for current version."
	local TOOLTIP_EXPORT_TOOL_VERSION_BUTTON_UPDATE = "This version of Studio Toolkit is not the latest! Please check download page for current version."


	-- STRUCTS ----------------------------------------------------

	-- Toolkit tool state struct - A structure for managing the state of each individual tool in the toolkit
	struct ToolkitTool
	(
		public
		id,
		toolRollout, -- <rollout> The rollout for the tool
		compatibleTools = #(), -- <array> of tool struct instances that can be active along with this one.
		controlsEnabledWhenToolEnabled = #(), -- <array> of tool UI controls that get enabled when the tool state is set to ENABLED
		controlsDisabledWhenToolActive = #(), -- <array> of tool UI controls that get disabled when the tool state is set to ACTIVE -- Did it this way because this is typically a shorter list
		exportLayerName = "", -- <string> constant of the name of the export layer associated with the toolkit tool, EXPORT_LAYER_NAME_
		
		private
		state = toolState_DISABLED,
		function setEnabledControls toolControls enable =
		(
			if toolControls != undefined then
				for cntrl in toolControls do
					if cntrl != undefined  then cntrl.enabled = enable
		),
		
		public
		function setState newState =
		(
			state = newState
			--format "B : % : % : %\n" state toolRollout &toolRollout -- DEBUG
			if toolRollout != undefined then
			(
				-- Disable all tool controls to start with. This will be the default if the state is not getting set to ACTIVE or ENABLED
				setEnabledControls toolRollout.controls false
				case state of
				(
					-- If tool is being activated, enable all of the controls and then disable the necessary controls
					toolState_ACTIVE:
					(
						--format "C : %\n" controlsDisabledWhenToolActive -- DEBUG
						if not toolRollout.open then toolRollout.open = true
						setEnabledControls toolRollout.controls true
						setEnabledControls controlsDisabledWhenToolActive false
					)
					-- If tool is being enabled, enable only the necessary controls to activate the tool.
					toolState_ENABLED: setEnabledControls controlsEnabledWhenToolEnabled true
				)
			)
		),
		function getState = (state)
	)

	-- Toolkit tool variable structs for storing tool specific variable during toolkit session
	struct VisualsVars
	(
		public
		setVisibleAvatar,
		setFreezeAvatars
	)

	struct AttachmentObject
	(
		public
		obj,
		objParent,
		pos,
		rot
	)

	struct AttachmentVars
	(
		public
		reorientState,
		attachmentRoot,
		attachmentObjects = #()
	)

	struct AvatarAnimVars
	(
		public
		allWidgets,
		ikLeftArmWidgets,
		ikRightArmWidgets,
		ikLeftLegWidgets,
		ikRightLegWidgets,
		fkLeftArmWidgets,
		fkRightArmWidgets,
		fkLeftLegWidgets,
		fkRightLegWidgets,
		customAttributesNode,
		alignPairs,
		alignWidgets,
		setKinematicsMode
	)


	-- VARIABLES ----------------------------------------------------
	global imvuStudioToolkitFloater

	-- Define regular expressions object
	local rgx = dotNetClass "System.Text.RegularExpressions.RegEx"

	-- Declare variables for tool objects
	local visualsTool, roomTool, furnitureTool, attachmentTool, clothingTool, avatarAnimTool, exportTool

	-- Declare variables for tool variable structs
	local visualsToolVars = VisualsVars setVisibleAvatar:undefined setFreezeAvatars:undefined
	local attachmentToolVars = AttachmentVars attachmentRoot:undefined reorientState:undefined
	local avatarAnimToolVars = AvatarAnimVars \
		allWidgets:#() \
		ikLeftArmWidgets:undefined \
		ikRightArmWidgets:undefined \
		ikLeftLegWidgets:undefined \
		ikRightLegWidgets:undefined \
		fkLeftArmWidgets:undefined \
		fkRightArmWidgets:undefined \
		fkLeftLegWidgets:undefined \
		fkRightLegWidgets:undefined \
		customAttributesNode:undefined \
		alignPairs:#() \
		alignWidgets:undefined
	local allTools = #()

	-- Rollout variables
	local roomRollout, furnitureRollout, attachmentRollout, clothingRollout, avatarAnimationRollout, exportRollout

	-- Declare variables for Utility functions that are used by Update functions or multiple toolkit tools
	-- CFUFT - Moved to imvuToolLib.ms local printDebug = undefined
	local getAnimLength = undefined
	local disableAnimationModes = undefined
	-- CFUFT - Moved to imvuToolLib.ms local areSystemUnitsCorrect = undefined

	-- Variables used by multiple tools
	local allAvatarExportBones = #()

	-- Registered callbacks in the form of <name> e.g.: #filesave
	local toolkitCallbackIds = #()


	-- BEGIN - TOOL UPDATE FUNCTIONS - Functions to set/update the state of the UI --------------------------------------------------------------------------------------------------------------------

	-- Update Function
	-- Global function that returns the local version of the toolkit. Must be called like this '::getImvuStudioToolkitVerison()'
	--		It is a global function because it is called by the installer builder script (mzpbuilder.ms)
	-- TODO: When/if I refactor ImvuStudioToolkit to be a singleton(?) struct, this will turn into an accessor function
	-- Parameters: None
	-- Returns: <string> the version number
	global getImvuStudioToolkitVerison
	fn getImvuStudioToolkitVerison = (LOCAL_VERSION)


	-- Update Function
	-- Checks the version against what version is available for download
	-- Parameters: None
	-- Returns: None
	fn checkForUpdate =
	(
		local PRINT_DEBUG = false
		local servicePointManager, httpWebRequest, response, responseStream, stream, body, pattern, match, onlineVersion
		local url = "https://create.imvu.com/api/studio/3dsmax"
		--local url = "https://www.google.com/404" -- TEST DEBUG
		local isUpdateButtonEnabled = false
		local updateButtonText = LOCAL_VERSION
		local updateButtonTooltip = TOOLTIP_EXPORT_TOOL_VERSION_BUTTON

		servicePointManager = dotNetClass "System.Net.ServicePointManager"
		servicePointManager.Expect100Continue = true
		servicePointManager.SecurityProtocol = (dotNetClass "System.Net.SecurityProtocolType").Tls12


		httpWebRequest = (dotNetClass "System.Net.HttpWebRequest").Create(url)
		httpWebRequest.Method = "GET"
		httpWebRequest.Timeout = 1000
		httpWebRequest.UserAgent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36" -- NOTE: This parameter appears not needed. If omitted we still get the response from the endpoint.
		httpWebRequest.proxy = (dotNetClass "System.Net.WebProxy").GetDefaultProxy()

		try -- Safeguard for 404 error from server
		(
			response = httpWebRequest.GetResponse()
			responseStream = response.GetResponseStream()
			stream = dotNetObject "System.IO.StreamReader" responseStream
			body = stream.ReadToEnd()
			pattern = "[^version\"]([0-9.]+)"
			match = rgx.Match body pattern

			if body != undefined then
				onlineVersion = match.value

			response.Close()
		)
		catch
		(
			imvuTools.printMsg #("IMVU Studio Toolkit failed to retrieve current version from server!") msgType:#ERROR

			isUpdateButtonEnabled = true
			updateButtonText = (LOCAL_VERSION + " (!)")
			updateButtonTooltip = TOOLTIP_EXPORT_TOOL_VERSION_BUTTON_UNK
		)

		if onlineVersion != undefined then
		(
			if onlineVersion > LOCAL_VERSION then
			(
				isUpdateButtonEnabled = true
				updateButtonText = "UPDATE AVAILABLE!"
				updateButtonTooltip = TOOLTIP_EXPORT_TOOL_VERSION_BUTTON_UPDATE
			)
			else if onlineVersion < LOCAL_VERSION then
			(
				updateButtonText = ("Beta " + LOCAL_VERSION)
				updateButtonTooltip = TOOLTIP_EXPORT_TOOL_VERSION_BUTTON_BETA
			)
		)
		if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #(("LocalVerison:" + LOCAL_VERSION), ("OnlineVersion:" + onlineVersion)) msgType:#INFO inScript:THIS_SCRIPT_FILENAME inFunction:checkForUpdate

		if exportRollout != undefined then
		(
			if exportRollout.versionButton != undefined then
			(
				exportRollout.versionButton.text = updateButtonText
				exportRollout.versionButton.toolTip = updateButtonTooltip
				exportRollout.versionButton.enabled = isUpdateButtonEnabled
			)
		)
	)
	
	
	-- Update Function
	-- Validates active tool compatibility.
	-- Parameters:
	--		toolkitToolInst <ToolkitTool>
	-- Returns: <boolean> 
	fn canToolBeActiveOrEnabled toolkitToolInst =
	(
		local toolCanBeActiveOrEnabled = true

		-- Look through all of the tools and if the tool is not the one supplied and it is not in the compatibility list of the one supplied and it is active then we have a problem.
		for aTool in allTools do
			if aTool != toolkitToolInst and findItem toolkitToolInst.compatibleTools aTool == 0 and aTool.getState() == toolState_ACTIVE then
				toolCanBeActiveOrEnabled = false
		
		--format "F %\n"  toolCanBeActiveOrEnabled -- DEBUG
		toolCanBeActiveOrEnabled
	)


	-- Update Function
	-- Check to see if the current scene has the custom attributes on the scene 'rootNode' used to store values we want to save withe the scene file
	--	NOTE: 'rootNode' is Max default node that is guaranteed to be in every scene file.
	-- Parameters: None
	-- Returns: <boolean> True if attributes exist, false if not
	fn sceneHasToolkitCustomAttributes =
	(
		local PRINT_DEBUG = false
		local attribsExist = false
		local custAttribsDefs = custAttributes.getDefs rootNode

		if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #(custAttribsDefs) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:sceneHasToolkitCustomAttributes
		if custAttribsDefs != undefined then
			for ii = 1 to custAttribsDefs.count while not attribsExist do attribsExist = custAttribsDefs[ii].name == #toolkitData
		attribsExist
	)


	-- Update Function
	-- Add toolkit custom attributes on the scene 'rootNode' used to store values we want to save withe the scene file
	--	NOTE: 'rootNode' is Max default node that is guaranteed to be in every scene file.
	-- More info on Max Custom Attributes: https://help.autodesk.com/view/3DSMAX/2017/ENU/?guid=__files_GUID_ADFD29E4_9751_4F55_98CC_F7C721C9AEE2_htm
	--	And a great video on using custom attributes: https://www.youtube.com/watch?v=GGfxeymS85w
	-- Parameters: None
	-- Returns: None
	fn addToolkitCustomAttributesToScene =
	(
		local PRINT_DEBUG = false
		local custAttribsNode = custAttributes.getDefs rootNode
		local attribDef
		
		attribDef = attributes toolkitData attribID:#(0x115e615e, 0x7b021399)
		(
			parameters main
			(
				avatarAnimName type:#string default:""
				furnitureAnimName type:#string default:""
			)
		)
		
		if not sceneHasToolkitCustomAttributes() then
		(
			if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("Adding toolkit custom attributes to scene!") msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:addToolkitCustomAttributesToScene
			custAttributes.add rootNode attribDef
		)
	)
	

	-- Update Function
	-- Updates all toolkit tool variables.
	-- Parameters: None
	-- Returns: None
	fn setToolkitToolVariables =
	(
		local source, target, nodePairs

		for toolkitToolInst in allTools do
		(
			case toolkitToolInst.id of
			(
				toolId_VISUALS: ()
				toolId_ROOM: ()
				toolId_FURNITURE: ()
				toolId_ATTACHMENT:
				(
					allAvatarExportBones = execute ("$" + MEGASCENE_NODE_NAME_AVATAR_SKELETON_ROOT + "...*") as array
				)
				toolId_CLOTHING:
				(
					allAvatarExportBones = execute ("$" + MEGASCENE_NODE_NAME_AVATAR_SKELETON_ROOT + "...*") as array
				)
				toolId_AVATARANIM:
				(
					allAvatarExportBones = execute ("$" + MEGASCENE_NODE_NAME_AVATAR_SKELETON_ROOT + "...*") as array

					avatarAnimToolVars.allWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_WIDGETS + "*") as array

					avatarAnimToolVars.ikLeftArmWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_IK_LEFT_ARM_WIDGETS + "*") as array
					avatarAnimToolVars.ikRightArmWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_IK_RIGHT_ARM_WIDGETS + "*") as array
					avatarAnimToolVars.ikLeftLegWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_IK_LEFT_LEG_WIDGETS + "*") as array
					avatarAnimToolVars.ikRightLegWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_IK_RIGHT_LEG_WIDGETS + "*") as array

					avatarAnimToolVars.fkLeftArmWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_FK_LEFT_ARM_WIDGETS + "*") as array
					avatarAnimToolVars.fkRightArmWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_FK_RIGHT_ARM_WIDGETS + "*") as array
					avatarAnimToolVars.fkLeftLegWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_FK_LEFT_LEG_WIDGETS + "*") as array
					avatarAnimToolVars.fkRightLegWidgets = execute ("$" + MEGASCENE_NODE_NAME_PREFIX_RIG_FK_RIGHT_LEG_WIDGETS + "*") as array

					avatarAnimToolVars.customAttributesNode = getNodeByName MEGASCENE_NODE_NAME_RIG_ATTRIBUTES

					-- Rebuild the list of rig widgets pairs
					avatarAnimToolVars.alignPairs = #()
					for pairIndex = animAlign_IK2FKARMS_L to animAlign_FK2IKLEGS_R do
					(
						nodePairs = #()
						for pairNames in ANIM_ALIGN_PAIR_NODE_NAMES[pairIndex] do
						(
							source = getNodeByName pairNames[const_SOURCE]
							target = getNodeByName pairNames[const_TARGET]

							append nodePairs #(source,target)
						)
						append avatarAnimToolVars.alignPairs nodePairs
					)
				)
			)
		)
	)


	-- Update Function
	-- Validates active tool compatibility, disables tools which are incompatible, enable which are compatible
	-- Parameters:
	--		updateAfter [<int>] Optional integer to control when UI updates and subsequent scene updates occur
			--	updateAfter_FILESAVE = 0
			--	updateAfter_FILEOPEN = 1
			--	updateAfter_FILENEW = 2
			--	updateAfter_TOOLSTART = 3
	fn updateToolkitSceneAndUIState updateAfter:updateAfter_FILESAVE =
	(
		local okToEnableExportButtons = false

		-- Add toolkitData custom attributes if they don't already exist
		addToolkitCustomAttributesToScene()

		-- Things we always want to do regardless of tool state
		for toolkitToolInst in allTools do
		(
			case toolkitToolInst.id of
			(
				toolId_FURNITURE:
				(
					if updateAfter == updateAfter_FILESAVE then rootNode.toolkitData.furnitureAnimName = toolkitToolInst.toolRollout.animationName.text
					else
					(
						-- updateAfter_FILEOPEN, updateAfter_TOOLSTART
						toolkitToolInst.toolRollout.animationName.text = rootNode.toolkitData.furnitureAnimName
						if updateAfter == updateAfter_FILENEW then rootNode.toolkitData.furnitureAnimName = toolkitToolInst.toolRollout.animationName.text = FURNITURE_ANIM_DEFAULT_NAME
					)
				)
				toolId_AVATARANIM:
				(
					if updateAfter == updateAfter_FILESAVE then rootNode.toolkitData.avatarAnimName = toolkitToolInst.toolRollout.animationName.text
					else
					(
						-- updateAfter_FILEOPEN, updateAfter_TOOLSTART
						toolkitToolInst.toolRollout.animationName.text = rootNode.toolkitData.avatarAnimName
						if updateAfter == updateAfter_FILENEW then rootNode.toolkitData.avatarAnimName = toolkitToolInst.toolRollout.animationName.text = AVATAR_ANIM_DEFAULT_NAME
					)
				)
			)
		)

		-- Rollup all tool rollouts after anything but a file save
		if not updateAfter == updateAfter_FILESAVE then for toolkitToolInst in allTools do toolkitToolInst.toolRollout.open = false

		for toolkitToolInst in allTools do toolkitToolInst.setState toolState_ENABLED

		visualsTool.setState toolState_ACTIVE -- Visuals Tool is always active
		
		if LayerManager.getLayerFromName EXPORT_LAYER_NAME_ROOM != undefined then
		(
			roomTool.setState toolState_ACTIVE
			okToEnableExportButtons = true
		)

		if LayerManager.getLayerFromName EXPORT_LAYER_NAME_FURNITURE != undefined then
		(
			furnitureTool.setState toolState_ACTIVE
			okToEnableExportButtons = true
		)

		if LayerManager.getLayerFromName EXPORT_LAYER_NAME_ATTACHMENT != undefined then
		(
			attachmentTool.setState toolState_ACTIVE
			okToEnableExportButtons = true
		)

		if LayerManager.getLayerFromName EXPORT_LAYER_NAME_CLOTHING != undefined then
		(
			clothingTool.setState toolState_ACTIVE
			okToEnableExportButtons = true
		)
		
		if LayerManager.getLayerFromName EXPORT_LAYER_NAME_AVATAR_ANIMATION != undefined then
		(
			avatarAnimTool.setState toolState_ACTIVE
			okToEnableExportButtons = true
		)
	

		for toolkitToolInst in allTools do
		(
			--toolkitToolInst.setState toolState_ENABLED

			if toolkitToolInst.getState() == toolState_ACTIVE then
			(
				--format "E-A Rollout: %\n" toolkitToolInst.toolRollout.name -- DEBUG
				-- Check if tool was activated when it should not have been
				if not canToolBeActiveOrEnabled toolkitToolInst then -- We have a problem					
					messageBox ("WARNING! Tool " + toolkitToolInst.toolRollout.name + " should not be active but is!") title:TOOLKIT_NAME icon:#warning
				else -- Update specific tool UI states
				(
					case toolkitToolInst.id of
					(
						toolId_VISUALS:
						(
							if attachmentTool.getState() == toolState_ACTIVE or clothingTool.getState() == toolState_ACTIVE or avatarAnimTool.getState() == toolState_ACTIVE then
								toolkitToolInst.toolRollout.displayAvatar.enabled = true
						)
						toolId_ROOM:
						(
							-- If ROOM_REF_MESH_NAME exists set the Room Size XSlider, YSlider, and ZSlider values to the bounding volume of the ROOM_REF_MESH_NAME
							-- TODO: Fix how the size of the room is set. Currently it modifies the scale, but this is problematic, because if the users modifies the verts then scale is no longer relevant.
							local roomRef = getNodeByName ROOM_REF_MESH_NAME
							local roomRefLocalBoundingBox
							local roomRefBoundingSize
							
							if roomRef != undefined then
							(
								roomRefLocalBoundingBox = nodeLocalBoundingBox roomRef
								roomRefLocalBoundingBox = (roomRefLocalBoundingBox[1] - roomRefLocalBoundingBox[2])
								roomRefLocalBoundingBox = [abs roomRefLocalBoundingBox.x, abs roomRefLocalBoundingBox.y, abs roomRefLocalBoundingBox.z]
								toolkitToolInst.toolRollout.XSlider.value = roomRefLocalBoundingBox.x
								toolkitToolInst.toolRollout.YSlider.value = roomRefLocalBoundingBox.y
								toolkitToolInst.toolRollout.ZSlider.value = roomRefLocalBoundingBox.z
							)
						)
						toolId_FURNITURE: ()
						toolId_ATTACHMENT:
						(
							unhide allAvatarExportBones
						)
						toolId_CLOTHING: ()
						toolId_AVATARANIM:
						(
							unhide avatarAnimToolVars.allWidgets

							-- Update after a New file load or Open file. Ignore file saves
							if updateAfter != updateAfter_FILESAVE then
							(
								local animLength = getAnimLength()
								-- Safeguard against no keys or keys only on frame 0
								if animLength == 0 then animLength = ANIM_DEFUALT_LENGTH

								-- Set the animation frame range
								animationRange = interval 0 animLength
								-- Set the UI to match the frame range
								toolkitToolInst.toolRollout.animationLength.value = animLength

								-- Turn off all animation modes that may be active
								disableAnimationModes()

								avatarAnimToolVars.setKinematicsMode ANIM_IK_LEGS_CAPTION true
								avatarAnimToolVars.setKinematicsMode ANIM_IK_ARMS_CAPTION true
								toolkitToolInst.toolRollout.armsIK.checked = true
								toolkitToolInst.toolRollout.legsIK.checked = true

								visualsToolVars.setVisibleAvatar visAvatar_PROXY setUI:true
							)
						)
					)
				)
			)
			else if toolkitToolInst.getState() == toolState_ENABLED then
			(
				--format "E-E Rollout: %\n" toolkitToolInst.toolRollout.name -- DEBUG
				if not canToolBeActiveOrEnabled toolkitToolInst then
					toolkitToolInst.setState toolState_DISABLED
				else
				(
					if toolkitToolInst.id == toolId_EXPORT then
					(
						if not maxFileName == "" and okToEnableExportButtons then
						(
							toolkitToolInst.toolRollout.exportButton.enabled = true
							toolkitToolInst.toolRollout.openDirButton.enabled = true
							toolkitToolInst.toolRollout.open = true
						)
					)
				)
			)
		)
	)

	
	-- Global function so it can be accessed when 3ds Max application callback events occur.
	-- Called after file save event
	-- Parameters:
	--		callbackParams <array> a 2 element array: Element 1 is 1 if doing a normal file save, 2 if doing an Edit>Hold 3 if doing an Autobackup More info: https://help.autodesk.com/view/MAXDEV/2023/ENU/?guid=GUID-2D8606DF-EF5D-4F5B-A00C-B10D29A14947
	-- Returns: None
	global updateToolkitPostFileSave
	fn updateToolkitPostFileSave callbackParams:#(CALLBACKPARAM_NORMALSAVE,"") =
	(
		--format "ImvuStudioToolkit::updateToolkitPostFileSave() callbackParams:[%,%]\n" callbackParams[1] callbackParams[2]-- DEBUG

		if callbackParams[1] == CALLBACKPARAM_NORMALSAVE then
		(
			updateToolkitSceneAndUIState updateAfter:updateAfter_FILESAVE
		)
	)

	-- Global function so it can be accessed when 3ds Max application callback events occur.
	-- Called after file open event
	-- Parameters:
	--		callbackParams <array> a 2 element array: Element 1 is 1 if doing a normal file load, 2 if doing an Edit>Fetch More info: https://help.autodesk.com/view/MAXDEV/2023/ENU/?guid=GUID-2D8606DF-EF5D-4F5B-A00C-B10D29A14947
	-- Returns: None
	global updateToolkitPostFileOpen
	fn updateToolkitPostFileOpen callbackParams:#(CALLBACKPARAM_NORMALSAVE,"") =
	(
		setToolkitToolVariables()
		updateToolkitSceneAndUIState updateAfter:updateAfter_FILEOPEN
		-- Make sure that the user did not allow the file being opened to alter the system units.
		--	This happens if a mismatch of the system units is detected.
		--	An option pops up when opening a file with system unit settings that don't match the current system unit settings.
		--	If the user selects "Adopt the File's Unit Scale?" things will go badly :(
		imvuTools.areSystemUnitsCorrect()
	)


	-- Global function so it can be accessed when 3ds Max application callback events occur.
	-- Called after file new event
	-- Parameters: None
	-- Returns: None
	global updateToolkitPostFileNew
	fn updateToolkitPostFileNew =
	(
		setToolkitToolVariables()
		updateToolkitSceneAndUIState updateAfter:updateAfter_FILENEW
	)

	
	-- Update Utility Function
	-- Clear all registered callbacks. To be used on close of tool
	-- Parameters: None
	-- Returns: None
	fn clearAllCallbacks =
	(
		for callbackId in toolkitCallbackIds do callbacks.removeScripts id:callbackId

		deleteAllChangeHandlers id:#ChangeHandler_AvatarAnimWidgetArms
		deleteAllChangeHandlers id:#ChangeHandler_AvatarAnimWidgetLegs
	)


	-- Update Utility Function
	-- Set a callback
	-- Parameters:
	--		callbackType <#name> https://help.autodesk.com/view/MAXDEV/2023/ENU/?guid=GUID-C1F6495F-5831-4FC8-A00C-667C5F2EAE36
	--		functionToCall <string> function to be called
	--		callbackId <#name> to identify the callback
	-- Returns: None
	fn setCallback callbackType functionToCall callbackId =
	(
		callbacks.removeScripts id:callbackId
		callbacks.addScript callbackType functionToCall id:callbackId
		appendIfUnique toolkitCallbackIds callbackId
	)


	-- Update Function
	-- Set up tool callbacks
	-- Parameters: None
	-- Returns: None 
	fn setToolkitCallbacks =
	(
		-- Set up callbacks
		setCallback #filePostSaveProcess "updateToolkitPostFileSave callbackParams:(callbacks.notificationParam())" #toolkitFileSave
		setCallback #filePostOpenProcess "updateToolkitPostFileOpen callbackParams:(callbacks.notificationParam())" #toolkitFileOpen
		setCallback #systemPostNew "updateToolkitPostFileNew()" #toolkitFileNew
	)

	
	-- BEGIN - UTILITY FUNCTIONS - Functions called by other functions --------------------------------------------------------------------------------------------------------------------------------
	
	/* CFUFT - Moved to imvuToolLib.ms
	-- Utility Function
	-- Prints a consistently formatted debug string to the 3ds Max MAXScript Listener output window
	-- Parameters:
	--		msgArray <array> Things to print
	--		addSpaces <boolean> [optional] Defaults to true and Will automatically add spaces between things to print
	--		func <function> [optional] The function from which the print debug is being called
	-- Returns: None
	fn printDebug msgArray addSpaces:true func:undefined =
	(
		local msg
		local msgStr = ""
		for msg in msgArray do msgStr += (msg as string)
		if addSpaces then msgStr += " "
		if func != undefined then msgStr = (func as string) + ": " + msgStr
		format "%\n" msgStr
	)
	*/

	
	-- TODO: Implement this function
	-- Utility Function
	-- Check if object is a valid object to be used by IMVU
	-- Parameters:
	--		obj <node> 
	-- Returns: <boolean> true if is a valid IMVU object, false if not
	fn isValidImvuObject obj =
	(
	)


	/* CFUFT - Moved to imvuToolLib.ms
	-- Utility Function
	-- Check if 3ds Max System Units are set correctly and alerts user with a message box informing so.
	-- Parameters: None
	-- Returns: <boolean> True if units are set correctly, false if not
	fn areSystemUnitsCorrect =
	(
		local sysUnitsAreCorrect = (units.SystemType == SYSTEM_UNIT_TYPE and units.SystemScale == SYSTEM_UNIT_SCALE)
		local msg = "WARNING!  System unit settings are not compatible with IMVU Studio Toolkit!\n"

		if not sysUnitsAreCorrect then
		(
			if not units.SystemType == SYSTEM_UNIT_TYPE then msg += ("System Unit Type is set to " + units.SystemType + " and should be " + SYSTEM_UNIT_TYPE + "\n")
			if not units.SystemScale == SYSTEM_UNIT_SCALE then msg += ("System Unit Scale is set to " + (units.SystemScale as string) + " and should be " + (SYSTEM_UNIT_SCALE as string))

			messageBox msg title:TOOLKIT_NAME icon:#warning
		)
		sysUnitsAreCorrect
	)
	*/
	

	/* CFUFT - Moved to imvuToolLib.ms
	-- Utility Function
	-- Filter a list of objects and return only the "root most nodes" 
	-- Parameters:
	--		objArray <array> Array of nodes
	-- Returns: <array> Filtered array of "root most" nodes
	fn getRootNodes objArray =
	(
		local rootNodes = #()
		local obj, objParent
		local parentFound = false

		for obj in objArray do
		(
			-- If this node has a parent and ANY parent up the hierarchy is in objArray then skip it
			parentFound = false
			objParent = obj.parent
			while objParent != undefined and not parentFound do
			(
				if findItem objArray ObjParent > 0 then parentFound = true
				objParent = objParent.parent
			)
			if not parentFound then appendIfUnique rootNodes obj
		)
		rootNodes
	)
	*/


	-- Utility Function
	-- Get transform matrix from supplied source
	-- Parameters:
	--		tm <variable> 
	-- Returns: <matrix3>
	fn toMatrix3 tm = case of
	(
		(iskindof tm Matrix3): tm
		(iskindof tm Point3): transmatrix tm
		(isvalidnode tm): tm.transform
		(tm == #world): WORLD_TRANSFORM_MATRIX
			  default: WORLD_TRANSFORM_MATRIX
	)


	-- Utility Function
	-- TODO: We may want to add the ability to align position or rotation only. Position only would be useful
	-- 		if aligning the leg IK widgets because these are the foot widgets
	-- 		and it is likely the user would like to use these for foot placement only.
	--		There are other widgets on the foot which control IK rotation.
	-- Align array of nodes to a target node using optional transform matrix source
	-- Parameters:
	--		node <node> Node to be aligned to the target
	--		target <node> Target node to align each node in 'nodes' to.
	--		origin <variable> Point, Matrix, node, #WORLD source to be used as the origin transform space
	--		atFrames <array> Of integers or frames at which to do the alignment. Assumes that if you are doing the alignment at frames then you are setting animation keys
	-- Returns: None
	fn alignNodesToTarget node target origin:#world atFrames:#() =
	(
		local target_tm = toMatrix3 target
		local origin_tm = toMatrix3 origin

		if atFrames.count > 0 then
		(
			with animate on
			(
				for frame in atFrames do
				(
					at time frame
					(
						target_tm = toMatrix3 target
						origin_tm = toMatrix3 origin
						node.transform *= inverse origin_tm * target_tm
					)
				)
			)
		)
		else node.transform *= inverse origin_tm * target_tm
	)


	-- Utility Function
	-- Get the node type of an object from by looking at its name
	-- Parameters:
	--		obj <node> 
	-- Returns: nodeTypeFound <integer> a 'nodeType_' constant
	-- NOTE: we are using "?" in the search pattern because we are anticipating that the separator might be "." or "_" or something else
	-- TODO: This function might be more efficient if we passed the object name sting instead of the object
	fn getObjectNodeType obj =
	(
		local nodeTypeFound
		if matchPattern obj.name pattern:(NODE_NAME_SEAT + "*") ignoreCase:true then nodeTypeFound = nodeType_SEAT
		else if matchPattern obj.name pattern:(NODE_NAME_HANDLE + "*") ignoreCase:true then nodeTypeFound = nodeType_HANDLE
		else if matchPattern obj.name pattern:(NODE_NAME_CATCHER + "*") ignoreCase:true then nodeTypeFound = nodeType_CATCHER_PITCHER
		else if matchPattern obj.name pattern:(NODE_NAME_PITCHER + "*") ignoreCase:true then nodeTypeFound = nodeType_CATCHER_PITCHER
		else if matchPattern obj.name pattern:(NODE_NAME_FURNITURE + "?" + NODE_NAME_FLOOR + "*") ignoreCase:true then nodeTypeFound = nodeType_FLOOR
		else if matchPattern obj.name pattern:(NODE_NAME_FURNITURE + "?" + NODE_NAME_WALL + "*") ignoreCase:true then nodeTypeFound = nodeType_WALL
		else if matchPattern obj.name pattern:(NODE_NAME_FURNITURE + "?" + NODE_NAME_CEILING + "*") ignoreCase:true then nodeTypeFound = nodeType_CEILING
		else if matchPattern obj.name pattern:(NODE_NAME_SEETHROUGH + "*") ignoreCase:true then nodeTypeFound = nodeType_SEETHROUGH
		--print ("getObjectNodeType() obj:" + obj.name + " nodeTypeFound:" + nodeTypeFound as string)
		nodeTypeFound
	)


	-- Utility Function - Get all nodes by node type from a supplied array
	-- Parameters:
	--		nodeType <integer> is a 'nodeType_' constant
	--		nodeArray: <array> Array variable to be acted upon.
	--		&nodesFoundArray: <array> Array variable passed by reference to be filled with nodes found.
	-- Returns: None
	fn getNodesByNodeTypeInArray nodeType nodeArray &nodesFoundArray =
	(
		for obj in nodeArray do 
			if getObjectNodeType obj == nodeType then
				append nodesFoundArray obj
	)


	-- Utility Function - Recursively look for an object with the name specified by the nameStr and return an array of found nodes. nameStr may include wildcards
	-- Parameters:
	--		parentObj: <node> Node to act on
	--		nodeType <integer> is a 'nodeType_' constant
	--		&foundArray: <array> Array variable passed by reference and returned with nodes found.
	-- Returns: None
	fn findChildrenByNodeType parentObj nodeType &foundArray =
	(
		for child in parentObj.children do 
		(
			if getObjectNodeType child == nodeType then
				append foundArray child
			if child.children.count > 0 then
				findChildrenByNodeType child nodeType &foundArray
		)
	)


	-- Utility Function - Recursively look for an object with the name specified by the nameStr and return true if found and false if not found. nameStr may include wildcards
	-- Parameters:
	--		parentObj: <node> Node to act on
	--		nodeType <integer> is a 'nodeType_' constant
	-- Returns: nodeFound <node>
	fn hasChildOfNodeType parentObj nodeType =
	(
		local nodeFound = false
		for child in parentObj.children do 
		(
			nodeFound = getObjectNodeType child == nodeType
			if not nodeFound then
			(
				if child.children.count > 0 then
					nodeFound = hasChildOfNodeType child nodeType
			)
		)
		nodeFound
	)


	-- Utility Function
	fn gwDisplayObjectNames =
	(
		gw.setTransform (matrix3 1)
		for obj in objects where not obj.isHiddenInVpt do
		(
			gw.text obj.pos obj.name color:yellow
			gw.enlargeUpdateRect #whole
		)
	)

	
	-- Utility Function
	-- Get the root node of any object in a hierarchy
	-- Returns: None
	fn getRoot node =
	(
		if isvalidnode node do (while node.parent != undefined do node = node.parent; node)
	)
	
	
	-- Utility Function
	-- Get all of the nodes in export layer by layer name
	-- Parameters:
	--		layerName <string> The name of an export layer see EXPORT_LAYER_NAME_ constants at top of script
	-- Returns: nodeArray <array> Array of nodes
	fn getNodesInExportLayer layerName =
	(
		local nodeArray = #()
		local layer = LayerManager.getLayerFromName layerName
		if layer != undefined then
			layer.nodes &nodeArray
		nodeArray
	)
	
	
	-- Utility Function
	-- Add nodes to export layer
	-- Parameters:
	--		nodes <array> of nodes to be added
	--		exportLayerName <string>
	-- Returns: None
	fn addNodesToExportLayer nodes exportLayerName =
	(
		local exportLayer = LayerManager.getLayerFromName exportLayerName

		if exportLayer == undefined then
			exportLayer = LayerManager.newLayerFromName exportLayerName
				
		--format "addNodesToExportLayer() exportLayerName:% exportLayer:%\n" exportLayerName exportLayer
		
		for eachNode in nodes do exportLayer.addNode eachNode
	)
	
	
	-- Utility Function
	-- Get the seat name from a seat node
	-- Parameters:
	--		seatNode <node>
	-- Returns a string on the seat name e.g.: Sitting, Standing, Idle, custom. Basically it "should" always be the suffix of (seat## + NODE_NAME_SEPARATOR)
	fn getSeatName seatNode =
	(
		local seatName = ""
		local match = rgx.Match seatNode.name ("(?i)^" + NODE_NAME_SEAT + "[0-9]+[" + LEGACY_NODE_NAME_SEPARATORS +"](.+)")
		if match == undefined then
			imvuTools.printMsg #("Node name match failed!") msgType:#ERROR inScript:THIS_SCRIPT_FILENAME inFunction:getSeatName
		else if match.Groups.count > 1 then
			seatName = match.Groups.Item[1].Value
		if seatName == "" then
			imvuTools.printMsg #("Seat name not found! seatNode:", seatNode.name, " seatName:", seatName) msgType:#WARN addSpaces:false inScript:THIS_SCRIPT_FILENAME inFunction:getSeatName
		seatName
	)
	

	-- Utility Function - Recursively look for seat child nodes and return an array of found nodes.
	-- Parameters:
	--		seatNode: <node> Node to act on
	--		&seatChildren: <array> Array variable passed by reference and returned with nodes found.
	-- Returns: None
	fn getSeatChildren seatNode &seatChildren =
	(
		seatChildren = #()
		if getObjectNodeType seatNode == nodeType_SEAT then
		(
			for child in seatNode.children do 
			(
				if getObjectNodeType child == nodeType_HANDLE then
					append seatChildren child
				if getObjectNodeType child == nodeType_CATCHER_PITCHER then
				(
					append seatChildren child
					for subchild in child.children do
					(
						if getObjectNodeType subchild == nodeType_CATCHER_PITCHER then
							append seatChildren subchild
					)
				)
			)
		)
		seatChildren
	)


	-- Utility Function
	-- Get the number from a imvu node
	-- Parameters:
	--		imvuNode <node>
	-- Returns: nodeNum <integer>
	fn getNodeNumber imvuNode =
	(
		local nodeNum = 0
		local matchStr
		local match
		
		case getObjectNodeType imvuNode of
		(
			nodeType_SEAT: matchStr = ("(?i)^" + NODE_NAME_SEAT + ".*([0-9]+)")
			nodeType_FLOOR: matchStr = ("(?i)^" + NODE_NAME_FURNITURE + "." + NODE_NAME_FLOOR + ".*([0-9]+)")
			nodeType_WALL: matchStr = ("(?i)^" + NODE_NAME_FURNITURE + "." + NODE_NAME_WALL + ".*([0-9]+)")
			nodeType_CEILING: matchStr = ("(?i)^" + NODE_NAME_FURNITURE + "." + NODE_NAME_CEILING + ".*([0-9]+)")
			default: matchStr = ""
		)

		match = rgx.Match imvuNode.name matchStr
		if match == undefined then
			imvuTools.printMsg #("Node name match failed!") msgType:#ERROR inScript:THIS_SCRIPT_FILENAME inFunction:getNodeNumber
		else if match.Groups.count > 1 then
			nodeNum = match.Groups.Item[1].Value as integer
		--if nodeNum == 0 then
		--	print ("getNodeNumber() Node number not found! imvuNode:" + imvuNode.name + " nodeNum:" + (nodeNum as string))

		--format "getNodeNumber() imvuNode:% nodeNum:%\n" imvuNode.name nodeNum -- DEBUG

		nodeNum
	)
	
	
	-- Utility Function
	-- Sort function for 'qsort' to sort an array of nodes by name
	-- Returns: None
	fn compareNodeNames n1 n2 =
	(
		local nodeNum1 = getNodeNumber n1
		local nodeNum2 = getNodeNumber n2
		case of
		(
			(nodeNum1 < nodeNum2): -1
			(nodeNum1 > nodeNum2): 1
			default: 0
		)
	)
	
	
	-- Utility Function
	-- Correctly increments the names of all nodes that will be exported
	-- Nodes need to be named 01-09, 10 + no zero padding, no duplicates and no gaps in the sequence otherwise N* will error.
	-- Nodes that are NOT to be exported should be numbered beginning with the highest exported + 1 and increase sequentially
	-- TODO: Need to modify this to work for both ROOM and FURNITURE tools. The nodes will need to be renamed/ordered for each export layer, not globally.
	-- Returns: None
	fn rebuildNodeNames =
	(
		local allExportNodes = #()
		local exportNodes = #()
		local noexportNodes = #()
		local seatChildren = #()
		local seatNodeCount = floorNodeCount = wallNodeCount = ceilingNodeCount = 0
		local objNodeType
		local nodeNum
		local pattern = "[0-9]+"

		append allExportNodes (getNodesInExportLayer EXPORT_LAYER_NAME_ROOM)

		for obj in objects do
		(
			objNodeType = getObjectNodeType(obj)
			if 	objNodeType == nodeType_SEAT or objNodeType == nodeType_FLOOR or objNodeType == nodeType_WALL or objNodeType == nodeType_CEILING then
				if findItem allExportNodes obj > 0 then
					append exportNodes obj
				else
					append noexportNodes obj
		)
		qsort exportNodes compareNodeNames
		qsort noexportNodes compareNodeNames

		for obj in exportNodes do
		(
			objNodeType =  getObjectNodeType(obj)
			case objNodeType of
			(
				nodeType_SEAT: nodeNum = formattedPrint (seatNodeCount+=1) format:"02d"
				nodeType_FLOOR: nodeNum = formattedPrint (floorNodeCount+=1) format:"02d"
				nodeType_WALL: nodeNum = formattedPrint (wallNodeCount+=1) format:"02d"
				nodeType_CEILING: nodeNum = formattedPrint (ceilingNodeCount+=1) format:"02d"
			)
			
			obj.name = rgx.Replace obj.name pattern nodeNum
			getSeatChildren obj &seatChildren
			for seatChild in seatChildren do 
				seatChild.name = rgx.Replace seatChild.name pattern nodeNum
		)
		for obj in noexportNodes do
		(
			objNodeType =  getObjectNodeType(obj)
			case objNodeType of
			(
				nodeType_SEAT: nodeNum = formattedPrint (seatNodeCount+=1) format:"02d"
				nodeType_FLOOR: nodeNum = formattedPrint (floorNodeCount+=1) format:"02d"
				nodeType_WALL: nodeNum = formattedPrint (wallNodeCount+=1) format:"02d"
				nodeType_CEILING: nodeNum = formattedPrint (ceilingNodeCount+=1) format:"02d"
			)

			obj.name = rgx.Replace obj.name pattern nodeNum
			getSeatChildren obj &seatChildren
			for seatChild in seatChildren do 
				seatChild.name = rgx.Replace seatChild.name pattern nodeNum
		)
	)
	
	
	-- Utility Function
	-- Get the next available incremented node number
	-- Parameters
	--		nodeType <integer> is a 'nodeType_' constant
	-- Returns <integer>
	fn getNextAvailableNodeNumber nodeType =
	(
		local nextNodeNum = 1
		local index = 1
		local nodeNums = #()
		local found = false

		for obj in objects do
		(
			if getObjectNodeType(obj) == nodeType then
				append nodeNums (getNodeNumber obj)
		)

		sort nodeNums
		--print nodeNums -- DEBUG
		
		while not found and index <= nodeNums.count do
		(
			if nodeNums[index] == nextNodeNum then
			(
				--format "match: nodeNums[%]:% nextNodeNum:%\n" index nodeNums[index] nextNodeNum -- DEBUG
				nextNodeNum += 1
			)
			
			if nodeNums[index] > nextNodeNum then
			(
				--format "NO MATCH: nodeNums[%]:% nextNodeNum:%\n" index nodeNums[index] nextNodeNum -- DEBUG
				found = true
			)
			index += 1
		)

		--print ("getNextAvailableNodeNumber() nextNodeNum:" + (nextNodeNum as string)) -- DEBUG
		nextNodeNum
	)


	-- Utility Function - Add/Remove seat child node (Handle, Catcher\Pitcher)
	-- Parameters:
	--		seatNode <object> is a node object
	--		childNodeType <integer> is a 'nodeType_' constant
	--		toolkitToolInst <ToolkitTool>
	--		addNode <boolean> Defaults to true. Function will add the specified child node by default.
	-- Returns: None
	fn addRemoveSeatChild seatNode childNodeType toolkitToolInst addNewNode:true =
	(
		if classof seatNode == TOOLKIT_NODE_CLASS and getObjectNodeType(seatNode) == nodeType_SEAT then
		(
			local seatNum
			local seatName
			local newSeatChild
			local newSeatPictcher
			local childArray = #()
		
			-- Get the seat name and number from the node
			--seatNum = GetSeatNumber seatNode
			seatNum = getNodeNumber seatNode
			seatName = getSeatName seatNode
			-- Valid that we got a number. seatNum will be undefined if match.value cannot be cast to an integer
			if seatNum == 0 then
				imvuTools.printMsg #("Seat number not found!") msgType:ERROR inScript:THIS_SCRIPT_FILENAME inFunction:addRemoveSeatChild
			else if seatName == "" then
				imvuTools.printMsg #("Seat name not found!") msgType:ERROR inScript:THIS_SCRIPT_FILENAME inFunction:addRemoveSeatChild
			else
			(
				if childNodeType == nodeType_HANDLE then
				(
					-- Check for the existence of the child
					findChildrenByNodeType seatNode nodeType_HANDLE &childArray

					-- If we are DELETING and there are existing handles delete them
					if not addNewNode then
						for child in childArray do
							delete child

					-- If we are ADDING a handle and there are non existing, do it
					if addNewNode and childArray.count == 0 then
					(
						newSeatChild = TOOLKIT_NODE_CLASS name:(NODE_NAME_HANDLE + (formattedPrint seatNum format:"02d")) pos:(seatNode.pos + HANDLE_NODE_OFFSET) size:HANDLE_NODE_SIZE box:on cross:on
						newSeatChild.parent = seatNode
						addNodesToExportLayer #(newSeatChild) toolkitToolInst.exportLayerName
					)
				)
				else if childNodeType == nodeType_CATCHER_PITCHER then
				(
					-- Check for the existence of the child
					findChildrenByNodeType seatNode nodeType_CATCHER_PITCHER &childArray

					-- If we are DELETING and there are existing catcher/pitchers delete them
					if not addNewNode then
						for child in childArray do
							delete child

					-- If we are ADDING a catcher/pitcher and there are non existing, do it
					if addNewNode and childArray.count == 0 then
					(
						newSeatChild = TOOLKIT_NODE_CLASS name:(NODE_NAME_CATCHER + (formattedPrint seatNum format:"02d") + NODE_NAME_SEPARATOR + seatName) pos:(seatNode.pos + CATCHER_NODE_OFFSET) size:PITCHER_CATCHER_NODE_SIZE box:on cross:on
						newSeatChild.parent = seatNode
						newSeatPictcher = TOOLKIT_NODE_CLASS name:(NODE_NAME_PITCHER + (formattedPrint seatNum format:"02d") + NODE_NAME_SEPARATOR + seatName) pos:(newSeatChild.pos + PITCHER_NODE_OFFSET) size:PITCHER_CATCHER_NODE_SIZE box:on cross:on
						rotate newSeatPictcher (eulerangles 0 0 180)
						newSeatPictcher.parent = newSeatChild
						addNodesToExportLayer #(newSeatChild, newSeatPictcher) toolkitToolInst.exportLayerName
					)
				)
			)
		)
	)


	-- Utility Function - Check two bones to see if they are in the same world space location and orientation
	-- Parameters:
	--		boneA <object> is a node object
	--		boneB <object> is a node object
	--		atFrame <int> Defaults to frame 0
	-- Returns: bonesMatch <boolean> True if bones match in orientation and position. False if not
	fn doBoneTransformsMatch boneA boneB atFrame:0=
	(
		local PRINT_DEBUG = false
		local bonesMatch = true
		local deltaPos, deltaRot
		local bonePosA = at time atFrame boneA.position
		local bonePosB = at time atFrame boneB.position
		local boneRotA = at time atFrame (boneA.transform * inverse WORLD_TRANSFORM_MATRIX).rotation
		local boneRotB = at time atFrame (boneB.transform * inverse WORLD_TRANSFORM_MATRIX).rotation

		deltaPos = distance bonePosA bonePosB
		deltaRot = (boneRotB - boneRotA).angle

		if deltaPos > ANIM_POSITION_THRESHOLD or deltaRot > ANIM_ROTATION_THRESHOLD then bonesMatch = false
		
		if PRINT_DEBUG or PRINT_DEBUG_ALL and not bonesMatch then
		(
			format "doBoneTransformsMatch():\n"
			if deltaPos > ANIM_POSITION_THRESHOLD then format "POSITIONs do not match:\n%:\t%\n%:\t%\n" boneA.name bonePosA boneB.name bonePosB
			if deltaRot > ANIM_ROTATION_THRESHOLD then format "ROTATIONs do not match:\n%:\t%\n%:\t%\nANGLE:%\n" boneA.name boneRotA boneB.name boneRotB deltaRot
		)

		bonesMatch
	)

	
	-- Utility Function - Check for keys recursively through the controller hierarchy
	-- Parameters:
	--		cntrlr <controller> Node [root] controller to recursively look for keys on
	--		atTime <interval> [optional] Time / frame delete keys on. Defaults to slider time when not provided
	-- Return value: hasKeys <boolean> True if keys were found, False if not. Controls the recursion. Stops after first keyed controller is found
	fn isControllerKeyedAtTime cntrlr atTime:sliderTime = if iscontroller cntrlr do
	(
		local PRINT_DEBUG = false
		local hasKeys = getKeyIndex cntrlr atTime > 0

		if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("A -", cntrlr, (if hasKeys then ": HAS KEYS" else ": NOPE")) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:isControllerKeyedAtTime

		if not hasKeys do --and cntrlr.numsubs > 0 do
		(
			for ii = 1 to cntrlr.numsubs
				while not hasKeys where iscontroller cntrlr[ii].controller do
				(
					hasKeys = isControllerKeyedAtTime cntrlr[ii].controller atTime:atTime

					if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("B -", cntrlr, (if hasKeys then ": HAS KEYS" else ": NOPE")) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:isControllerKeyedAtTime
				)
		)
		if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("C -", cntrlr, (if hasKeys then ": HAS KEYS" else ": NOPE")) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:isControllerKeyedAtTime

		hasKeys
	)


	-- Utility Function - Look for keys recursively through the controller hierarchy and collect this in an array passed by reference
	-- Parameters:
	--		cntrlr <controller> Node [root] controller to recursively look for keys on
	--		&keyframes <array> List of keyframes. Passed by reference and gets appended if unique frames are found.
	-- Return value: hasKeys <boolean> True if keys were found, False if not. Controls the recursion.
	fn collectControllerKeyframes cntrlr &keyframes = if iscontroller cntrlr do
	(
		local PRINT_DEBUG = false
		local keyCount = numkeys cntrlr
		local hasKeys = keyCount > 0
		local tempKeys = #()

		if hasKeys then
		(
			for ii = 1 to keyCount do append tempKeys (getKeyTime cntrlr ii)
			if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #(cntrlr, ":", tempKeys) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:collectControllerKeyframes

			join keyframes tempKeys
			keyframes = makeUniqueArray keyframes
			--for ii = 1 to keyCount do appendIfUnique keyframes (getKeyTime cntrlr ii)
		)

		if not hasKeys do
		(
			for ii = 1 to cntrlr.numsubs
				while not hasKeys where iscontroller cntrlr[ii].controller do
					hasKeys = collectControllerKeyframes cntrlr[ii].controller &keyframes
		)
		hasKeys
	)


	-- Utility Function - Delete all keys on an object recursively through the transform controller hierarchy
	-- Background: 3ds Max Key Functions only allow you to delete keys #all or by #selection and by selection you still need the controller
	-- Parameters:
	--		cntrlr <controller> Node [root] controller to recursively look for keys on
	--		atFrames <array of integers or interval> [optional] Time / frame delete keys on. Defaults to 'all frames' when not provided
	-- Return value: None
	fn deleteAllKeysOnControllerAtTime cntrlr atFrames:#() = if isController cntrlr do
	(
		--format "deleteAllKeysOnControllerAtTime() cntrlr:% atFrames:%\n" cntrlr atFrames -- DEBUG
		local keyIndex

		if atFrames.count == 0 then
		(
			deleteKeys cntrlr #allKeys

			if cntrlr.numsubs > 0 do
				for ii = 1 to cntrlr.numsubs do deleteAllKeysOnControllerAtTime cntrlr[ii].controller
		)
		else
		(
			for ii = 1 to atFrames.count do
			(
				keyIndex = getKeyIndex cntrlr atFrames[ii]

				if keyIndex > 0 then deleteKey cntrlr keyIndex

				if cntrlr.numsubs > 0 do
					for jj = 1 to cntrlr.numsubs do deleteAllKeysOnControllerAtTime cntrlr[jj].controller atFrames:#(atFrames[ii])
			)
		)
	)


	/* CRUFT - Not needed as reduceKeys will act on a collection, but keep around for now for reference.
				TODO: write my own key reduction algorithm because for reduceKeys
					"As with the reducer in the track view, a progress bar and cancel button is displayed."
					See Maxscript documentation for more info: https://help.autodesk.com/view/3DSMAX/2017/ENU/?guid=__files_GUID_E6615B92_8BC4_4A13_BC31_86E006657A5C_htm
	-- Utility Function - Reduce keys on and object recursively through the transform controller hierarchy
	-- Parameters:
	--		cntrlr <controller> Node [root] controller to recursively look for keys on
	-- Return value: None
	fn reduceKeysOnController cntrlr = if isController cntrlr do
	(
		local keyReductionThreshold = 0.1
		local timeStep = 1f -- every frame

		reduceKeys cntrlr keyReductionThreshold timeStep

		if cntrlr.numsubs > 0 do
			for ii = 1 to cntrlr.numsubs do reduceKeysOnController cntrlr[ii].controller
	)
	*/


	-- Utility Function -  Get the highest key on any widget or export skeleton bone and return the animation length
	--		nodes <array> Array of nodes. Defaults to all export skeleton bones and widgets
	-- Return value: highestKeyTime <frame_integer> which could be 0 if there are no keys or only keys on frame 0
	fn getAnimLength nodes:(avatarAnimToolVars.allWidgets + allAvatarExportBones) =
	(
		local PRINT_DEBUG = false
		local allFrames = #()
		local highestKeyTime

		for node in nodes do
			collectControllerKeyframes node.transform.controller &allFrames

		sort allFrames

		highestKeyTime = if allFrames.count > 0 then allFrames[allFrames.count] else 0
		if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("Animation range length is ", highestKeyTime) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:getAnimLength
		highestKeyTime
	)


	-- Utility Function -  Disable all animation modes 'Auto Key' and 'Set Key'
	-- Return value: None
	fn disableAnimationModes =
	(
		-- Always set 'Auto Key' state to the default (off). 'animButtonState' is a maxscript global variable.
		animButtonState = ANIM_AUTO_DEFAULT_OFF -- Button and UI
		set animate off -- Actual auto animation state
		-- Always set the manual 'Set Key' state to the default (off). 'max set key mode' is a built in 3ds Max command that toggles the 'Set Key'
		if maxOps.setKeyMode then max set key mode
	)


	-- BEGIN - TOOL FUNCTIONS - Functions called from tool UI  ----------------------------------------------------------------------------------------------------------------------------------------
	
 	-- Tool Function - Set up the selected object(s) for export for the appropriate tool
	-- Used by the FURNITURE and TBD. Could be used by all tools(?)
	-- Parameters:
	--		toolExportLayer <string>
	-- Return value: None
 	fn setSelectedForExport rootNode toolExportLayerName =
 	(
 		local selObjArray = selection as array
 		local obj
 		local nodesToActOn = #()
 		local nodesToParentToAttachmentRoot = #()

		-- Collect a list of nodes in the selection we "actually" need to move to the export layer.
		--	That would be everything from the selected 'obj' down in the hierarchy including children that are NOT in the selection.
 		for obj in selObjArray do imvuTools.getHierarchy obj &nodesToActOn
		addNodesToExportLayer nodesToActOn toolExportLayerName

		-- Collect a list of what in our selection we "actually" need to parent to the "root" node.
		nodesToActOn = imvuTools.getRootNodes selObjArray
		if rootNode != undefined then for obj in nodesToActOn do obj.parent = rootNode
 	)


 	-- Tool Function - Add seat node
	-- Used by the ROOM and FURNITURE Tools
	-- Parameters:
	-- 		seatTypeName <string> One of the IMVU base types Standing, Sitting, Idle, or a custom name
	--		toolkitToolInst <ToolkitTool>
	--		addHandle <boolean>
	--		addCatcherPitcher <boolean>
	--		atPosition <point3>
	-- Return value: None
	fn addSeat seatTypeName toolkitToolInst addHandle:true addCatcherPitcher:true atPosition:[0.0, 0.0, 0.0] =
	(
		-- print ("addSeat() seatTypeName:" + seatTypeName + " addHandle:" +  (addHandle as string) + " addCatcherPitcher:" + (addCatcherPitcher  as string) + " atPosition:" + atPosition as string) -- DEBUG

		--local seatNodeParent = getNodeByName NODE_NAME_SEAT_PARENT
		local seatNodeParent
		local newSeatNum
		local newSeat
		local newSeatChild
		local newSeatPictcher
	
		if toolkitToolInst.id == toolId_ROOM then
			seatNodeParent = execute ("$'" + NODE_NAME_ROOM_ROOT + "'/" + NODE_NAME_SEAT_PARENT)
		else if toolkitToolInst.id == toolId_FURNITURE then
			seatNodeParent = execute ("$'" + NODE_NAME_FURNITURE_ROOT + "'/" + NODE_NAME_SEAT_PARENT)

		-- TODO: Add a provision for either creating the seatParentNode if not found OR put seats under currently selected node OR something else maybe?
		if seatNodeParent != undefined then
		(
			newSeatNum = getNextAvailableNodeNumber nodeType_SEAT
			newSeat = TOOLKIT_NODE_CLASS name:(NODE_NAME_SEAT + (formattedPrint newSeatNum format:"02d") + NODE_NAME_SEPARATOR + seatTypeName as string) pos:atPosition size:DEFAULT_NODE_SIZE box:on cross:on
			newSeat.parent = seatNodeParent

			if newSeat  != undefined then
			(
				if addHandle then
				(
					addRemoveSeatChild newSeat nodeType_HANDLE toolkitToolInst
				)
					
				if addCatcherPitcher then
				(
					addRemoveSeatChild newSeat nodeType_CATCHER_PITCHER toolkitToolInst
				)

				rebuildNodeNames()
			)
		)
		addNodesToExportLayer #(newSeat) toolkitToolInst.exportLayerName
		newSeat
	)


	-- Tool Function - Create SeeThrough Node
	-- Used by the ROOM and FURNITURE Tools
	-- Parameters:
	--		toolkitToolInst <ToolkitTool>
	-- Return value: None
	fn createSeeThroughNode toolkitToolInst =
	(
		local exportLayer
		local seeThroughNodeParent
		local seeThroughNum = getNextAvailableNodeNumber nodeType_SEETHROUGH
		local seeThroughNode = Cone name:(NODE_NAME_SEETHROUGH + formattedPrint seeThroughNum format:"02d") radius1:SEE_THROUGH_NODE_SIZE.x radius2:SEE_THROUGH_NODE_SIZE.y height:SEE_THROUGH_NODE_SIZE.z

		seeThroughNode.material = standardMaterial name:"M_SeeThrough" opacity:50
		rotate seeThroughNode (eulerangles 90 0 0)
		select seeThroughNode
		
		if toolkitToolInst.id == toolId_ROOM then
			seeThroughNodeParent = getNodeByName NODE_NAME_ROOM_ROOT
		else if toolkitToolInst.id == toolId_FURNITURE then
			seeThroughNodeParent = getNodeByName NODE_NAME_FURNITURE_ROOT
		if seeThroughNodeParent != undefined then
			seeThroughNode.parent = seeThroughNodeParent
		
		addNodesToExportLayer #(seeThroughNode) toolkitToolInst.exportLayerName
	)
	
	
	-- Tool Function - Link SeeThrough Node to selected mesh(es)
	-- Used by the ROOM and FURNITURE Tools
	-- Return value: None
	fn linkSeeThroughNode =
	(
		local selObjArray = selection as array
		local seeThroughNodesInSelection = #()
		local seeThroughNum = getNextAvailableNodeNumber nodeType_SEETHROUGH
		local seeThroughNodeLinkSuccess = false
		local seeThroughSkinModifier
		
		getNodesByNodeTypeInArray nodeType_SEETHROUGH selObjArray &seeThroughNodesInSelection
		
		if selObjArray.count == 0 then
			messageBox "Nothing Selected!\nSelect one SeeThrough node and one or more meshes to bind to and try again." title:TOOLKIT_NAME
		else if seeThroughNodesInSelection.count > 1 then
			messageBox "Too many SeeThrough nodes selected!\nPlease select only one SeeThrough node and try again." title:TOOLKIT_NAME
		else if seeThroughNodesInSelection.count == 0 then
			messageBox "No SeeThrough nodes selected!\nPlease select one SeeThrough node and try again." title:TOOLKIT_NAME
		else
		(
			-- Remove the see through node from the selection array
			deleteItem selObjArray (findItem selObjArray seeThroughNodesInSelection[1])
			deselect objects
			max modify mode
			for selObj in selObjArray do
			(
				-- Is the object a valid object to add a skin modifier to?
				if validModifier selObj Skin then
				(
					-- See if there is already a see through skin modifier applied.
					-- TODO : Maybe we want to allow multiple bone assignments?? Does this work? Need to test it.
					--		For now do we just skip objects that already have the skin modifier applied?
					--		OR
					--		Do we remove the current bone assignment and reassign to the current one selected?
					--	For skip objects with existing see through skin modifier applied.
					seeThroughSkinModifier = selObj.modifiers[MODIFIER_NAME_SEETHROUGHSKIN as name]
					if seeThroughSkinModifier == undefined then
					(
						seeThroughSkinModifier = Skin name:MODIFIER_NAME_SEETHROUGHSKIN
						
						--suspendEditing which:#modify
						select selObj
						addModifier selObj seeThroughSkinModifier
						--resumeEditing which:#modify
						modPanel.setCurrentObject selObj.modifiers[MODIFIER_NAME_SEETHROUGHSKIN as name]
						skinOps.addBone seeThroughSkinModifier seeThroughNodesInSelection[1] -1
						
						seeThroughNodeLinkSuccess = true
					)
				)
			)
			if not seeThroughNodeLinkSuccess then
				messageBox "No valid objects selected to link to SeeThrough node!" title:TOOLKIT_NAME
		)
	)
	

	-- Tool Function - Unlink SeeThrough Node from selected mesh(es)
	-- Used by the ROOM and FURNITURE Tools
	-- Return value: None
	fn unlinkSeeThroughNode =
	(
		local selObjArray = selection as array
		local selObjModifier

		if selObjArray.count == 0 then
			messageBox "Nothing Selected!\nSelect one or more meshes unbind See Through nodes to and try again." title:TOOLKIT_NAME
		else
		(
			--suspendEditing which:#modify
			for selObj in selObjArray do
			(
				selObjModifier = selObj.modifiers[MODIFIER_NAME_SEETHROUGHSKIN as name]
				while (selObjModifier != undefined) do
				(
					deleteModifier selObj selObjModifier
					selObjModifier = selObj.modifiers[MODIFIER_NAME_SEETHROUGHSKIN as name]
				)
			)
			--resumeEditing which:#modify
		)
	)

	
	-- Tool Function - Duplicate furniture nodes
	-- Used by the ROOM Tool
	-- Parameters:
	--		dir <string> Direction the duplicates will be offset
	--		dist <float> The distance as specified in the UI to offset the duplicates. NOTE: right now the displayed number is a fraction of the actual amount. May want to change this.
	-- Return value: None
	fn duplicateFurnitureNodes dir dist =
	(
		--format "dir: % dist: %\n" dir dist --DEBUG
		-- Check for selection
		local selObjs = selection as array
		if selObjs.count > 0 then
		(			
			local offsetFactor = 1.0 -- in meters TODO: need to know if toolkit is in 'Legacy' (100x) mode or not
			local offset = [0.0, 0.0, 0.0]

			case dir of
			(
				"N": offset.y = dist
				"S": offset.y = -dist
				"E": offset.x = dist
				"W": offset.x = -dist
			)
			offset *= offsetFactor
			
			-- Build an array of nodes to act upon from the selection
			furnitureNodes = #()
			for obj in selObjs do
			(
				objNodeType = getObjectNodeType(obj)
				if 	 objNodeType == nodeType_FLOOR or objNodeType == nodeType_WALL or objNodeType == nodeType_CEILING then
					append furnitureNodes obj
			)
			if furnitureNodes.count > 0 then
			(
				maxOps.cloneNodes furnitureNodes cloneType:#copy newNodes:&newNodes
				for newNode in newNodes do
				(
					if getObjectNodeType(newNode) == nodeType_CEILING then
						in coordsys local newNode.pos += [offset.x * -1, offset.y, offset.z]
					else
						in coordsys local newNode.pos += offset
				)

				rebuildNodeNames()
			)
		)
	)


	-- Tool Function - Add room furniture node
	-- Used by the ROOM Tool
	-- Parameters:
	--		furnitureType <integer> The furniture type as specified by the radio button in the UI. Valid numbers are 1 (Floor), 2 (Wall), and 3 (Ceiling).
	--		furnitureDir <integer> The wall that a Wall furniture node should be placed on as specified by the UI. Valid numbers are 1 (North ), 2 (South), 3 (East), 4 (West).
	-- Return value: None
	fn addRoomFurnitureNode furnitureType furnitureDir =
	(
		local roomRef = getNodeByName ROOM_REF_MESH_NAME
		local roomRefBoundingBoxHalfSize
		local roomRefLocalBoundingBox
		local furnitureNodeParent = getNodeByName NODE_NAME_FURNITURE_PARENT
		local newFurnitureNode
		local newFurnitureNumString

		if furnitureNodeParent == undefined then
			messageBox ("ERROR: addRoomFurnitureNode() furnitureNodeParent: " + NODE_NAME_FURNITURE_PARENT + " not found!") title:TOOLNAME_ROOM
		else
		(
			if roomRef != undefined then
			(
				roomRefLocalBoundingBox = nodeLocalBoundingBox roomRef
				roomRefBoundingBoxHalfSize = (roomRefLocalBoundingBox[1] - roomRefLocalBoundingBox[2]) * 0.5
				roomRefBoundingBoxHalfSize = [abs roomRefBoundingBoxHalfSize.x, abs roomRefBoundingBoxHalfSize.y, abs roomRefBoundingBoxHalfSize.z]
			)

			local furnitureNodePos = (roomRefLocalBoundingBox[1] + roomRefLocalBoundingBox[2]) * 0.5
			local furnitureNodeRot = [0, 0, 0]
			
			--format "stats\nroomRefBoundingBoxHalfSize:%\nroomRefLocalBoundingBox:%\nfurnitureNodePos:%\n\n" roomRefBoundingBoxHalfSize roomRefLocalBoundingBox furnitureNodePos -- DEBUG

			case furnitureType of
			(
				1: -- Floor
				(
					furnitureNodePos.z -= roomRefBoundingBoxHalfSize.z
					newFurnitureNumString = formattedPrint (getNextAvailableNodeNumber nodeType_FLOOR) format:"02d"
					newFurnitureNode = TOOLKIT_NODE_CLASS name:(NODE_NAME_FURNITURE + NODE_NAME_SEPARATOR + NODE_NAME_FLOOR + NODE_NAME_SEPARATOR + newFurnitureNumString) pos:furnitureNodePos size:DEFAULT_NODE_SIZE box:on cross:on
				)
				2: -- Wall
				(
					furnitureNodeRot = [-90, 180, 0]
					-- furnitureDir labels:#("North   ","South   ","East   ","West")
					case furnitureDir of
					(
						1: -- North (-Y)
						(
							furnitureNodePos.y += roomRefBoundingBoxHalfSize.y
							furnitureNodeRot.z = -180
						)
						2: -- South (+Y)
							furnitureNodePos.y -= roomRefBoundingBoxHalfSize.y
						3: -- East (+X)
						(
							furnitureNodePos.x += roomRefBoundingBoxHalfSize.x
							furnitureNodeRot.z = 90
						)
						4: -- West (-X)
						(
							furnitureNodePos.x -= roomRefBoundingBoxHalfSize.x
							furnitureNodeRot.z = -90
						)
					)
					newFurnitureNumString = formattedPrint (getNextAvailableNodeNumber nodeType_WALL) format:"02d"
					newFurnitureNode = TOOLKIT_NODE_CLASS name:(NODE_NAME_FURNITURE + NODE_NAME_SEPARATOR + NODE_NAME_WALL + NODE_NAME_SEPARATOR + newFurnitureNumString) pos:furnitureNodePos size:DEFAULT_NODE_SIZE box:on cross:on
				)
				3: -- Ceiling
				(
					furnitureNodePos.z += roomRefBoundingBoxHalfSize.z
					furnitureNodeRot = [0, 180, 0]
					newFurnitureNumString = formattedPrint (getNextAvailableNodeNumber nodeType_CEILING) format:"02d"
					newFurnitureNode = TOOLKIT_NODE_CLASS name:(NODE_NAME_FURNITURE + NODE_NAME_SEPARATOR + NODE_NAME_CEILING + NODE_NAME_SEPARATOR + newFurnitureNumString) pos:furnitureNodePos size:DEFAULT_NODE_SIZE box:on cross:on
				)
			)
			in coordsys local rotate newFurnitureNode (eulerangles furnitureNodeRot.x furnitureNodeRot.y furnitureNodeRot.z)
			newFurnitureNode.parent = furnitureNodeParent
			
			-- Add the nodes to the export layer
			addNodesToExportLayer #(newFurnitureNode) roomTool.exportLayerName
		)
	)


	-- Tool Function - Resize the room
	-- Used by the ROOM Tool
	-- Return value: None
	-- TODO: Fix how the size of the room is set. Currently it modifies the scale, but this is problematic, because if the users modifies the verts then scale is no longer relevant.
	fn resizeRoom roomSize =
	(
		--local newScale
		local meshRef = getNodeByName ROOM_REF_MESH_NAME
		
		if meshRef != undefined then
		(
			meshRef.scale = roomSize
			--newScale = #((XSlider.value * DEFAULT_ROOM_SCALE_FACTOR), (YSlider.value * DEFAULT_ROOM_SCALE_FACTOR), (ZSlider.value * DEFAULT_ROOM_SCALE_FACTOR))
			--meshRef.scale = [newScale[1] /  ROOM_REF_SCALE_FACTOR.x, newScale[2] /  ROOM_REF_SCALE_FACTOR.y, newScale[3] /  ROOM_REF_SCALE_FACTOR.z]
		)
	)


	-- Tool Function - Create Room
	-- Used by the ROOM Tool
	-- Parameters:
	--		roomSize <point3> - The size of the room to create
	-- Return value: None
	fn createRoom roomSize =
	(
		local roomRootNode, lightingNodes, cameraTargetNode
		local exportNodes = #()

		local meshRef = getNodeByName ROOM_REF_MESH_NAME
		if meshRef != undefined then
			delete meshRef

		-- Create room nodes
		roomRootNode = TOOLKIT_NODE_CLASS name:NODE_NAME_ROOM_ROOT pos:WORLD_ORIGIN_POSITION size:SKELETON_ROOT_NODE_SIZE box:on cross:on
		append exportNodes roomRootNode
		append exportNodes (TOOLKIT_NODE_CLASS name:NODE_NAME_SEAT_PARENT parent:roomRootNode pos:WORLD_ORIGIN_POSITION size:NODE_SIZE_ZERO box:off cross:on)
		append exportNodes (TOOLKIT_NODE_CLASS name:NODE_NAME_FURNITURE_PARENT parent:roomRootNode pos:WORLD_ORIGIN_POSITION size:NODE_SIZE_ZERO box:off cross:on)
		lightingNodes = TOOLKIT_NODE_CLASS name:NODE_NAME_LIGHTING parent:roomRootNode pos:WORLD_ORIGIN_POSITION size:NODE_SIZE_ZERO box:off cross:on
		append exportNodes lightingNodes

		-- Create the camera
		append exportNodes (cameraTargetNode = Targetobject name:CAMERA_TRAGET_NAME parent:roomRootNode transform:(matrix3 [1,0,0] [0,1,0] [0,0,1] CAMERA_TARGET_POS))
		append exportNodes (Targetcamera name:CAMERA_NAME parent:roomRootNode fov:45 nearclip:1 farclip:1000 nearrange:0 farrange:1000 mpassEnabled:off mpassRenderPerPass:off pos:CAMERA_POS isSelected:off target:(cameraTargetNode))

		-- Create Lighting
		append exportNodes (Omnilight name:OMNI_LIGHT_NAME  parent:lightingNodes rgb:COLOR_WHITE shadowColor:COLOR_BLACK multiplier:1 contrast:0 softenDiffuseEdge:0 nearAttenStart:0 nearAttenEnd:40 farAttenStart:80 farAttenEnd:200 decayRadius:40 atmosOpacity:100 atmosColorAmt:100 shadowMultiplier:1 pos:OMNI_LIGHT_POSITION)

		-- Create room mesh
		roomMesh = Box name:ROOM_REF_MESH_NAME parent:roomRootNode width:ROOM_REF_BOX_SIZE.x length:ROOM_REF_BOX_SIZE.y height:ROOM_REF_BOX_SIZE.z
		roomMesh.scale = roomSize
		append exportNodes roomMesh
		convertTo roomMesh TriMeshGeometry
		meshop.flipNormals roomMesh #all
		roomMesh.backfacecull = on
		roomMesh.material = standardMaterial name:"M_Room" diffuse:ROOM_COLOR

		-- Create scale reference object
		append exportNodes (Box name:FURNITURE_REF_MESH_NAME parent:roomRootNode width:SCALE_REFERENCE_SIZE.x length:SCALE_REFERENCE_SIZE.y height:SCALE_REFERENCE_SIZE.z pos:SCALE_REFERENCE_POS)

		-- Make 6 default seats without a catcher/pitcher
		for ii = 1 to 6 do
		(
			append exportNodes (addSeat NODE_NAME_SEAT_STANDING  roomTool addHandle:false addCatcherPitcher:false atPosition:((ii-1)*NEW_SEAT_OFFSET))
		)
		
		-- Set all nodes to the Export Layer 
		addNodesToExportLayer exportNodes roomTool.exportLayerName
	)
	
	
	-- Tool Function - Create Furniture
	-- Used by the FURNITURE Tool
	-- Return value: None
	fn createFurniture =
	(
		local furnitureRootNode
		local exportNodes = #()

		furnitureRootNode = TOOLKIT_NODE_CLASS name:NODE_NAME_FURNITURE_ROOT pos:WORLD_ORIGIN_POSITION size:SKELETON_ROOT_NODE_SIZE box:on cross:on
		append exportNodes furnitureRootNode
		append exportNodes (Box name:FURNITURE_REF_MESH_NAME parent:furnitureRootNode width:FURNITURE_REFERENCE_CUBE_SIZE.x length:FURNITURE_REFERENCE_CUBE_SIZE.y height:FURNITURE_REFERENCE_CUBE_SIZE.z pos:WORLD_ORIGIN_POSITION)
		append exportNodes (TOOLKIT_NODE_CLASS name:NODE_NAME_FURNITURE_NODES_PARENT parent:furnitureRootNode pos:WORLD_ORIGIN_POSITION size:DEFAULT_NODE_SIZE box:on cross:on)
		append exportNodes (TOOLKIT_NODE_CLASS name:NODE_NAME_SEAT_PARENT parent:furnitureRootNode pos:WORLD_ORIGIN_POSITION size:NODE_SIZE_ZERO box:off cross:on)
		
		-- Set all nodes to the Export Layer 
		addNodesToExportLayer exportNodes furnitureTool.exportLayerName
	)
	
	
	-- Tool Function - Set the visibility of the avatar mesh
	-- Used by the ATTACHMENT and CLOTHING Tools
	-- Parameters:
	--		visAvatar <integer> is a 'visAvatar_' constant
	--		setUI <bool> defaults to false. Used to set the UI when not called by a UI change
	-- Return value: None
	fn setVisAvatar visAvatar setUI:false =
	(
		hide (execute ("$" + MEGASCENE_NODE_NAME_MALE_MESH + "*"))
		hide (execute ("$" + MEGASCENE_NODE_NAME_FEMALE_MESH + "*"))
		hide (execute ("$" + MEGASCENE_NODE_NAME_PROXY_MESH + "*"))
			
		case visAvatar of
		(
			visAvatar_MALE: unhide (execute ("$" + MEGASCENE_NODE_NAME_MALE_MESH + "*"))
			visAvatar_FEMALE: unhide (execute ("$" + MEGASCENE_NODE_NAME_FEMALE_MESH + "*"))
			visAvatar_PROXY: unhide (execute ("$" + MEGASCENE_NODE_NAME_PROXY_MESH + "*"))
		)
		redrawViews()

		if setUI then visualsTool.toolRollout.displayAvatar.state = (findItem VIS_AVATAR_LABEL_ORDER visAvatar)

	)
	-- Store a pointer to the function in the tool var
	visualsToolVars.setVisibleAvatar = setVisAvatar

	
	-- Tool Function - Set the freeze state of the avatar meshes
	-- Parameters:
	--		freeze <bool> defaults to false.
	-- Return value: None
	fn setFreezeStateAvatars doFreeze:false =
	(
		if doFreeze then freeze (execute ("$" + MEGASCENE_NODE_NAME_MALE_MESH + "*")) else unfreeze (execute ("$" + MEGASCENE_NODE_NAME_MALE_MESH + "*"))
		if doFreeze then freeze (execute ("$" + MEGASCENE_NODE_NAME_FEMALE_MESH + "*")) else unfreeze (execute ("$" + MEGASCENE_NODE_NAME_FEMALE_MESH + "*"))
			
		redrawViews()
	)
	-- Store a pointer to the function in the tool var
	visualsToolVars.setFreezeAvatars = setFreezeStateAvatars

	
	-- Tool Function - Create Attachment
	-- Used by the ATTACHMENT Tool
	-- Parameters: None
	-- Return value: None
	fn createAttachment =
	(
		local exportNodes = #()

		mergeMaxFile MEGASCENE MEGASCENE_GROUPS
		ungroup (getNodeByName MEGASCENE_GROUPS[megaSceneGroup_AVATAR_RIG])

		visualsToolVars.setVisibleAvatar visAvatar_FEMALE setUI:true

		-- Freeze the avatar meshes because there is no need for the user to select them and they will get in the way when picking bones
		visualsToolVars.setFreezeAvatars doFreeze:true

		setToolkitToolVariables()

		-- Unhide the avatar export skeleton bones so that the user can pick from them when setting up the attachment
		unhide allAvatarExportBones

		-- Create root node
		append exportNodes (TOOLKIT_NODE_CLASS name:NODE_NAME_ATTACHMENT_ROOT pos:WORLD_ORIGIN_POSITION size:SKELETON_ROOT_NODE_SIZE box:on cross:on)
		
		-- Set all nodes to the Export Layer 
		addNodesToExportLayer exportNodes attachmentTool.exportLayerName

		redrawViews()
	)
	
	
	-- Tool Function - Test if object is a bone
	-- Used by the ATTACHMENT Tool to filter picker
	-- Parameters:
	--		object <node>
	-- Return value: <boolean>
	fn isBone obj = classof obj == BoneGeometry or superclassOf obj == Helper
	

	-- Tool Function - Setups up the user's assets to be export ready as an attachment product
	-- Used by the ATTACHMENT Tool
	-- Parameters:
	--		pickedBone <node> The bone that the user has select and is set in the tool UI
	--		setupType <integer> The type of set up to be performed as specified by the UI. Valid numbers are 1 (Mesh(es)), 2 (Skeleton)
	-- Return value: None
	--
	-- TODO: Error checking for selected objects within the same hierarchy. The desired result would be to alert the user that they have
	--		parents and children within the same hierarchy selected and warn them that this may have undesirable results and maybe they should select the hierarchy root?
	fn setUpAttachment pickedBone reorient =
	(
		local selObjArray = selection as array
		local exportLayer = LayerManager.getLayerFromName EXPORT_LAYER_NAME_ATTACHMENT

		-- Later if we need absolute node path. Use "'" around any single object name to safeguard the case where there is a "." (dot) in the object name
		--local pickedBoneObject = execute("$'<the_root_node_name>'/.../$" + pickedBone)
		local pickedBoneObject = getNodeByName pickedBone
		local attachmentRootNode = execute ("$'" + NODE_NAME_ATTACHMENT_ROOT + "*'") as array -- This has to be cast to an array because otherwise we get a PathName object and not a Node
		
		if pickedBoneObject == undefined then
			messageBox "No bone picked!\nUse the Pick Bone button to select a bone to align to and try again." title:TOOLNAME_ATTACHMENT
		else if selObjArray.count == 0 then
			messageBox "Nothing Selected!\nSelect one or more meshes to setup as an attachment and try again." title:TOOLNAME_ATTACHMENT
		else if attachmentRootNode.count != 1 then
			messageBox "We found more than one Attachment Root Node!\nSomething is wrong." title:TOOLNAME_ATTACHMENT
		else
		(
			-- Store object positions and rotations so that we can reset them if the user wishes to.
			attachmentToolVars.attachmentRoot = AttachmentObject obj:attachmentRootNode[1] pos:attachmentRootNode[1].position rot:attachmentRootNode[1].rotation
			attachmentToolVars.attachmentObjects = #()
			for selObj in selObjArray do
			(
				append attachmentToolVars.attachmentObjects (AttachmentObject obj:selObj objParent:selObj.parent pos:selObj.position rot:selObj.rotation)
				-- Add the object to the export layer
				exportLayer.addNode selObj
			)
			attachmentToolVars.reorientState = reorient

			if reorient then -- User wants to parent the selected object(s) to the Attachment Root bone and then align the Attachment Root bone to the specified avatar bone.
			(
				-- Parent the object to the Attachment Root node. This will allow for easy adjustment of the object after setup, as opposed to if it was skinned to the Attachment Root node.
				for selObj in selObjArray do selObj.parent = attachmentRootNode[1]
				-- Move the attachment root bone into position
				attachmentRootNode[1].rotation = pickedBoneObject.rotation
				attachmentRootNode[1].position = pickedBoneObject.position
			)
			else
			(
				-- Move the attachment root bone into position
				attachmentRootNode[1].rotation = pickedBoneObject.rotation
				attachmentRootNode[1].position = pickedBoneObject.position
				-- Parent the object to the Attachment Root node. This will allow for easy adjustment of the object after setup, as opposed to if it was skinned to the Attachment Root node.
				for selObj in selObjArray do selObj.parent = attachmentRootNode[1]
			)
			-- Rename the bone for ingestion into IMVU Studio so that the skeleton Node gets automatically assigned to the 'pickedBone' 
			attachmentRootNode[1].name = NODE_NAME_ATTACHMENT_ROOT + "[" + pickedBone + "]"
		)
	)


	-- Tool Function - Resets the Attachment Root and user's assets to the position(s) and rotation(s) they were in before the Setup
	-- Used by the ATTACHMENT Tool
	-- Parameters: None
	-- Return value: None
	fn resetAttachment =
	(
		if attachmentToolVars.reorientState then
		(
			attachmentToolVars.attachmentRoot.obj.pos = attachmentToolVars.attachmentRoot.pos
			attachmentToolVars.attachmentRoot.obj.rotation = attachmentToolVars.attachmentRoot.rot
			for obj in attachmentToolVars.attachmentObjects do
			(
				obj.obj.parent = obj.objParent
				obj.obj.pos = obj.pos
				obj.obj.rotation = obj.rot
			)
		)
		else
		(
			for obj in attachmentToolVars.attachmentObjects do
			(
				obj.obj.parent = obj.objParent
				obj.obj.pos = obj.pos
				obj.obj.rotation = obj.rot
			)
			attachmentToolVars.attachmentRoot.obj.pos = attachmentToolVars.attachmentRoot.pos
			attachmentToolVars.attachmentRoot.obj.rotation = attachmentToolVars.attachmentRoot.rot
		)
	)


	-- Tool Function - Set up a clothing
	-- Used by the CLOTHING Tool
	-- Parameters: None
	-- Return value: None
	fn createClothing =
	(
		mergeMaxFile MEGASCENE MEGASCENE_GROUPS
		ungroup (getNodeByName MEGASCENE_GROUPS[megaSceneGroup_AVATAR_RIG])
		
		visualsToolVars.setVisibleAvatar visAvatar_FEMALE setUI:true

		setToolkitToolVariables()

		-- Set all nodes to the Export Layer 
		addNodesToExportLayer allAvatarExportBones clothingTool.exportLayerName
	)
	
	
	-- Tool Function - Remove the unweighted bones from the skin modifier
	-- Used by the CLOTHING Tool
	-- Parameters: skinModifier <skin_modifier>
	-- Return value: None
	fn removeUnweightedBones skinModifier =
	(
		local selObjArray = selection as array
		local selObj = selObjArray[1]
		local boneVerts = #{} -- Bit array of weighted bones
		local vertexBoneInfluencesCount
		local vertBoneIndex
		local boneIndex
		local numBones
		local bonesToRemove = #()
		local bonesToKeep = #()
		local boneName

		if selObjArray.count == 0 then
			messageBox "Nothing Selected!\n" title:TOOLNAME_CLOTHING
		else if selObjArray.count > 1 then
			messageBox "More than one object Selected!\n" title:TOOLNAME_CLOTHING
		-- Is the object a valid object to paint weights?
		else if not validModifier selObj Skin then
			messageBox "No skinned object selected!\n" title:TOOLNAME_CLOTHING
		else if not skinModifier == undefined then
		(
			if skinModifier.name == MODIFIER_NAME_CLOTHINGSKIN then
			(
				-- Unfortunately this is very slow, but we need to check all vertices for bone influences in order to know which bones have been weighted to vertices
				for vertIndex = 1 to selObj.numverts do
				(
					vertexBoneInfluencesCount = skinOps.getVertexWeightCount skinModifier vertIndex
					for vertBoneIndex = 1 to vertexBoneInfluencesCount do
					(
						boneIndex = skinOps.getVertexWeightBoneID skinModifier vertIndex vertBoneIndex
						boneVerts[boneIndex] = true
					)
				)

				-- Now we need to get the bone names for the unweighted bones
				numBones = skinOps.getNumberBones skinModifier
				for boneIndex = 1 to numBones do
				(
					if not boneVerts[boneIndex] then append bonesToRemove (skinOps.getBoneName skinModifier boneIndex 0)
					else append bonesToKeep (skinOps.getBoneName skinModifier boneIndex 0)
				)

				/* DEBUG
				--format "bonesToKeep:\n"
				--for boneName in bonesToKeep do format "%\n" boneName

				-- All bones
				for boneIndex = 1 to numBones do
				(
					boneName = skinOps.getBoneName skinModifier boneIndex 0
					--if findItem bonesToRemove boneName > 0 then
						format "boneIndex %=%\n" boneIndex boneName
				)
				*/

				-- Do the bone removal
				boneIndex = 1
				for i = 1 to numBones do
				(
					boneName = skinOps.getBoneName skinModifier boneIndex 0
					if findItem bonesToRemove boneName > 0 then
					(
						skinOps.selectBone skinModifier boneIndex
						skinOps.removeBone skinModifier
						--skinOps.removeBone skinModifier boneIndex  <-- (GKS) This does not work. Maybe because the bone list has become dirty. Using this the first bone is deleted by not of the subsequent ones.
					)
					else boneIndex += 1
				)
			)
		)
	)


	-- Used by the CLOTHING Tool
	--	TODO: Re-evaluate the need to 'If we are acting on a hierarchy then we need to process the whole thing from root most selected object down'
	--		One valid check would be if two or more objects are in a descendants <-> ancestors relationship
	--	TODO: Improvement: Check for transforms on parent objects. This will produce unpredictable results when importing into Studio.
	-- Parameters: None
	-- Return value: None
	fn bindClothing =
	(
		local PRINT_DEBUG = false
		local QUERYBOX_TITLE = "Bind Clothing"

		local userMessage
		local rootNodes, obj
		local objArray = selection as array
		local morphTargetsArray = #()
		local tempArray = #()
		local skinObjects = #()
		local skinObjectsModifiers = #() 
		local skip = false
		local haltBind = false
		local modifier
		local morphModifierCount
		local skinModifierCount
		local skinModifierIndex
		local exportLayer = LayerManager.getLayerFromName clothingTool.exportLayerName
		local skinModifier
		local linkSuccess
		local aBone

		-- First validate the selection list.
		if objArray.count == 0 then
		(
			messageBox "Nothing Selected!\nSelect one or more clothing meshes to bind to and try again." title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
			haltBind = true
		)

		-- More than one object selected: Not ideal but allowed. Query user and suggest user collapse hierarchy. Allow them to halt bind.
		if objArray.count > 1 then
			haltBind = queryBox "More than one object selected!\nIt is recommended that you combine your meshes before binding your clothing.\n\nDo you want to stop and combine your meshes?" title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)

		if not haltBind then
		(
			-- Get root nodes - This will cull child objects cherry picked from a hierarchy and we want to be acting on the root most nodes
			-- DEV CRUFT?: Not sure we really want to do this. The user may want to cherry pick what they are binding?
			--	Maybe do a check for descendants <-> ancestors relationships, and warn them that this is potentially problematic, but allow them to continue?
			--rootNodes = imvuTools.getRootNodes objArray

			-- Collect all of the objects we want to potentially perform the bind on
			-- DEV CRUFT?: Not sure we really want to do this. The user may want to cherry pick what they are binding?
			--for node in rootNodes do
			--(
			--	-- If we are acting on a hierarchy then we need to process the whole thing from root most selected object down
			--	-- DEV TODO: Maybe alert the user and give them the option to process the whole hierarchy?
			--	if node.children.count > 0 then imvuTools.getHierarchy node &objArray
			--	else insertItem node objArray 1
			--)

			if PRINT_DEBUG then
			(
				format "objArray BEFORE:\n" -- DEV_TEMP
				for obj in objArray do format "\t%\n" obj.name -- DEV_TEMP
			)

			-- CULLING ROUND 1
			-- Now collect the nodes we want don't want to perform the bind on
			for obj in objArray while not haltBind do
			(
				if PRINT_DEBUG then format "----- Checking: %\n" obj.name -- DEV_TEMP

				skip = false
				-- While 3ds Max allows you apply a Skin modifier to non-GeometryClass objects, only GeometryClass objects will be able to be imported into Studio. We are going to skip them.
				if superclassOf obj != GeometryClass then
				(
					skip = true
					append tempArray obj
					userMessage = "The object \"" + obj.name + "\" is not a mesh object and will be skipped!"
					messageBox userMessage title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
					imvuTools.printMsg #(userMessage) msgType:WARN addSpaces:false
				)

				-- Are there potential conditions where adding a Skin modifier will fail? I have not found them yet, but we should probably do the check just in case and halt if the condition is found.
				if not validModifier obj Skin and not skip then
				(
					skip = true
					append tempArray obj
					userMessage = "Object \"" + obj.name + "\" is not a valid object to add a skin modifier to. Skipping it."
					messageBox userMessage title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
					imvuTools.printMsg #(userMessage) msgType:WARN addSpaces:false
				)

				-- Check morph animated objects and add any target objects to the cull list. Morph targets should not be bound or exported.
				if not skip then
				(
					morphModifierCount = 0
					for ii = 1 to obj.modifiers.count do
					(
						-- Once we find the object has at least one modifier, validate the "morph object"
						if classof obj.modifiers[ii] == Morpher then
						(
							morphModifierCount += 1

							-- Collect morph targets and add them to the cull list
							join morphTargetsArray (imvuTools.getMorphTargetsFromMorphModifier obj.modifiers[ii])
						)
					)

					if morphModifierCount > 0 then
					(
						if PRINT_DEBUG then format "-- Validate Morpher --\n" -- DEV_TEMP

						-- Validate the morpher object
						haltBind = not imvuTools.isValidMorphObject obj

						if PRINT_DEBUG then format "imvuTools.isValidMorpher haltBind:%\n" haltBind -- DEV_TEMP

						if not haltBind and not imvuTools.isZeroTransformMatrix obj.transform then
						(
							-- If the object has been transformed and has a morpher modifier it will be very risky to try and preserve and re-apply or rebuild
							--	the morpher after the reset transform has been done. Therefore we are just going to warn the user about the situation.
							--	If they choose to continue they will be warned again when the reset transform is performed.
							userMessage = "Object \"" + obj.name + "\" has a Morpher modifier and has also been transformed (moved, rotated or scaled)!\n"
							userMessage = userMessage + "The morpher modifier will be destroyed when the object has its transforms reset."
							haltBind = not queryBox (userMessage + "\n\nDo you wish to continue?") title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
							imvuTools.printMsg #(userMessage) msgType:#WARN addSpaces:false
						)
					)
				)
			)

			if PRINT_DEBUG then
			(
				format "\ntempArray - ROUND 1:\n" -- DEV_TEMP
				for obj in tempArray do format "\t%\n" obj.name -- DEV_TEMP
			)

			if not haltBind then
			(
				skip = false

				-- Remove the objects we don't want to perform the bind on
				imvuTools.removeFromArray &objArray tempArray

				if PRINT_DEBUG then
				(
					format "\nobjArray AFTER:\n" -- DEV_TEMP
					for obj in objArray do format "\t%\n" obj.name -- DEV_TEMP
				)

				-- Get an intersection of objArray and morph targets. We are going to notify the user that
				--	they have selected morph targets to bind  
				tempArray = #()

				-- Alert the user to any morph targets that might be in their selection [less the culled objects]
				tempArray = for obj in morphTargetsArray where findItem objArray obj > 0 collect obj
				tempArray = makeUniqueArray tempArray
				if tempArray.count > 0 then
				(
					userMessage = "Object(s):\n"
					for obj in tempArray do userMessage = userMessage + obj.name + "\n"
					userMessage = userMessage + "are also morph targets and should not usually be bound to a skeleton and do not need to be exported."
					skip = queryBox (userMessage + "\n\nDo you wish to skip binding these objects?") title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
					imvuTools.printMsg #(userMessage + "\nskip bind: " + skip as string) msgType:#WARN addSpaces:false
					if skip then imvuTools.removeFromArray &objArray tempArray
				)

				-- Reset the temp list for another round of culling
				tempArray = #()

				-- CULLING ROUND 2
				-- Now do some checks on whether we can add a Skin modifier to the culled set of objects
				for obj in objArray while not haltBind do
				(
					morphModifierCount = 0
					skinModifierCount = 0
					skinModifierIndex = 0
					for ii = 1 to obj.modifiers.count do
					(
						-- Do some counting of Morpher modifiers
						if classof obj.modifiers[ii] == Morpher then morphModifierCount += 1

						-- Do some counting of Skin modifiers
						if classof obj.modifiers[ii] == Skin then
						(
							skinModifierCount += 1
							-- Get the index of the first Skin modifier we find.
							if skinModifierIndex == 0 then skinModifierIndex = ii
						)
					)

					if obj.modifiers.count > (morphModifierCount + skinModifierCount) then
					(
						-- The presence of certain modifiers other than Morpher and/or Skin will cause Max to crash when the Skin modifier is added.
						append tempArray obj
						userMessage = "Object \"" + obj.name + "\" has modifiers which may conflict with adding a Skin modifier and will be skipped!\n"
						for modifier in obj.modifiers do userMessage += ("\n" + modifier.name)
						haltBind = not queryBox (userMessage + "\n\nDo you wish to continue?") title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
						imvuTools.printMsg #(userMessage) msgType:#WARN addSpaces:false
					)
					else
					(
						if skinModifierCount > 1 then
						(
							-- Presence of multiple Skin modifiers -- THIS IS BAD
							append tempArray obj
							userMessage = "Object \"" + obj.name + "\" has multiple Skin modifiers which is a problem and will be skipped!"
							haltBind = not queryBox (userMessage + "\n\nDo you wish to continue?") title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
							imvuTools.printMsg #(userMessage) msgType:#WARN addSpaces:false
						)
						else
						(
							if skinModifierIndex == 1 then
							(
								-- If the Skin modifier is in the, correct, top most modifier position (1), query user asking if they would like to reapply the bind or halt.
								userMessage = "Object \"" + obj.name + "\" has a skin modifier.\n\nDo you want to delete the modifier and re-bind it?"
								skip = not queryBox userMessage title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
								imvuTools.printMsg #(userMessage + "\nskip:" + (skip as string)) msgType:#INFO addSpaces:false

								if not skip then
								(
									append skinObjects obj
									append skinObjectsModifiers obj.modifiers[skinModifierIndex]
									format "Delete modifier:% at index:% from object:%\n" obj.modifiers[skinModifierIndex].name skinModifierIndex obj.name -- DEV_TEMP
									deleteModifier obj skinModifierIndex
								)
								else append tempArray obj
							)
							else if skinModifierIndex > 1 then
							(
								-- Skin modifier is not in the correct position in the modifier stack
								append tempArray obj
								userMessage = "Object \"" + obj.name + "\" Skin modifier is not in the correct position in the modifier stack and will be skipped!"
								haltBind = not queryBox (userMessage + "\n\nDo you wish to continue?") title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE)
								imvuTools.printMsg #(userMessage) msgType:#WARN addSpaces:false
							)
						)
					)
				)
			)

			if PRINT_DEBUG then
			(
				format "\ntempArray - ROUND 2:\n"
				for obj in tempArray do format "\t%\n" obj.name -- DEV_TEMP
			)

			if not haltBind then
			(

				imvuTools.removeFromArray &objArray tempArray

				if PRINT_DEBUG then
				(
					format "\nobjArray AFTER:\n"
					for obj in objArray do format "\t%\n" obj.name -- DEV_TEMP
				)

				--haltBind = true -- DEV_DEBUG
				
				-- Get root nodes - This will cull child objects cherry picked from a hierarchy and we want to be acting on the root most nodes
				rootNodes = imvuTools.getRootNodes objArray

				--	Query reset transforms on the provided selection root nodes. This will alert the user to potential issues without actually doing the reset.  Halt if query failed.
				if not haltBind then haltBind = not imvuTools.resetTransforms suppliedObjArray:rootNodes mode:#QUERY
			)

			if haltBind then
			(
				-- If after going through all of the object checks we found any objects which had skin modifier deletes, add them back.
				-- Add back any top of stack skin modifiers
				for ii = 1 to skinObjects.count do
				(
					obj = skinObjects[ii]
					addModifier obj skinObjectsModifiers[ii]
				)
			)
			else
			(
				--	Reset transforms on the provided selection root nodes
				--	Object(s) with transforms will need to have them reset otherwise the mesh will be offset when imported into Studio. Halt if reset failed.
				haltBind = not imvuTools.resetTransforms suppliedObjArray:rootNodes mode:#FORCE
			)

			if not haltBind then
			(
				-- Finally perform the bind
				deselect objects
				max modify mode
				for obj in objArray do
				(
					linkSuccess = false

					--suspendEditing which:#modify
					select obj

					-- Add a skin modifier to each of the selected objects
					skinModifier = Skin name:MODIFIER_NAME_CLOTHINGSKIN
					addModifier obj skinModifier

					--resumeEditing which:#modify
					modPanel.setCurrentObject obj.modifiers[MODIFIER_NAME_CLOTHINGSKIN as name]
					
					if allAvatarExportBones == undefined then allAvatarExportBones = execute ("$" + MEGASCENE_NODE_NAME_AVATAR_SKELETON_ROOT + "...*") as array
					for aBone in allAvatarExportBones do
						skinOps.addBone skinModifier aBone -1
					
					skinModifier.showNoEnvelopes = on

					-- Get into skin modifier vert mode and select all of the vertices
					skinModifier.filter_vertices = on
					--skinOps.SelectVertices skinModifier #{1..(selObj.numverts)}
					skinOps.SelectVertices skinModifier #{1..(obj.numverts)}
					skinModifier.bone_Limit = BONES_PER_VERTEX_LIMIT

					-- Run the Weight Solver
					local MODE_INTEGER = 1 -- The <mode_integer> argument defines the Weight Solver to use, where 0 = Voxel, 1 = Heat Map
					local FALLOFF_FLOAT = 0.2 -- The <falloff_float> argument defines the Falloff, UI default is 0.2.
					-- BONES_PER_VERTEX_LIMIT -- The <max_influence_integer> argument defines the maximum number of bones that can influence a vertex. UI default is 0 which means no limit.
					local MAX_RES_INTEGER = 0 -- The <max_res_integer> argument defines the Maximum resolution of the voxel grid. The UI default is 256. Only used by the Voxel mode, but must be specified even when performing Heat Map Weighting.
					-- The <winding_bool> argument corresponds to the "Use Winding Numbers" checkbox in the UI which is unchecked by default. Only used by the Voxel mode, but must be specified even when performing Heat Map Weighting.
					-- The <envelope_off_bool> argument corresponds to the "Turn Off Envelope Gizmos" checkbox in the UI which is checked by default.
					skinOps.voxelWeighting skinModifier MODE_INTEGER  FALLOFF_FLOAT  BONES_PER_VERTEX_LIMIT  MAX_RES_INTEGER off on

					removeUnweightedBones skinModifier

					-- Add the object(s) to the export layer
					exportLayer.addNode obj

					linkSuccess = true

					if not linkSuccess then messageBox ("ERROR: Mesh " + obj.name + " failed to bind!") title:(TOOLNAME_CLOTHING + ": " + QUERYBOX_TITLE) icon:#critical
				)
			)
		)
	)


	-- Tool Function - Switch to Paint Weights mode
	-- ONLY Used by the CLOTHING Tool
	-- Parameters: None
	-- Return value: None
	fn paintWeightsMode =
	(
		local selObjArray = selection as array
		local selObj = selObjArray[1]
		local skinModifier

		if selObjArray.count == 0 then
			messageBox "Nothing Selected!\nSelect one mesh to paint weight and try again." title:TOOLNAME_CLOTHING
		else if selObjArray.count > 1 then
			messageBox "More than one object Selected!\nSelect one mesh to paint weight and try again." title:TOOLNAME_CLOTHING
		-- Is the object a valid object to paint weights?
		else if not validModifier selObj Skin then
			messageBox "No skinned object selected!\nSelect one clothing mesh to paint weight and try again." title:TOOLNAME_CLOTHING
		else
		(
			skinModifier = selObj.modifiers[MODIFIER_NAME_CLOTHINGSKIN as name]
			if skinModifier != undefined then
			(
				max modify mode
				--local disp = NitrousGraphicsManager.GetActiveViewportSetting()
				--disp.ViewportViewSettingImpl.VisualStyleMode = #Shaded
				viewport.SetRenderLevel #smooth
				subObjectLevel = 1

				-- Get into skin modifier vert mode and select all of the vertices
				skinModifier.filter_vertices = on
				skinOps.SelectVertices skinModifier #{1..(selObj.numverts)}

				--if skinOps.isWeightToolOpen skinModifier == 0 then skinOps.weightTool skinModifier
				skinOps.weightTool skinModifier
				skinOps.paintOptionsButton skinModifier
				skinOps.paintWeightsButton skinModifier
			)
			else
				messageBox "No clothing skinned object selected!\nSelect one clothing mesh to paint weight and try again." title:TOOLNAME_CLOTHING
		)
	)

			
	-- Tool Function - Limit the bones per vertex for all objects with CLOTHING skin modifiers
	-- ONLY Used by the CLOTHING Tool
	-- Parameters: None
	-- Return value: None
	fn limitBoneWeights =
	(
		local skinModifier
		
		for obj in objects do
		(
			if validModifier obj Skin then
				skinModifier = obj.modifiers[MODIFIER_NAME_CLOTHINGSKIN as name]
				if skinModifier != undefined then
					skinModifier.bone_Limit = BONES_PER_VERTEX_LIMIT
		)
	)


	-- Tool Function - Set up a Avatar animation
	-- Used by the AVATAR ANIMATION Tool
	-- Parameters: None
	-- Return value: None
	fn createAvatarAnimation =
	(
		mergeMaxFile MEGASCENE MEGASCENE_GROUPS
		ungroup (getNodeByName MEGASCENE_GROUPS[megaSceneGroup_AVATAR_RIG])

		visualsToolVars.setVisibleAvatar visAvatar_PROXY setUI:true

		setToolkitToolVariables()

		frameRate = ANIM_FRAMERATE

		-- Set all nodes to the Export Layer 
		addNodesToExportLayer allAvatarExportBones avatarAnimTool.exportLayerName
	)


	-- Tool Function - Aligns Avatar widgets and key necessary widgets based on the animation 'Auto Key' and 'Set Key' states in the main 3ds Max UI.
	-- ONLY Used by the AVATAR ANIMATION Tool
	-- Parameters:
	--		selectedWidgetPairIndex <integer> The widget pair group index which the object who's transform is changing belongs to
	--		widgetPairIndices <array> Array of widget pairs indices (i.e.: #(animAlign_IK2FKARMS_L, animAlign_IK2FKLEGS_L, etc.)) which should be acted on for alignment
	-- Return value: None
	fn alignAvatarWidgets selectedWidgetPairIndex widgetPairIndices =
	(
		local PRINT_DEBUG = false

		if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("selectedWidgetPairIndex:", selectedWidgetPairIndex) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:alignAvatarWidgets

		local pairIndex, pair
		local widgetIsKeyed = false

		-- Look for any "source" widgets that have been keyed. This will be at least one that the user has currently selected and is moving or rotating 
		for pair in avatarAnimToolVars.alignPairs[selectedWidgetPairIndex] where not widgetIsKeyed do
			widgetIsKeyed = isControllerKeyedAtTime pair[const_SOURCE].transform.controller

		for pairIndex in widgetPairIndices do
		(
			if PRINT_DEBUG or PRINT_DEBUG_ALL then imvuTools.printMsg #("pairIndex=", pairIndex) msgType:#DEBUG inScript:THIS_SCRIPT_FILENAME inFunction:alignAvatarWidgets

			for pair in avatarAnimToolVars.alignPairs[pairIndex] do
			(
				-- If the 'Set Key' animation button is active (maxOps.setKeyMode) and the widget being manipulated is keyed, temporarily suspend "Set Key" mode
				-- because 'with animate on' does not seem to work when in "Set Key" mode is active.
				-- Suspending it does not mess with the state of the Max UI whereas setting 'maxOps.setKeyMode' does.
				maxOps.setKeySuspend = maxOps.setKeyMode and widgetIsKeyed

				if PRINT_DEBUG or PRINT_DEBUG_ALL then format "pair[const_SOURCE]=% pair[const_TARGET]=% widgetIsKeyed=% SetKeyMode=% AutoAnim=% " pair[const_SOURCE].name pair[const_TARGET].name widgetIsKeyed maxOps.setKeyMode animButtonState

				-- If 'Auto' animation button is active (animButtonState)
				if animButtonState or maxOps.setKeySuspend then
				(
					if PRINT_DEBUG or PRINT_DEBUG_ALL then format "ANIMATE!"
					alignNodesToTarget pair[const_SOURCE] pair[const_TARGET] origin:pair[const_SOURCE] atFrames:#(sliderTime)
				)
				else alignNodesToTarget pair[const_SOURCE] pair[const_TARGET] origin:pair[const_SOURCE]

				if PRINT_DEBUG or PRINT_DEBUG_ALL then format "\n"

				maxOps.setKeySuspend = false
			)
		)
	)
	-- Store a pointer to the function in the tool var
	avatarAnimToolVars.alignWidgets = alignAvatarWidgets


	-- Tool Function -  Sets the kinematics mode for left and right arms and legs
	-- ONLY Used by the AVATAR ANIMATION Tool
	-- Parameters:
	--		cbCntrlCaption <string>
	--		ikEnabled <boolean>
	-- Return value: None
	fn setIkFkMode cbCntrlCaption ikEnabled =
	(
		--format "cbCntrlCaption=% ikEnabled=%\n" cbCntrlCaption ikEnabled -- DEBUG

		if avatarAnimToolVars.customAttributesNode == undefined or
			avatarAnimToolVars.ikLeftArmWidgets == undefined or
			avatarAnimToolVars.ikRightArmWidgets == undefined or
			avatarAnimToolVars.ikLeftLegWidgets == undefined or
			avatarAnimToolVars.ikRightLegWidgets == undefined or
			avatarAnimToolVars.fkLeftArmWidgets == undefined or
			avatarAnimToolVars.fkRightArmWidgets == undefined or
			avatarAnimToolVars.fkLeftLegWidgets == undefined or
			avatarAnimToolVars.fkRightLegWidgets == undefined then imvuTools.printMsg #("Control or widgets undefined!") msgType:WARN inScript:THIS_SCRIPT_FILENAME inFunction:setIkFkMode
		else
		(
			case cbCntrlCaption of
			(
				ANIM_IK_ARMS_CAPTION:
				(
					-- Set IK/FK control rig weighting for arms
					with animate off avatarAnimToolVars.customAttributesNode.Custom_Attributes.lfArmIK = if ikEnabled then 100 else 0
					with animate off avatarAnimToolVars.customAttributesNode.Custom_Attributes.rtArmIK = if ikEnabled then 100 else 0

					-- Set the visibility of the IK/FK rig arm widgets
					for widget in avatarAnimToolVars.ikLeftArmWidgets do widget.isNodeHidden = not ikEnabled
					for widget in avatarAnimToolVars.ikRightArmWidgets do widget.isNodeHidden = not ikEnabled
					for widget in avatarAnimToolVars.fkLeftArmWidgets do widget.isNodeHidden = ikEnabled
					for widget in avatarAnimToolVars.fkRightArmWidgets do widget.isNodeHidden = ikEnabled

					-- Remove and establish/re-establish widget transform change handlers i.e.: 3ds Max callbacks when the widget is transformed in the 3d scene
					--	Used to keep IK and FK bones aligned at all times.
					deleteAllChangeHandlers id:#ChangeHandler_AvatarAnimWidgetArms

					if ikEnabled then
					(
						when transform avatarAnimToolVars.alignPairs[animAlign_IK2FKARMS_L][1][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetArms handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_IK2FKARMS_L #(animAlign_FK2IKARMS_L)
						when transform avatarAnimToolVars.alignPairs[animAlign_IK2FKARMS_R][1][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetArms handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_IK2FKARMS_R #(animAlign_FK2IKARMS_R)
					)
					else
					(
						when transform avatarAnimToolVars.alignPairs[animAlign_FK2IKARMS_L][3][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetArms handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_FK2IKARMS_L #(animAlign_IK2FKARMS_L)
						when transform avatarAnimToolVars.alignPairs[animAlign_FK2IKARMS_R][3][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetArms handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_FK2IKARMS_R #(animAlign_IK2FKARMS_R)
					)

				)
				ANIM_IK_LEGS_CAPTION:
				(
					-- Set IK/FK control rig weighting for legs
					with animate off avatarAnimToolVars.customAttributesNode.Custom_Attributes.lfLegIK = if ikEnabled then 100 else 0
					with animate off avatarAnimToolVars.customAttributesNode.Custom_Attributes.rtLegIK = if ikEnabled then 100 else 0

					-- Set the visibility of the IK/FK rig leg widgets
					for widget in avatarAnimToolVars.ikLeftLegWidgets do widget.isNodeHidden = not ikEnabled
					for widget in avatarAnimToolVars.ikRightLegWidgets do widget.isNodeHidden = not ikEnabled
					for widget in avatarAnimToolVars.fkLeftLegWidgets do widget.isNodeHidden = ikEnabled
					for widget in avatarAnimToolVars.fkRightLegWidgets do widget.isNodeHidden = ikEnabled

					-- Remove and establish/re-establish widget transform change handlers i.e.: 3ds Max callbacks when the widget is transformed in the 3d scene
					--	Used to keep IK and FK bones aligned at all times.
					deleteAllChangeHandlers id:#ChangeHandler_AvatarAnimWidgetLegs

					if ikEnabled then
					(
						when transform avatarAnimToolVars.alignPairs[animAlign_IK2FKLEGS_L][1][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetLegs handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_IK2FKLEGS_L #(animAlign_FK2IKLEGS_L)
						when transform avatarAnimToolVars.alignPairs[animAlign_IK2FKLEGS_R][1][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetLegs handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_IK2FKLEGS_R #(animAlign_FK2IKLEGS_R)
					)
					else
					(
						when transform avatarAnimToolVars.alignPairs[animAlign_FK2IKLEGS_L][4][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetLegs handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_FK2IKLEGS_L #(animAlign_IK2FKLEGS_L)
						when transform avatarAnimToolVars.alignPairs[animAlign_FK2IKLEGS_R][4][const_SOURCE] changes id:#ChangeHandler_AvatarAnimWidgetLegs handleAt:#redrawViews do
							avatarAnimToolVars.alignWidgets animAlign_FK2IKLEGS_R #(animAlign_IK2FKLEGS_R)
					)
				)
			)
		)
	)
	-- Store a pointer to the function in the tool var
	avatarAnimToolVars.setKinematicsMode = setIkFkMode


	-- TODO: This is not yet implemented
	/*
	-- Tool Function -  Sets the rotation isolation mode for left and right arms and legs
	-- ONLY Used by the AVATAR ANIMATION Tool
	-- Parameters:
	--		cbCntrlName <string>
	--		isEnabled <boolean>
	-- Return value: None
	fn setRotationIsolationMode cbCntrlName isEnabled =
	(
		format "ImvuStudioToolkit::setRotationIsolationMode() control=% isEnabled=%\n" cbCntrlName isEnabled
		case cbCntrlName of
		(
			"lArmRotIso": TBD
			"rArmRotIso": TBD
			"lLegRotIso": TBD
			"rLegRotIso": TBD
		)
	)
	*/


	-- Tool Function - Deletes keys on selected avatar rig widgets at timeslider frames if provided. I no frames are provided then all keys are deleted.
	-- ONLY Used by the AVATAR ANIMATION Tool
	-- Parameters:
	--		frames <array> [optional] Array of keyframes to act on. Defaults to empty array which results in acting on all keys
	-- Return value: None
	fn deleteAvatarKeys atFrames:#() =
	(
		local selObjArray = selection as array
		local selectedWidgets = imvuTools.intersectionOfArrays #(selObjArray, avatarAnimToolVars.allWidgets)

		if selObjArray.count > 0 and selectedWidgets.count == 0 then messageBox ("No widgets in your selection!\nPlease select one or more avatar rig widget and try again.") title:TOOLKIT_NAME icon:#warning 
		else
		(
			if selectedWidgets.count == 0 then for widget in avatarAnimToolVars.allWidgets do deleteAllKeysOnControllerAtTime widget.transform.controller atFrames:atFrames
			else for widget in selectedWidgets do deleteAllKeysOnControllerAtTime widget.transform.controller atFrames:atFrames
		)
	)



	-- Tool Function -  Sets the rotation and position transforms for the selected avatar rig widgets (or all widgets if none are selected) back to zero
	-- ONLY Used by the AVATAR ANIMATION Tool and adapted from the 3ds Max supplied macroscript Macro_FreezeTransform.mcr
	-- Parameters: None
	-- Return value: None
	fn rigWidgetTransformsToZero =
	(
		local selObjArray = selection as array
		local widgets = #()
		local widget
		local activeController

		if selObjArray.count > 0 then
			for selObj in selObjArray do
				if matchPattern selObj.name pattern:(MEGASCENE_NODE_NAME_PREFIX_RIG_WIDGETS + "*") then append widgets selObj

		if widgets.count == 0 then widgets = avatarAnimToolVars.allWidgets

		if widgets.count > 0 then
		(
			suspendEditing
			for widget in widgets do
			(
				Try
				(
					if classof widget.Position.controller == Position_Layer then
					(
						activeController = widget.position.controller.active
						widget.position.controller[activeController].value = (Point3 0 0 0)
					)
					else
					(
						widget.position.controller[2].x_position = 0
						widget.position.controller[2].y_position = 0
						widget.position.controller[2].z_position = 0
					)

					if classof widget.rotation.controller == Rotation_Layer then
					(
						activeController = widget.rotation.controller.active
					  	widget.rotation.controller[activeController].value = (quat 0 0 0 1)
					)
					else
					(
						widget.rotation.controller[2].x_rotation = 0
						widget.rotation.controller[2].y_rotation = 0
						widget.rotation.controller[2].z_rotation = 0
					)
				)
				Catch( messageBox "One of the object's transform was never frozen." title:"Toolkit - Transforms to Zero")
			)
			resumeEditing
		)
	)


	-- Tool Function - Optionally copy positions and rotations of all Control_ bones to all export skeleton bones and
	--	key all export skeleton bones at optionally supplied keyframe.
	-- ONLY Used by the EXPORT Tool
	-- Parameters:
	--		bones <array> List of bones to be keyed.
	--		atFrames <array> Defaults to all frames in the time slider. Not usually what you want, but it is here as an option.
	--		keyReduction <boolean> Defaults to 'false'. If 'true' perform a key reduction pass after setting keys.
	--			Used for primarily for removing bones with keys that are all the same. There is a use case for keying all bones however
	-- Return value: None
	fn keyAvatarExportBones bones:#() atFrames:(for ii = 0 to animationRange.end.frame collect ii) keyReduction:false  =
	(
		local aBone, controlBone
		local KEY_REDUCTION_THRESHOLD = 0.1
		local TIME_STEP = 1f -- every frame

		-- sort the frames in case the frame array passed was unsorted
		sort atFrames

		-- Turn off all animation modes that may be active
		disableAnimationModes()

		for aBone in bones do
		(
			controlBone = getNodeByName (MEGASCENE_NODE_NAME_PREFIX_CONTROL_BONE + aBone.name)
			if controlBone == undefined then  imvuTools.printMsg #("Failed to find control bone!:", (MEGASCENE_NODE_NAME_PREFIX_CONTROL_BONE + aBone.name)) msgType:#WARN inScript:THIS_SCRIPT_FILENAME inFunction:keyAvatarExportBones
			else alignNodesToTarget aBone controlBone origin:aBone atFrames:atFrames
		)

		-- NOTE: reduceKeys is a built-in maxscript function which acts on controller keys
		--	It has the undesirable side effect of showing a progress bar and cancel button and ends up closing Studio Toolkit
		-- TODO: Write my own function (see reduceKeysOnController) to do the same but limit it to position and rotation controllers
		if keyReduction then reduceKeys bones KEY_REDUCTION_THRESHOLD TIME_STEP
	)


	-- Tool Function - Set up avatar animation export
	-- ONLY Used by the EXPORT Tool. On export:
	-- Parameters: None
	-- Return value: None
	fn setupAvatarAnimForExport =
	(
		local aBone, widget, widgetKey, widgetFrame
		local nodeFrames = #()
		local deltaPos, deltaRot

		-- Reset bone positions and rotations.
		-- Key all export skeleton bones at frames 0 (this should be the t-pose)
		for aBone in allAvatarExportBones do
		(
			-- Allow for custom, user generated skeletons by skipping all non-base IMVU skeleton bones
			if hasINISetting MEGASCENE_INIFILE aBone.name then
			(
				-- Clear all keys on export skeleton.
				deleteAllKeysOnControllerAtTime aBone.transform.controller

				aBone.rotation = execute (getINISetting MEGASCENE_INIFILE aBone.name "Rotation")
				aBone.position = execute (getINISetting MEGASCENE_INIFILE aBone.name "Position")
			)

			addNewKey aBone.rotation.controller 0 #interpolate
			addNewKey aBone.position.controller 0 #interpolate
		)

		--	If 'makePose' then copy positions and rotations of all Control_ bones to all export skeleton bones and key them at frames 1 and 3
		if avatarAnimTool.toolRollout.makePose.checked then
			keyAvatarExportBones bones:allAvatarExportBones atFrames:#(1,3)
		else
		(
			-- Loop through all of the widgets and collect keyframes.
			--		We are doing this on position and rotation separately instead of transform.controller because we want to ignore scale
			for widget in avatarAnimToolVars.allWidgets do
			(
				collectControllerKeyframes widget.position.controller &nodeFrames
				collectControllerKeyframes widget.rotation.controller &nodeFrames
			)

			if avatarAnimTool.toolRollout.exportAllBones.checked then
				keyAvatarExportBones bones:allAvatarExportBones atFrames:nodeFrames
			else
				keyAvatarExportBones bones:allAvatarExportBones atFrames:nodeFrames keyReduction:true
		)
	)


	-- Tool Function - Perform file export
	--	An interesting tool I stumbled upon while looking into splitting takes and naming them: https://github.com/gildor2/UEViewer/tree/master/Tools/MaxActorXImport
	-- ONLY Used by the EXPORT Tool
	-- Parameters: None
	-- Return value: None
	fn doExport =
	(
		local newFileName
		local exportFilePath = ""
		local selObjArray = selection as array
		local obj, modifier
		local hasModifier = false
		local curLayerNodes = #()
		local haltExport =  false
		local exportFileNamePrefix = ""
		local exportFileNameSuffix = ""
		local exportMsg = ""
		local exportLayersFound = false
		local animLength = getAnimLength()
		local exportLayer
		local exportLayerNodes = #()
		
		if maxFileName == "" then
			messageBox "Scene not saved! Please save your project before exporting" title:TOOLNAME_EXPORT
		else
		(
			for layerName in exportLayerNames do
			(
				if (LayerManager.getLayerFromName layerName) != undefined then
				(
					-- Get objects to be exported from the export layer
					(LayerManager.getLayerFromName layerName).nodes &curLayerNodes

					-- Perform some validation on what we are about to export.
					haltExport = not imvuTools.validateObjects curLayerNodes

					if not haltExport then
					(
						-- Select the objects in the Export layer
						clearSelection()
						select curLayerNodes
						
						fbxExporterSetParam "ASCII" true
						fbxExporterSetParam "ScaleFactor" EXPORT_SCALE_FACTOR
						fbxExporterSetParam "UpAxis" "Z" --"Y" This seems to have no effect on the resulting XSFs or XAFs

						-- Animation params we always want
						fbxExporterSetParam "Animation" true
						fbxExporterSetParam "BakeAnimation" false
						fbxExporterSetParam "UseSceneName" false
						fbxExporterSetParam "SplitAnimationIntoTakes" "-clear"

						fbxExporterSetParam "Cameras" false
						fbxExporterSetParam "Skin" true
						fbxExporterSetParam "Shape" true -- for morph animation
						fbxExporterSetParam "GeomAsBone" false
						
						case layerName of
						(
							EXPORT_LAYER_NAME_ROOM:
							(
								local exportSeats = #()
								getNodesByNodeTypeInArray nodeType_SEAT objects &exportSeats
								if exportSeats.count < 6 then
									messageBox "Not enough seat nodes! You need at least 6" title:TOOLNAME_EXPORT 
								else
								(
									exportFileNamePrefix = "Room_"
									
									fbxExporterSetParam "Cameras" true
								)
							)
							EXPORT_LAYER_NAME_FURNITURE:
							(
								exportLayer = LayerManager.getLayerFromName EXPORT_LAYER_NAME_FURNITURE
								exportLayerNodes = #()

								exportLayer.nodes &exportLayerNodes
								animLength = getAnimLength nodes:exportLayerNodes

								exportFileNamePrefix = "Furniture_"

								if animLength > 1 then
								(
									exportFileNameSuffix = ("_" + rootNode.toolkitData.furnitureAnimName)
									fbxExporterSetParam "SplitAnimationIntoTakes" rootNode.toolkitData.furnitureAnimName 0 animLength
								)
							)
							EXPORT_LAYER_NAME_ATTACHMENT:
							(
								exportFileNamePrefix = "Attachment_"
							)
							EXPORT_LAYER_NAME_CLOTHING:
							(
								exportFileNamePrefix = "Clothing_"
								
								--fbxExporterSetParam "Animation" false
							)
							EXPORT_LAYER_NAME_AVATAR_ANIMATION:
							(
								exportFileNamePrefix = "AviAnim_"
								
								exportLayer = LayerManager.getLayerFromName EXPORT_LAYER_NAME_AVATAR_ANIMATION
								exportLayerNodes = #()

								exportLayer.nodes &exportLayerNodes
								animLength = getAnimLength nodes:exportLayerNodes

								if animLength > 1 then
								(
									exportFileNameSuffix = ("_" + rootNode.toolkitData.avatarAnimName)
									fbxExporterSetParam "SplitAnimationIntoTakes" rootNode.toolkitData.avatarAnimName 0 animLength
								)
								
								-- fbxExporterSetParam "GeomAsBone" true
								fbxExporterSetParam "Removesinglekeys" true
								-- fbxExporterSetParam "Resampling" 1

								setupAvatarAnimForExport()
							)
						)
						
						newFileName = exportFileNamePrefix + (filterString maxFileName ".")[1] + exportFileNameSuffix + ".FBX"
						exportFilePath = (substituteString maxFilePath "\\" "/") + newFileName
						exportFile exportFilePath #noPrompt using:FBXEXP selectedOnly:true
						
						exportLayersFound = true
						exportMsg += "File exported: " +  newFileName + "\n"
					)
				)
			)
			if haltExport then
			(
				messageBox ("Export failed!\n") title:TOOLNAME_EXPORT icon:#critical
				imvuTools.printMsg #("Export failed!") msgType:#ERROR
			)
			else
			(
				if not exportLayersFound then
					exportMsg = "Nothing Exported! No export layers found!"
				else
				(
					messageBox ("Export complete!\n" + exportMsg) title:TOOLNAME_EXPORT icon:#information
					imvuTools.printMsg #("File exported:", exportFilePath) msgType:#INFO
				)
			)
		)
	)


	-- BEGIN - TOOL UI - Rollouts  --------------------------------------------------------------------------------------------------------------------------------------------------------------------
	
	-- MAX HELPER TOOLS Rollout
	rollout visualsRollout "Visuals" rolledup:false 
	(
		-- UI definition
		group ""
		(
			radiobuttons displayAvatar labels:VIS_AVATAR_LABELS offsets:#([0, 0], [5, 0], [10, 0], [15, 0]) align:#center enabled:false
			button showNames "Show Bone Names" width:110 toolTip:TOOLTIP_VISUALS_SHOWNAMES_BUTTON align:#right across:2 
			button hideNames "Hide Bone Names" width:110 toolTip:TOOLTIP_VISUALS_HIDENAMES_BUTTON align:#left
			button freezeAvatars "Freeze Avatars" width:110 toolTip:TOOLTIP_VISUALS_FREEZEAVATARS_BUTTON align:#right across:2 
			button unfreezeAvatars "Unfreeze Avatars" width:110 toolTip:TOOLTIP_VISUALS_UNFREEZEAVATARS_BUTTON align:#left
			label spacer01 "" height:1 width:0
		)
		
		-- UI controls
		on displayAvatar changed val do
		(
			visualsToolVars.setVisibleAvatar VIS_AVATAR_LABEL_ORDER[val]
		)

		-- TODO: FIX THIS: This is not the right way to handle hiding/showing bone names. Can get 3ds Max in a bad state after closing Toolkit with names displayed
		on showNames pressed do
		(
			unregisterRedrawViewsCallback gwDisplayObjectNames
			registerRedrawViewsCallback gwDisplayObjectNames
		)
		
		on hideNames pressed do
		(
			unregisterRedrawViewsCallback gwDisplayObjectNames
		)

		on freezeAvatars pressed do visualsToolVars.setFreezeAvatars doFreeze:true

		on unfreezeAvatars pressed do visualsToolVars.setFreezeAvatars doFreeze:false
	)

	-- ROOM TOOL Rollout
	rollout roomRollout "Room Tool" rolledup:true 
	(
		-- UI definition
		group ""
		(
			spinner XSlider "X" range:[0, MAX_ROOM_REF_SIZE, DEFAULT_ROOM_REF_SIZE] width:80 align:#left enabled:false across:3
			spinner YSlider "Y" range:[0, MAX_ROOM_REF_SIZE, DEFAULT_ROOM_REF_SIZE] width:80 align:#center enabled:false
			spinner ZSlider "Z" range:[0, MAX_ROOM_REF_SIZE, DEFAULT_ROOM_REF_SIZE] width:80 align:#right enabled:false

			button createRoomButton "Create Room" width:220 toolTip:TOOLTIP_ROOM_TOOL_CREATE_BUTTON	enabled:true
			label spacer01 "" width:0 height:1
		)
		
		group ""
		(
			radiobuttons furnitureType labels:#("Floor      ","Wall        ","Ceiling") enabled:false
			radiobuttons furnitureDir labels:#("North   ","South   ","East   ","West") enabled:false
			button addFurnitureButton "+  Furniture" width:220 toolTip:TOOLTIP_ROOM_TOOL_ADDFURNITURE_BUTTON enabled:false
		)
		
		group ""
		(
			button roomN "Up" width:50 toolTip:TOOLTIP_ROOM_TOOL_ROOMN_BUTTON enabled:false
			button roomW "Left" width:50 toolTip:TOOLTIP_ROOM_TOOL_ROOMW_BUTTON enabled:false align:#right across:3
			spinner furnDist "" width:50 toolTip:TOOLTIP_ROOM_TOOL_FURNDIST_BUTTON enabled:false range:[0, 100, 3] align:#center
			button roomE "Right" width:50 toolTip:TOOLTIP_ROOM_TOOL_ROOME_BUTTON enabled:false align:#left
			button roomS "Down" width:50 toolTip:TOOLTIP_ROOM_TOOL_ROOMS_BUTTON enabled:false
		)

		group ""
		(
			button addIdleButton "+  Idle" across:1 width:220 align:#center toolTip:TOOLTIP_ROOM_TOOL_ADDIDLE_BUTTON enabled:false
			button addStandingButton "+  Standing" across:1 width:220 align:#center toolTip:TOOLTIP_ROOM_TOOL_ADDSTANDING_BUTTON enabled:false
			button addSittingButton "+  Sitting" across:1 width:220 align:#center toolTip:TOOLTIP_ROOM_TOOL_ADDSITTING_BUTTON enabled:false
			
		)
		
		group ""
		(
			edittext customSeat "" width:220 align:#center toolTip:TOOLTIP_ROOM_TOOL_CUSTOMSEAT enabled:false
			button addCustomButton "+  Custom" width:220 align:#center toolTip:TOOLTIP_ROOM_TOOL_ADDCUSTOMSEAT_BUTTON enabled:false
		)
		
		group ""
		(
			button addCatcher "+  Catcher" width:110 toolTip:TOOLTIP_ROOM_TOOL_ADDCATCHER_BUTTON enabled:false align:#right across:2 
			button addHandle "+  Handle" width:110 toolTip:TOOLTIP_ROOM_TOOL_ADDHANDLE_BUTTON enabled:false align:#left
			button removeCatcher "-  Catcher" width:110 toolTip:TOOLTIP_ROOM_TOOL_REMOVECATCHER_BUTTON enabled:false align:#right across:2 
			button removeHandle "-  Handle" width:110 toolTip:TOOLTIP_ROOM_TOOL_REMOVEHANDLE_BUTTON enabled:false align:#left
		)
		
		group ""
		(
			button createRoomST "Create See Through Node" width:220 align:#center enabled:false
			button linkRoomST "Link To SeeThrough" width:220 align:#center enabled:false
			button unlinkRoomST "Unlink From SeeThrough" width:220 align:#center enabled:false
		)
		
		-- ROOM TOOL Rollout actions
		on createRoomButton pressed do --Create Room
		(
			if not canToolBeActiveOrEnabled roomTool then
				messageBox ("ERROR: Room cannot be created because there is another tool already active!") title:TOOLNAME_ROOM
			else
			(
				createRoom [XSlider.value * ROOM_REF_SCALE_FACTOR, YSlider.value * ROOM_REF_SCALE_FACTOR, ZSlider.value * ROOM_REF_SCALE_FACTOR]
				max zoomext sel
				roomTool.setState toolState_ACTIVE
				updateToolkitSceneAndUIState()
			)
		)

		-- Room controls
		on XSlider changed val do
			resizeRoom [XSlider.value * ROOM_REF_SCALE_FACTOR, YSlider.value * ROOM_REF_SCALE_FACTOR, ZSlider.value * ROOM_REF_SCALE_FACTOR]
		on YSlider changed val do
			resizeRoom [XSlider.value * ROOM_REF_SCALE_FACTOR, YSlider.value * ROOM_REF_SCALE_FACTOR, ZSlider.value * ROOM_REF_SCALE_FACTOR]
		on ZSlider changed val do
			resizeRoom [XSlider.value * ROOM_REF_SCALE_FACTOR, YSlider.value * ROOM_REF_SCALE_FACTOR, ZSlider.value * ROOM_REF_SCALE_FACTOR]
		
		-- Furniture controls
		on furnitureType changed state do
			if furnitureType.state == 2 then
				furnitureDir.enabled = true
			else
				furnitureDir.enabled = false
		
		--Add Furniture
		on addFurnitureButton pressed do addRoomFurnitureNode  furnitureType.state furnitureDir.state

		on roomN pressed do --Duplicate Furniture North
			duplicateFurnitureNodes "N" furnDist.value
		on roomS pressed do --Duplicate Furniture South
			duplicateFurnitureNodes "S" furnDist.value
		on roomE pressed do --Duplicate Furniture East
			duplicateFurnitureNodes "E" furnDist.value
		on roomW pressed do --Duplicate Furniture West
			duplicateFurnitureNodes "W" furnDist.value

		-- Seat controls
		on addSittingButton pressed do --Add Sitting
			addSeat NODE_NAME_SEAT_SITTING roomTool
		on addStandingButton pressed do --Add Standing
			addSeat NODE_NAME_SEAT_STANDING roomTool
		on addIdleButton pressed do --Add Idle
			addSeat NODE_NAME_SEAT_IDLE roomTool
		on addCustomButton pressed do --Add Custom
			addSeat customSeat.text roomTool
		
		-- Seat Catch/Pitcher and Handle controls
		on addCatcher pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_CATCHER_PITCHER roomTool
		on addHandle pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_HANDLE roomTool
		on removeCatcher pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_CATCHER_PITCHER roomTool addNewNode:false
		on removeHandle pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_HANDLE roomTool addNewNode:false
			
		-- See through node controls
		on createRoomST pressed do
			createSeeThroughNode roomTool
		on linkRoomST pressed do
			linkSeeThroughNode()
		on unlinkRoomST pressed do
			unlinkSeeThroughNode()
	) -- END ROOM TOOL Rollout


	-- FURNITURE TOOL Rollout
	rollout furnitureRollout "Furniture Tool" rolledup:true 
	(
		-- UI definition
		group ""
		(
			button createFurnitureButton "Create Furniture" width:220 toolTip:TOOLTIP_FURNITURE_TOOL_CREATE_BUTTON
			label spacer01 "" height:1 width:0
		)
		
		group ""
		(
			button F_addIdleButton "+  Idle" across:1 width:220 align:#center toolTip:TOOLTIP_FURNITURE_TOOL_ADDIDLE_BUTTON enabled: false
			button F_addStandingButton "+  Standing" across:1 width:220 align:#center toolTip:TOOLTIP_FURNITURE_TOOL_ADDSTANDING_BUTTON enabled: false
			button F_addSittingButton "+  Sitting" across:1 width:220 align:#center toolTip:TOOLTIP_FURNITURE_TOOL_ADDSITTING_BUTTON enabled: false
		)
		
		group ""
		(
			edittext F_customText "" width:220 align:#center toolTip:TOOLTIP_FURNITURE_TOOL_CUSTOMSEAT enabled: false
			button F_addCustomButton "+  Custom" across:1 width:220 align:#center toolTip:TOOLTIP_FURNITURE_TOOL_ADDCUSTOMSEAT_BUTTON enabled: false
		)
		
		group ""
		(
			button F_addCatcher "+  Catcher" width:110 across:2 align:#right toolTip:TOOLTIP_FURNITURE_TOOL_ADDCATCHER_BUTTON enabled:false
			button F_addHandle "+  Handle" width:100 align:#left toolTip:TOOLTIP_FURNITURE_TOOL_ADDHANDLE_BUTTON enabled:false
			button F_removeCatcher "-  Catcher"	width:110 across:2 align:#right toolTip:TOOLTIP_FURNITURE_TOOL_REMOVECATCHER_BUTTON enabled:false
			button F_removeHandle "-  Handle" width:100 align:#left toolTip:TOOLTIP_FURNITURE_TOOL_REMOVEHANDLE_BUTTON enabled:false
		)
		
		group ""
		(
			button F_createRoomST "Create See Through Node" width:220 align:#center enabled:false
			button F_linkRoomST "Link To SeeThrough" width:220 align:#center enabled:false
			button F_unlinkRoomST "Unlink From SeeThrough" width:220 align:#center enabled:false
		)
		
		group ""
		(
			button setSelectedForExportButton "Set Selected For Export" width:220 align:#center enabled:false toolTip:TOOLTIP_FURNITURE_TOOL_SETSELECTEDFOREXPORT
			editText animationName "Furniture Animation Name:" labelOnTop:true text:"" width:220 align:#center toolTip:TOOLTIP_FURNITURE_TOOL_ANIMNAME
			button animNameFromObject "Animation Name From Selected" width:220 align:#center enabled:false toolTip:TOOLTIP_FURNITURE_TOOL_ANIMNAMEFROMOBJECT
			label spacer03 "" height:1 width:0
		)
		
		-- FURNITURE TOOL Rollout actions
		on createFurnitureButton pressed do --Create Furniture
		(
			if not canToolBeActiveOrEnabled furnitureTool then
				messageBox ("ERROR: Furniture cannot be created because there is another tool already active!") title:TOOLNAME_FURNITURE
			else
			(
				createFurniture()
				max zoomext sel
				furnitureTool.setState toolState_ACTIVE
				updateToolkitSceneAndUIState()
			)
		)
		
		-- Seat controls
		on F_addSittingButton pressed do --Add Sitting
			addSeat NODE_NAME_SEAT_SITTING furnitureTool
		on F_addStandingButton pressed do --Add Standing
			addSeat NODE_NAME_SEAT_STANDING furnitureTool
		on F_addIdleButton pressed do --Add Idle
			addSeat NODE_NAME_SEAT_IDLE furnitureTool
		on F_addCustomButton pressed do --Add Custom
			addSeat F_customText.text furnitureTool
		
		-- Seat Catch/Pitcher and Handle controls
		on F_addCatcher pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_CATCHER_PITCHER furnitureTool
		on F_addHandle pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_HANDLE furnitureTool
		on F_removeCatcher pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_CATCHER_PITCHER furnitureTool addNewNode:false
		on F_removeHandle pressed do
			for obj in selection as array do
				addRemoveSeatChild obj nodeType_HANDLE furnitureTool addNewNode:false
			
		-- See through node controls
		on F_createRoomST pressed do
			createSeeThroughNode furnitureTool
		on F_linkRoomST pressed do
			linkSeeThroughNode()
		on F_unlinkRoomST pressed do
			unlinkSeeThroughNode()

		on setSelectedForExportButton pressed do
		(
			local rootNode = getNodeByName NODE_NAME_FURNITURE_NODES_PARENT
			if rootNode == undefined then rootNode = getNodeByName NODE_NAME_FURNITURE_ROOT
			if rootNode == undefined then messageBox ("Could not find objects '" + NODE_NAME_FURNITURE_NODES_PARENT + "' or '" + NODE_NAME_FURNITURE_ROOT + "' to parent selected object(s) to!") title:"Furniture Tool"
			else setSelectedForExport rootNode EXPORT_LAYER_NAME_FURNITURE
		)
		on animationName entered val do rootNode.toolkitData.furnitureAnimName = val
		on animNameFromObject pressed do
		(
			local selObjArray = selection as array
			if selObjArray.count > 0 then
			(
				rootNode.toolkitData.furnitureAnimName = animationName.text = selObjArray[1].name
			)
			else messageBox "Nothing Selected!"
		)
	) -- END FURNITURE TOOL Rollout

	
	-- ATTACHMENT TOOL Rollout
	rollout attachmentRollout "Attachment Tool" rolledup:true 
	(
		-- UI definition
		group ""
		(
			button createAttachmentButton "Create Attachment" width:220 toolTip:TOOLTIP_ATTACHMENT_TOOL_CREATE_BUTTON
			label spacer01 "" height:1 width:0
		)

		group ""
		(
			label boneLabel "Bone:" align:#left offset:[20, 0] width:30 across:2 enabled:false
			pickbutton pickBone "Pick Bone" align:#right offset:[-20, 0] width:190 filter:isBone message:"Select a Bone" enabled:false

			--radio buttons setupType "Setup Type:" align:#left offset:[20, 0] labels:#("Mesh(es)","Skeleton") offsets:#([0, 0], [10, 0]) enabled:false
			checkbox setupOrientationCheckbox "Reorient On Setup" align:#left offset:[20, 0] toolTip:TOOLTIP_ATTACHMENT_TOOL_SETUPORIENTATION_CHECKBOX enabled:false
			button setupButton "Setup Attachment" width:220 toolTip:TOOLTIP_ATTACHMENT_TOOL_SETUP_BUTTON enabled:false
			button resetAttachmentButton "Reset Attachment" width:220 toolTip:TOOLTIP_ATTACHMENT_TOOL_RESET_BUTTON enabled:false
			label spacer02 "" height:5 width:0
		)

		-- ATTACHMENT TOOL Rollout actions
		on createAttachmentButton pressed do
		(
			if not canToolBeActiveOrEnabled attachmentTool then
				messageBox ("ERROR: Attachment cannot be created because there is another tool already active!") title:TOOLNAME_ATTACHMENT
			else
			(
				createAttachment()
				max zoomext sel
				attachmentTool.setState toolState_ACTIVE
				updateToolkitSceneAndUIState()
			)
		)
		
		on pickBone picked obj do
		(
			-- TODO: Can the pick box be emptied leaving the setupButton.enabled == true?
			pickBone.text = (obj.name as string)
			setupButton.enabled = true
		)

		on setupButton pressed do setUpAttachment pickBone.text setupOrientationCheckbox.checked

		on resetAttachmentButton pressed do resetAttachment()
	) -- END ATTACHMENT TOOL Rollout


	-- CLOTHING TOOL Rollout
	rollout clothingRollout "Clothing Tool" rolledup: true
	(
		-- UI definition
		group ""
		(
			button createClothingButton "Create Clothing" width:220 toolTip:TOOLTIP_CLOTHING_TOOL_CREATE_BUTTON
			label spacer01 "" height:1 width:0
		)
		
		group ""
		(
			button bindButton "Bind to Skeleton" width:220 toolTip:TOOLTIP_CLOTHING_TOOL_BIND_BUTTON enabled:false
			button paintWeights "Paint Weights" width:110 toolTip:TOOLTIP_CLOTHING_TOOL_PAINTWEIGHTS_BUTTON across:2 align:#right enabled:false
			button poseRig "Pose" width:110 toolTip:TOOLTIP_CLOTHING_TOOL_POSE_BUTTON align:#left enabled:false
			button limitButton "Limit Bones" width:220 toolTip:TOOLTIP_CLOTHING_TOOL_LIMITBONES_BUTTON enabled:false
			label spacer02 "" height:1 width:0
		)
		
		
		-- CLOTHING TOOL Rollout actions
		on createClothingButton pressed do
		(
			if not canToolBeActiveOrEnabled clothingTool then
				messageBox ("ERROR: Clothing cannot be created because there is another tool already active!") title:TOOLNAME_CLOTHING
			else
			(
				createClothing()
				max zoomext sel
				clothingTool.setState toolState_ACTIVE
				updateToolkitSceneAndUIState()
			)
		)
		
		on bindButton pressed do bindClothing()

		on poseRig pressed do
		(
			viewport.SetRenderLevel #wireframe
			subobjectLevel = 0
		)

		on paintWeights pressed do paintWeightsMode()

		on LimitButton Pressed do limitBoneWeights()
	) -- END CLOTHING TOOL Rollout 


	-- ANIMATION TOOL Rollout
	rollout avatarAnimationRollout "Avatar Animation Tool" rolledup:true 
	(
		-- UI definition
		group ""
		(
			button createAvatarAnimationButton "Create Avatar Animation" width:220 toolTip:TOOLTIP_AVATARANIM_TOOL_CREATE_BUTTON
			label spacer01 "" height:1 width:0
		)
		
		group""
		(
			checkbox makePose "Make Pose" width:220 align:#center toolTip:TOOLTIP_AVATARANIM_TOOL_MAKEPOSE
			checkbox exportAllBones "Export All Bones" width:220 align:#center checked:true toolTip:TOOLTIP_AVATARANIM_TOOL_EXPORTALLBONES
			label spacer02 "" height:1 width:0
		)

		group ""
		(
			spinner animationLength "Animation Length    " width:220 align:#center range:[1, 1000, 50] type:#integer
			editText animationName "Avatar Animation Name:" labelOnTop:true text:"" width:220 align:#center toolTip:TOOLTIP_AVATARANIM_TOOL_ANIMNAME
			label spacer03 "" height:1 width:0
		)
		
		group ""
		(
			checkbox armsIK ANIM_IK_ARMS_CAPTION align:#right checked:true across:2 toolTip:TOOLTIP_AVATARANIM_TOOL_USEIK
			checkbox legsIK ANIM_IK_LEGS_CAPTION align:#left checked:true toolTip:TOOLTIP_AVATARANIM_TOOL_USEIK
			label spacer04 "" height:1 width:0

			--checkbox rArmIK "Right Arm" align:#right checked:true across:2
			--checkbox lArmIK "Left Arm" align:#left checked:true
			--checkbox rLegIK "Right Leg" align:#right checked:true across:2
			--checkbox lLegIK "Left Leg" align:#left checked:true
		)

		
		-- TODO: This is not yet implemented
		/*
		group ""
		(
			label RotIso "Rotation Isolation"
			checkbox lArmRotIso "Left Arm" across:2
			checkbox rArmRotIso "Right Arm"
			checkbox lLegRotIso "Left Leg" across:2
			checkbox rLegRotIso "Right Leg"
		)
		*/
		

		group ""
		(
			button selectAllRigWidgets "Select All Rig Widgets" width:220 toolTip:TOOLTIP_AVATARANIM_TOOL_SELECTALLRIGWIDGETS_BUTTON
			button deleteAvatarKeysButton "Delete Avatar Keys" width:220 toolTip:TOOLTIP_AVATARANIM_TOOL_DELETEAVATARKEYS_BUTTON
			button deleteAllAvatarKeysButton "Delete All Avatar Keys" width:220 toolTip:TOOLTIP_AVATARANIM_TOOL_DELETEALLAVATARKEYS_BUTTON
			button resetPose "Reset Pose" width:220 toolTip:TOOLTIP_AVATARANIM_TOOL_RESETPOSE_BUTTON
			label spacer05 "" height:1 width:0
		)

		-- ANIMATION TOOL Rollout actions
		on createAvatarAnimationButton pressed do
		(
			if not canToolBeActiveOrEnabled avatarAnimTool then
				messageBox ("ERROR: Attachment cannot be created because there is another tool already active!") title:TOOLNAME_AVATARANIM
			else
			(
				createAvatarAnimation()
				max zoomext sel
				avatarAnimTool.setState toolState_ACTIVE
				updateToolkitSceneAndUIState()
			)
		)

		on makePose changed val do
		(
			exportAllBones.checked = val
			exportAllBones.enabled = not val
		)
		
		on animationLength changed val do animationRange = interval 0 val

		on animationName entered val do rootNode.toolkitData.avatarAnimName = val
		
		on armsIK changed val do
		(
			clearSelection()
			avatarAnimToolVars.setKinematicsMode armsIK.caption val
		)

		on legsIK changed val do
		(
			clearSelection()
			avatarAnimToolVars.setKinematicsMode legsIK.caption val
		)

		-- TODO: This is not yet implemented
		/*
		on lArmRotIso changed val do setRotationIsolationMode lArmRotIso.name val
		on rArmRotIso changed val do setRotationIsolationMode rArmRotIso.name val
		on lLegRotIso changed val do setRotationIsolationMode lLegRotIso.name val
		on rLegRotIso changed val do setRotationIsolationMode rLegRotIso.name val
		*/

		on selectAllRigWidgets pressed do select avatarAnimToolVars.allWidgets 

		on deleteAvatarKeysButton pressed do deleteAvatarKeys atFrames:#(sliderTime)

		on deleteAllAvatarKeysButton pressed do deleteAvatarKeys()

		on resetPose pressed do rigWidgetTransformsToZero()
	) -- END ANIMATION TOOL Rollout


	-- EXPORT TOOL Rollout
	rollout exportRollout TOOLNAME_EXPORT rolledup:false
	(
		button exportButton "Export" width:220 toolTip:TOOLTIP_EXPORT_TOOL_EXPORT_BUTTON
		button openDirButton "Open Export Directory" width:220 toolTip:TOOLTIP_EXPORT_TOOL_OPENDIR_BUTTON
		button helpButton "Help" width:220 toolTip:TOOLTIP_EXPORT_TOOL_HELP_BUTTON
		
		label spacer01 ""  width:0 height:5
		button versionButton "" width:150 toolTip:TOOLTIP_EXPORT_TOOL_VERSION_BUTTON enabled: false
		
		-- EXPORT TOOL Rollout actions
		on openDirButton pressed do ShellLaunch "explorer" maxFilePath
		on exportButton pressed do doExport()
		on helpButton pressed do ShellLaunch "https://create.imvu.com/articles/studio/using-imvu-studio-toolkit-for-3ds-max/" "" -- TODO: Get this from the mzpbuilderini 'support'
		on versionButton pressed do ShellLaunch "https://create.imvu.com/articles/studio/toolkit/#3ds-max" "" -- TODO: Get this from the mzpbuilderini 'version'

		-- Called when Toolkit is close. Does not get called when the export tool is "rolled up"
		on exportRollout close do clearAllCallbacks()
	)

	
	-- TOOL  MAIN -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
	
	-- Main Function
	-- Global function that starts the toolkit. Must be called like this '::startImvuStudioToolkit()'
	-- TODO: When/if I refactor ImvuStudioToolkit to be a singleton(?) struct, this will turn into a public function
	-- Parameters: None
	-- Returns: None
	global startImvuStudioToolkit
	fn startImvuStudioToolkit =
	(
		-- Destroy the Toolkit floater window if it already exists
		try (closeRolloutFloater imvuStudioToolkitFloater) catch()

		-- Block Toolkit from starting if system units are not set correctly. I think this is the behavior we want.
		if imvuTools.areSystemUnitsCorrect() then
		(
			-- Create Toolkit floater window
			imvuStudioToolkitFloater = newrolloutfloater (TOOLKIT_NAME + "  " + LOCAL_VERSION) DEFAULT_WINDOW_WIDTH DEFAULT_WINDOW_HEIGHT
			addrollout visualsRollout imvuStudioToolkitFloater
			addrollout roomRollout imvuStudioToolkitFloater
			addrollout furnitureRollout imvuStudioToolkitFloater
			addrollout attachmentRollout imvuStudioToolkitFloater
			addrollout clothingRollout imvuStudioToolkitFloater
			addrollout avatarAnimationRollout imvuStudioToolkitFloater
			addrollout exportRollout imvuStudioToolkitFloater
			
			-- Create tool objects
			visualsTool = ToolkitTool id:toolId_VISUALS toolRollout:visualsRollout controlsEnabledWhenToolEnabled:#(visualsRollout.showNames,visualsRollout.hideNames) controlsDisabledWhenToolActive:#(visualsRollout.displayAvatar)
			roomTool = ToolkitTool id:toolId_ROOM toolRollout:roomRollout controlsEnabledWhenToolEnabled:#(roomRollout.createRoomButton) controlsDisabledWhenToolActive:#(roomRollout.createRoomButton) exportLayerName:EXPORT_LAYER_NAME_ROOM
			furnitureTool = ToolkitTool id:toolId_FURNITURE toolRollout:furnitureRollout controlsEnabledWhenToolEnabled:#(furnitureRollout.createFurnitureButton) controlsDisabledWhenToolActive:#(furnitureRollout.createFurnitureButton) exportLayerName:EXPORT_LAYER_NAME_FURNITURE
			attachmentTool = ToolkitTool id:toolId_ATTACHMENT toolRollout:attachmentRollout controlsEnabledWhenToolEnabled:#(attachmentRollout.createAttachmentButton) controlsDisabledWhenToolActive:#(attachmentRollout.createAttachmentButton) exportLayerName:EXPORT_LAYER_NAME_ATTACHMENT
			clothingTool = ToolkitTool id:toolId_CLOTHING toolRollout:clothingRollout controlsEnabledWhenToolEnabled:#(clothingRollout.createClothingButton) controlsDisabledWhenToolActive:#(clothingRollout.createClothingButton) exportLayerName:EXPORT_LAYER_NAME_CLOTHING
			avatarAnimTool = ToolkitTool id:toolId_AVATARANIM toolRollout:avatarAnimationRollout controlsEnabledWhenToolEnabled:#(avatarAnimationRollout.createAvatarAnimationButton) controlsDisabledWhenToolActive:#(avatarAnimationRollout.createAvatarAnimationButton) exportLayerName:EXPORT_LAYER_NAME_AVATAR_ANIMATION
			exportTool = ToolkitTool id:toolId_EXPORT toolRollout:exportRollout controlsEnabledWhenToolEnabled:#(exportRollout.helpButton)

			allTools = #(visualsTool, roomTool, furnitureTool, attachmentTool, clothingTool, avatarAnimTool, exportTool)
			
			-- Tool compatibility - Tools that can be run in conjunction with each other
			visualsTool.compatibleTools = #(roomTool, furnitureTool, attachmentTool, clothingTool, avatarAnimTool)
			roomTool.compatibleTools = #(visualsTool)
			furnitureTool.compatibleTools = #(visualsTool, avatarAnimTool)	
			attachmentTool.compatibleTools = #(visualsTool)
			clothingTool.compatibleTools = #(visualsTool)
			avatarAnimTool.compatibleTools = #(visualsTool, furnitureTool)
			exportTool.compatibleTools = #(visualsTool, roomTool, furnitureTool, attachmentTool, clothingTool, avatarAnimTool)

			setToolkitCallbacks()
			setToolkitToolVariables()
			updateToolkitSceneAndUIState updateAfter:updateAfter_TOOLSTART
			checkForUpdate()
		)
	)
)
