/*------------------------------------------------------------------------------------------
//  Creation Date:  8/7/24
//  Author:		Greg Smith (gks)
//  Name:		ImvuToolsLib.ms
//  Description:  This script is a collection of tools and functions which are meant
//		to be accessible between various Imvu tools such as ImvuStudioToolkit and the ImvuMenu
//  Limitations:
//  Revisions:
------------------------------------------------------------------------------------------*/
(
	struct ImvuToolsLibrary
	(
		private
		-- Required system unit settings for Toolkit to work correctly. Toolkit relies on these for consistency from IMVUMegaScene to import into Studio
		-- These units also make moving assets back and forth between Blender and Maya consistent without issue with scale.
		SYSTEM_UNIT_TYPE = #meters,
		SYSTEM_UNIT_SCALE = 1.0,

		/* -- If we ever need a constructor
		-- Constructor
		fn initToolsLib =
		(
			--format "TOOLS INIT\n"
		),
		init = initToolsLib(),
		*/


		--	Description:
		--		Checks for an object being negatively scale along any axis
		--	Parameters:
		--		<node> obj: any max scene object
		--	Returns: <boolean> True if the object has a negative scale value for any axis, false if not
		function isNegativelyScaled obj =
		(
			obj.scale.x < 0.0 or obj.scale.y < 0.0 or obj.scale.z < 0.0
		),


		public
		printDebug = true,


		--	Description:
		--		Prints a message to the console with the optional message type and function name from where it was called.
		--	Parameters:
		--		msgArray <array> Things to print
		--		<name> msgType [optional]: Valid types are #ERROR, #WARN, #INFO, and #DEBUG. Defaults to no label
		--		addSpaces <boolean> [optional] Defaults to true and Will automatically add spaces between things to print
		--		<string> inScript [optional]: script name/path from where it was called. e.g.: filenameFromPath(getThisScriptFilename())
		--		<function> inFunction [optional]: function from where it was called
		--	Returns: None
		function printMsg msgArray msgType: addSpaces:true inScript: inFunction: =
		(
			local msg
			local msgStr = ""
			case msgType of
			(
				#ERROR: msgStr = "ERROR: "
				#WARN: msgStr = "WARNING: "
				#INFO: msgStr = "INFO: "
				#DEBUG: msgStr = "DEBUG: "
				default: msgStr = ""
			)

			if (printDebug and msgType == #DEBUG) or msgType != #DEBUG then
			(
				for msg in msgArray do
				(
					msgStr += (msg as string)
					if addSpaces then msgStr += " "
				)
				if inFunction != unsupplied then msgStr = (inFunction as string) + ": " + msgStr
				if inScript != unsupplied then msgStr = (inScript as string) + ": " + msgStr
				format "%\n" msgStr
			)
		),
		

		--	Description:
		--		Converts input to an array. Primarily used handle objects and selections it will also create an array of one if
		--			supplied any variable type, string, float, etc.
		--	Parameters: <type-free_var> &input (passed by reference)
		--	Returns: None
		function toArray &input =
		(
			if not classof input == Array then
			(
				if classof input == ObjectSet then input = input as array
				else input = #(input)
			)
		),


		--	Description:
		--		Returns a new array that is the interection of the supplied arrays 
		--	Parameters:
		--		<array> sourceArray - An array of arrays or items to perform the intersection on
		--	Returns: <array> The interection of the supplied arrays 
		function intersectionOfArrays sourceArray = if classof sourceArray == Array then
		(
			local element
			local i, ii, j, jj
			local itemCount
			local itemFound
			local returnArray = #()
			
			-- Make sure all source array elements are converted to arrays and all duplicate elements have been removed 
			for element in sourceArray do
			(
				toArray &element
				element = makeUniqueArray element
			)

			for i = 1 to sourceArray.count do
			(
				for ii = 1 to sourceArray[i].count do
				(
					itemCount = 1
					for j = sourceArray.count to i+1 by -1 do
					(
						itemFound = false
						for jj = 1 to sourceArray[j].count while not itemFound do
						(
							if sourceArray[i][ii] == sourceArray[j][jj] then
							(
								itemFound = true
								itemCount += 1
							)
						)
					)
					if itemCount == sourceArray.count then append returnArray sourceArray[i][ii]
				)
			)
			returnArray
		),


		--	Description:
		--		Removes all occurrences the supplied item or array of items from the source array
		--	Parameters:
		--		<array> &sourceArray (passed by reference) The array to be modified
		--		<type-free_var> remove - The item(s) to be removed. Will convert a single item to an array
		--	Returns: None
		function removeFromArray &sourceArray remove =
		(
			local item, itemIndex
			
			toArray &remove
			remove = makeUniqueArray remove
			for item in remove do while (itemIndex = findItem sourceArray item) != 0 do deleteItem sourceArray itemIndex
		),


		--	Description:
		--		Checks if supplied transform matrix is a zero transform matrix
		--		Since matrix3 comparisons, e.g.: <matrix3> == <matrxi3> are supported only in 3ds Max 2018 and beyond, we have this function for 2017 support.
		--	Parameters:
		--		<matrix3> objTransformMatrix: any max scene object
		--	Returns: <boolean> True if the object has a negative scale value for any axis, false if not
		function isZeroTransformMatrix objTransformMatrix =
		(
			local ZERO_TRANSFROM_MATRIX = (matrix3 [1,0,0] [0,1,0] [0,0,1] [0,0,0])
			local isZeroTransform = true

			for ii = 1 to 4 while isZeroTransform do isZeroTransform = objTransformMatrix[ii] == ZERO_TRANSFROM_MATRIX[ii]
			isZeroTransform
		),


		--	Description:
		--		Check if 3ds Max System Units are set correctly and alerts user with a message box informing so.
		--	Parameters: None
		--	Returns: <boolean> True if units are set correctly, false if not
		function areSystemUnitsCorrect =
		(
			local sysUnitsAreCorrect = (units.SystemType == SYSTEM_UNIT_TYPE and units.SystemScale == SYSTEM_UNIT_SCALE)
			local msg = "WARNING!  System unit settings are not compatible with IMVU Studio Toolkit!\n"
			local MESSAGEBOX_TITLE = "System Units Check"

			if not sysUnitsAreCorrect then
			(
				if not units.SystemType == SYSTEM_UNIT_TYPE then msg += ("System Unit Type is set to " + units.SystemType + " and should be " + SYSTEM_UNIT_TYPE + "\n")
				if not units.SystemScale == SYSTEM_UNIT_SCALE then msg += ("System Unit Scale is set to " + (units.SystemScale as string) + " and should be " + (SYSTEM_UNIT_SCALE as string))

				messageBox msg title:MESSAGEBOX_TITLE icon:#warning
			)
			sysUnitsAreCorrect
		),
	

		--	Description:
		--		Wrapper for evaluating a script and reporting errors
		--	Parameters:
		--		scriptPath <string> Path to the script to be evaluated
		--	Returns: None
		function evalScript scriptPath =
		(
			if doesFileExist scriptPath then fileIn scriptPath
			else print ("ERROR: Could not locate \"" + scriptPath + "\"")
		),


		--	Description:
		--		Filter a list of objects and return only the "root most nodes" 
		--	Parameters:
		--		objArray <array> Array of nodes
		--	Returns: <array> Filtered array of "root most" nodes
		fn getRootNodes objArray =
		(
			local rootNodes = #()
			local obj, objParent
			local parentFound = false

			for obj in objArray do
			(
				-- If this node has a parent and ANY parent up the hierarchy is in objArray then skip it
				parentFound = false
				objParent = obj.parent
				while objParent != undefined and not parentFound do
				(
					if findItem objArray ObjParent > 0 then parentFound = true
					objParent = objParent.parent
				)
				if not parentFound then appendIfUnique rootNodes obj
			)
			rootNodes
		),


		--	Description:
		--		Build a list of objects by hierarchy.
		--	Parameters:
		--		parentObj <node> Root/parent node to begin get on
		--		&objArray <array> of nodes that gets passed by reference to be populated
		--		sortDepthFirst <boolean> [optional]: Defaults to false if not provided
		--	Returns: None
		function getHierarchy parentObj &objArray sortDepthFirst:false =
		(
			local child

			if parentObj.children.count > 0 then
				for child in parentObj.children do getHierarchy child &objArray sortDepthFirst:sortDepthFirst
			if sortDepthFirst then
				append objArray parentObj
			else
				insertItem parentObj objArray 1
		),


		--	Description:
		--		Get a list of valid morph targets from a morpher modifier.
		--	Parameters:
		--		morphModifier <modifier> The morph modifier to parse
		--	Returns: <array> of morph target nodes
		function getMorphTargetsFromMorphModifier morphModifier =
		(
			local morphTargets = #()

			if classof morphModifier == Morpher then
			(
				channelCount = WM3_NumberOfChannels morphModifier
				morphTargets = for ii = 1 to channelCount where WM3_MC_IsActive morphModifier ii and WM3_MC_HasTarget morphModifier ii collect (WM3_MC_GetTarget morphModifier ii)
			)
			morphTargets
		),


		--	Description:
		--		Perform validation of a "Morph Animated Object" (object with a "morpher" modifier) to be sure it is set up correctly for morph animation.
		--		NOTE: Morph channels which are not active are ignored
		--		Some info on correct morph animation setup for FBX export:
		--			Multiple morph modifiers are completely okay! The resulting FBX will contain the combined morph result!
		--			NOTE: Historically morph animations were exported directly to Cal3D format as XPF files using Cal3D exporters. The most recent
		--				supported version of 3ds Max for these exporters is 2017. The following info is valid for both FBX and Cal3D (XPF) export.
		--			- Must have at least one morph target mesh that has the exact same number and order of vertices.
		--			- Must have a 'Morpher' modifier with assigned morph target(s) and keyed target weight(s) in the channel list.
		--			- Must have a 'Skin' modifier and weighted to the avatar skeleton IF it is going to be imported to a clothing product.
		--				- Skin modifier must be the top most modifier, typically with the Morpher modifier directly below the Skin modifier
		--				- Attachment, Furniture, and Room products do not require a 'Skin' modifier. In this case the Morph Animated Object
		--					can be parented to any bone in the skeleton.
		--			- A morph target mesh(es) must have the same transforms as the Morph Animated Object.
		--			- Related to same transformations, the position of the Morph Animated Object and the morph target(s) in a hierarchy is important. There are two options:
		--				- Parenting the morph target mesh(es) to the Morph Animated Object (THIS IS THE IMVU CONVENTION).
		--					NOTE: There is an issue with this structure in that if a morph target object becomes up-parented after assignment in the morpher modifier
		--						and the user tries to re-parent the morph target mesh they will encounter an error: "Cannot link: would create a dependency loop".
		--				- Parenting the morph target mesh(es) and the Morph Animated Object to the same node (siblings). (this is likely the scene root for skinned objects).
		--			- Morph names in the channel list must follow an IMVU name suffix convention. Valid suffixes are: .Clamped .Average .Exclusive .Additive
		--				NOTE: Morph target mesh(es) do not need to follow this naming convention.
		--			- The morph target mesh(es) do not need to be exported. If they are it is not a problem because even if they are imported into IMVU Studio, they
		--				are not used by or submitted as part of the product.
		--	Parameters:
		--		node <node> object to be checked
		--	Returns: <boolean> True if object set up correctly, false if not.
		function isValidMorphObject node =
		(
			local MORPH_OBJECT = "Morph Object"
			local QUERYBOX_TITLE = "Validate Morpher"
			local VALID_MORPH_TARGET_SUFFIXES = #("Clamped", "Average", "Exclusive", "Additive")

			local child, obj
			local userMessage
			local valid = true
			local modifier, morphModifier
			local morphModifierCount = 0
			local hasValidNameSuffix
			local channelCount, targetCount
			local morphTarget, morphName, morphNameArray
			local nodeVertsCount, numMorphVerts, numVerts

			for modifier in node.modifiers do if classof modifier == Morpher then
			(
				--format "modifier name: %\n" modifier.name -- DEV_DEBUG

				morphModifierCount += 1
				morphModifier = modifier
				nodeVertsCount = node.verts.count
				channelCount = WM3_NumberOfChannels morphModifier
				targetCount = 0
				-- Validate the Morpher modifier setup
				-- Compare the lists of ACTIVE and VALID objects in the morph channel list, doing validation along the along the way, and compare it with the Morph Animated Object sibling and child object lists.
				for ii = 1 to channelCount where WM3_MC_IsActive morphModifier ii do
				(
					morphName = WM3_MC_GetName morphModifier ii
					morphNameArray = filterString morphName "."
					hasValidNameSuffix = if morphNameArray.count > 0 then findItem VALID_MORPH_TARGET_SUFFIXES morphNameArray[morphNameArray.count] != 0 else false

					if WM3_MC_HasTarget morphModifier ii then
					(
						targetCount += 1
						morphTarget = WM3_MC_GetTarget morphModifier ii
						-- Not sure what the difference is between these two? And why they wouldn't be the same as morphTarget.verts.count
						numMorphVerts = WM3_MC_NumMPts morphModifier ii
						numVerts = WM3_MC_NumPts morphModifier ii
	
						-- Has valid name and has target (GOLDEN)
						if hasValidNameSuffix then
						(
							-- Check if morph target object is a child OR sibling of the Morph Animated Object. This is NOT a requirement, but makes for consistent successful results.
							--	The target object "could" be somewhere else in the scene hierarchy, but if the Morph Animated Object is skinned or in an animated skeletal hierarchy things will break once imported into IMVU Studio
							--	The target object could also be descendant in a hierarchy below the Morph Animated Object and that "might" be okay, but it would be better to enforce a standard hierarchy.
							if (findItem node.children morphTarget != 0) or (morphTarget.parent == node.parent) then
							(
								-- Make sure the morph target is geometry and is valid
								if superclassOf morphTarget == GeometryClass and WM3_MC_IsValid morphModifier ii then
								(
									-- CRUDE CHECK: Make sure all vert counts match the vertices count of the Morph Animated Object
									--if nodeVertsCount != numMorphVerts or nodeVertsCount != numVerts or nodeVertsCount != morphTarget.verts.count  then
									if nodeVertsCount != morphTarget.verts.count  then
									(
										userMessage = ("Morph target \"" + morphTarget.name + "\" assigned to Morph name \"" + morphName + "\" does not have the same number of vertices as the " + MORPH_OBJECT + " \"" + node.name + "\"\n\n")
										userMessage = (userMessage + node.name + ": " + nodeVertsCount as string + "\n")
										--userMessage = userMessage + "Morph  vertices:" + numMorphVerts as string + "\n"
										--userMessage = userMessage + "Vertices:" + numVerts as string + "\n"
										userMessage = (userMessage + morphTarget.name + ": " + morphTarget.verts.count as string)
										userMessage = (userMessage + "\n\nDo you wish to continue?")
										valid = queryBox userMessage title:(QUERYBOX_TITLE)
										printMsg #(userMessage) msgType:ERROR addSpaces:false
									)
								)
								else
								(
									userMessage = ("Morph target \"" + morphTarget.name + "\" assigned to Morph name \"" + morphName + "\" is not a mesh object or valid morph target!")
									valid = queryBox (userMessage + "\n\nDo you wish to continue?") title:QUERYBOX_TITLE
									printMsg #(userMessage) msgType:ERROR addSpaces:false
								)
							)
							else
							(
								userMessage = ("Morph target \"" + morphTarget.name + "\" assigned to Morph name \"" + morphName + "\" is not parented to \"" + node.name + "\" or have the same parent as \"" + node.name + "\"")
								valid = queryBox (userMessage + "\n\nDo you wish to continue?") title:QUERYBOX_TITLE
								printMsg #(userMessage) msgType:ERROR addSpaces:false
							)
						)
						else -- Has invalid name and has target (REPORT)
						(
							userMessage = ("Morph name \"" + morphName + "\" of " + MORPH_OBJECT + " \"" + node.name + "\" does not appear to be named correctly!  Valid names are:\n<morph_name>.Clamped\n<morph_name>.Average\n<morph_name>.Exclusive\n<morph_name>.Additive")
							valid = queryBox (userMessage + "\n\nDo you wish to continue?") title:QUERYBOX_TITLE
							printMsg #(userMessage) msgType:ERROR addSpaces:false
						)
					)
					else
					(
						-- Has valid name and missing target (REPORT)
						if hasValidNameSuffix then
						(
							userMessage = ("Morph name \"" + morphName + "\" of " + MORPH_OBJECT + " \"" + node.name + "\" is missing a morph target")
							valid = queryBox (userMessage + "\n\nDo you wish to continue?") title:QUERYBOX_TITLE
							printMsg #(userMessage) msgType:ERROR addSpaces:false
						)
					)
					-- Has invalid name and missing target (IGNORE)
				)

				-- Has no active channels with targets (REPORT)
				if targetCount == 0 then
				(
					userMessage = ("Morpher modifier \"" + morphModifier.name + "\" of " + MORPH_OBJECT + " \"" + node.name + "\" has no active morphs with targets!")
					valid = queryBox (userMessage + "\n\nDo you wish to continue?") title:QUERYBOX_TITLE
					printMsg #(userMessage) msgType:ERROR addSpaces:false
				)
			)

			if morphModifierCount == 0 then
			(
				valid = false
				userMessage = ("Object \"" + node.name + "\" does not have a Morpher modifier!")
				printMsg #(userMessage) msgType:WARN addSpaces:false
			)

			valid
		),


		--	Description:
		--		Performs a reset of the transforms for each of the nodes in the supplied list.
		--		Zeros out position, rotation, scale without altering the object
		-- 		Resetting transforms will have unpredictable results when the selection involves a hierarchy!!!!
		--		Especially if the selection is cherry picked from a hierarchy. In this case the best option is
		--		to reset ALL objects in the hierarchy starting with selected child closest to the root. e.g.:
		--			root_node (not selected)
		--				child0 (selected - Reset from here down, even the un-selected nodes)
		--					child1 (not selected)
		--						child2 (selected)
		--		There is a potential use case where the user might want to preserve the transforms of the parent and above nodes. So we won't reset the entire hierarchy up the hierarchy to the root parent.
		-- Parameters:
		--		objArray <array> [optional] of nodes. If not supplied uses the current selection.
		--		mode <name> [optional] #QUERY just does the validation and does not perform the reset. #FORCE bypasses the validation and just does the reset
		-- Returns: <boolean> True if all objects in objArray are successfully reset. False if any one fails or if any issues are found.
		function resetTransforms suppliedObjArray:(selection as array) mode: =
		(
			local PRINT_DEBUG = false
			local QUERYBOX_TITLE = "Reset XForms"

			local obj, node
			local rootNodes = #()
			local objArray = #()
			--local filteredObjectArrayHasHierarchy = false CRUFT
			local modifier
			local msgStr = ""
			local halt = false
			local hasAncestorWithTransform
			local objArrayCulled = #()
			local objectsSorted = #()
			local objCount
			local ntm, piv, xformMod

			if PRINT_DEBUG then
			(
				format "resetTransforms() suppliedObjArray:\n" -- DEV_TEMP
				for obj in suppliedObjArray do format "\t%\n" obj.name -- DEV_TEMP
				format "\n" -- DEV_TEMP
			)

			-- Make sure there is something to reset
			if suppliedObjArray.count == 0 then
			(
				if not mode == #QUERY then messageBox "Nothing to process!" title:QUERYBOX_TITLE
				halt = true
			)

			-- Get root nodes - This will cull child objects cherry picked from a hierarchy and we want to be acting on the root most nodes
			if not halt then rootNodes = getRootNodes suppliedObjArray

			if PRINT_DEBUG then
			(
				format "resetTransforms() rootNodes:\n" -- DEV_TEMP
				for obj in rootNodes do format "\t%\n" obj.name -- DEV_TEMP
				format "\n" -- DEV_TEMP
			)

			-- Act on the root nodes
			for node in rootNodes while not halt do
			(
				objArray = #()
				if node.children.count > 0 then
				(
					-- Query user
					if not mode == #FORCE then
						halt = not queryBox ("Object \"" + node.name + "\" has child objects.\n\nDo you want to attempt to reset the transforms on this object and all of it's children (recommended)?\nOnly objects with transforms will be reset.") title:QUERYBOX_TITLE

					if not halt then getHierarchy node &objArray
				)
				else objArray = #(node)
				hasAncestorWithTransform = false
				for obj in objArray while not halt do
				(
					if not isZeroTransformMatrix obj.transform then
					(
						--if obj.children.count > 0 then filteredObjectArrayHasHierarchy = true CRUFT
						hasAncestorWithTransform = true

						if not mode == #FORCE then
						(
							if isNegativelyScaled obj and superClassOf obj ==  GeometryClass then
							(
								halt = not queryBox ("Object \"" + obj.name + "\" has negative transforms! The mesh normals may get flipped when transforms are reset. Continue Yes/No?") title:QUERYBOX_TITLE
							)

							if not halt then
							(
								if obj.modifiers.count > 0 then
								(
									msgStr = "Object \"" + obj.name + "\" has transforms and modifiers. These modifiers will get destroyed when transforms are reset:\n\n"
									for modifier in obj.modifiers do msgStr += (modifier.name + "\n")
									msgStr += "\nContinue Yes/No?"
									halt = not queryBox msgStr title:QUERYBOX_TITLE
								)
							)
						)
					)
					if hasAncestorWithTransform then append objArrayCulled obj
				)
			)

			if not halt and not mode == #QUERY then
			(
				if objArrayCulled.count == 0 then printMsg #(QUERYBOX_TITLE, ": No objects found needing to have their transforms reset!") addSpaces:false
				else
				(
					-- Actually perform the reset
					for obj in objArrayCulled do
					(
						-- collapse the modifier stack to remove all existing modifiers
						-- NOTE: Collapsing the modifier stack "may" not be the same as sequentially (top down) performing a'deleteModifier' on each modifier in the modifier stack!
						if (classOf obj.baseObject == Editable_Poly) then convertTo obj Editable_Poly
						else if (classOf obj.baseObject == PolyMeshObject or classOf obj.baseObject == Editable_mesh) then convertToMesh obj
						else if (classOf obj.baseObject == line or classOf obj.baseObject == SplineShape) then convertTo obj SplineShape
						else collapseStack obj

						-- Get the original node transform matrix
						ntm = obj.transform
						
						-- The new object transform only inherits position
						obj.transform = transMatrix obj.pos
						
						-- Compute the pivot transform matrix
						piv = obj.objecttransform * inverse obj.transform
						
						-- Reset the object offsets to 0
						obj.pivot = [0,0,0]
						obj.objectoffsetPos = [0,0,0]
						obj.objectoffsetRot = (quat 0 0 0 1)
						obj.objectoffsetScale = [1,1,1]
						
						-- Apply the pivot transform matrix to the modified original node transform matrix
						-- to get the XForm gizmo transform
						ntm = piv * ntm
						
						-- apply an XForm modifier to the node
						xformMod = xform()
						addmodifier obj xformMod
						
						-- set the XForm modifier's gizmo tranform
						xformMod.gizmo.transform = ntm

						-- collapse the modifier stack to apply the xform modifier
						if (classOf obj.baseObject == Editable_Poly) then convertTo obj Editable_Poly
						else if (classOf obj.baseObject == PolyMeshObject or classOf obj.baseObject == Editable_mesh) then convertToMesh obj
						else if (classOf obj.baseObject == line or classOf obj.baseObject == SplineShape) then convertTo obj SplineShape
						else collapseStack obj
					)

					/* CRUFT - Don't think we need this because we are going through the hierarchy from the root object down.
					-- If resets were performed on a hierarchy then we need to go through and all objects to reset values.
					--	This will be a bit redundant for objects without children, but whatever.
					--	And we need to perform this in the reverse hierarchical order so that the bottom most child is set last
					if filteredObjectArrayHasHierarchy then
					(
						objCount = objArrayCulled.count
						for ii = 0 to (objCount - 1) do
						(
							obj = objArrayCulled[objCount - ii]
							obj.pos = [0,0,0]
							obj.rotation = (quat 0 0 0 1)
							obj.scale = [1,1,1]
						)
					)
					*/
				)
			)
			-- Return success or fail
			not halt
		),

		--	Description:
		--		Perform validation of an object or list of objects.
		--		These are some checks that apply to all objects that will end up getting exported and imported into IMVU Studio
		--	Parameters: <object_or_object_array> objects
		--	Returns: <boolean> True if object(s) are free of issues, false if not.
		function validateObjects objects =
		(
			local obj
			local hasModifier, modifier
			local valid = true
			local validMorphObj = true

			-- Convert our input to an array if a single object/node was passed
			toArray &objects

			-- We want to perform a check on all objects even if issues are found so we can notify the user of all issues
			for obj in objects do
			(
				hasModifier = false
				for modifier in obj.modifiers while not hasModifier do
				(
					if classof modifier == Morpher then
					(
						hasModifier = true
						-- We want to perform a check on all Morpher modifiers even if issues are found so we can notify the user of all issues
						validMorphObj = isValidMorphObject obj
						-- Only set the valid flag if it has not already been set to false
						if valid then valid = validMorphObj 
					)
				)
			)
			valid
		)
	)

	global imvuTools = ImvuToolsLibrary()
)